#include <string.h>
#include <time.h>
#ifdef HAVE_LIBENDEAVOUR2
# include <endeavour2.h>
#endif
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>

#include "../include/string.h"

#include "imgio.h"
#include "guiutils.h"
#include "guirgbimg.h"
#include "img_info_dlg.h"
#include "config.h"

#include "images/icon_ok_20x20.xpm"
#include "images/icon_info_32x32.xpm"


static const gchar *img_info_get_deliminated_size_string(const gulong i);
#ifdef HAVE_LIBENDEAVOUR2
static const gchar *img_info_dlg_get_date_string(
	EDVContext *edv2_ctx,
	const gulong t
);
#else
static const gchar *img_info_dlg_get_date_string(const gulong t);
#endif

static gint img_info_dlg_delete_event_cb(
	GtkWidget *widget, GdkEvent *event, gpointer data
);
static gint img_info_dlg_thumb_event_cb(
	GtkWidget *widget, GdkEvent *event, gpointer data
);
static gint img_info_dlg_comments_text_event_cb(
	GtkWidget *widget, GdkEvent *event, gpointer data
);
static void img_info_dlg_ok_cb(GtkWidget *widget, gpointer data);
static void img_info_dlg_thumb_draw(ImgInfoDlg *d);

ImgInfoDlg *img_info_dlg_new(GtkWidget *ref_toplevel);
static void img_info_dlg_set_image(
	ImgInfoDlg *d,
	const guint8 *rgba,
	const gint width, const gint height,
	const gint bpp, const gint bpl,
	const guint8 *bg_color			/* 4 bytes RGBA */
);
void img_info_dlg_set_values(
	ImgInfoDlg *d,
#ifdef HAVE_LIBENDEAVOUR2
	EDVContext *edv2_ctx,
#endif
	const guint8 *rgba,
	const gint width, const gint height,
	const gint bpp, const gint bpl,
	const gint nframes,
	const gint nloops,			/* -1 for infinite */
	const gulong play_time_ms,		/* Play time per loop in
											 * milliseconds */
	const gulong total_play_time_ms,	/* Total play time in
											 * milliseconds or (gulong)-1
											 * for infinite */
	const gint alpha_channel,               /* 0 = None
											 * 1 = Allocated But Uniform
											 * 2 = Allocated & Defined */
	const guint8 *bg_color,                 /* 4 bytes RGBA */
	const gint x, const gint y,             /* Offset on base */
	const gint base_width,                  /* Dimensions of base (can be 0 for no base) */
	const gint base_height,
	const gboolean has_changes,

	const gchar *file_name,
	const gchar *file_format,
	const gulong file_size,                 /* File size in bytes */
	const gulong last_modified,             /* Last modified time in
											 * seconds since EPOCH */
	const gchar *title,
	const gchar *author,
	const gchar *creator,
	const gchar *comments,
	const gchar *display,                   /* X Display address */
	const gchar *visual_type,               /* X Visual type */
	const gint display_depth                /* X Depth in bits per pixel */
);
void img_info_dlg_map(ImgInfoDlg *d);
void img_info_dlg_unmap(ImgInfoDlg *d);
void img_info_dlg_delete(ImgInfoDlg *d);


#define IMG_INFO_DLG_DEF_TITLE		"Information"
#define IMG_INFO_DLG_DISPLAY_MAX_WIDTH	200
#define IMG_INFO_DLG_DISPLAY_MAX_HEIGHT	200


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : TRUE)


/*
 *      Image Info Dialog:
 */
struct _ImgInfoDlg {

	GtkWidget	*toplevel;
	GtkAccelGroup	*accelgrp;
	GdkGC		*gc;

	GtkWidget	*main_vbox,
			*thumb_da,		/* GtkDrawingArea a thumb
						 * of the image */
			*image_parent,
			*image_parameters_label,
			*image_values_label,
			*file_parent,
			*file_parameters_label,
			*file_values_label,
			*display_parameters_label,
			*display_value_label,
			*comments_frame,
			*comments_text,
			*ok_btn;

	GdkPixmap	*image_pixmap;
	GdkBitmap	*image_mask;
};


/*
 *	Returns a statically allocated string describing the value i
 *	with ',' deliminators at each thousands place.
 */
static const gchar *img_info_get_deliminated_size_string(const gulong i)
{
#ifdef HAVE_LIBENDEAVOUR2
	return(edv_str_size_delim(i));
#else
	gint comma_countdown, slen;
	gchar		ss[256],
			*ss_ptr;
	static gchar	ts[256],
			*ts_ptr;

	(void)g_snprintf(
		ss, sizeof(ss),
		"%ld",
		i
	);
	slen = strlen(ss);

	/* 3 digits or less? (no commas needed) */
	if(slen <= 3)
	{
		(void)strcpy(
			ts,
			ss
		);
		return(ts);
	}

	ts_ptr = ts;
	ss_ptr = ss;

	/* Initialize comma counter */
	comma_countdown = slen % 3;
	if(comma_countdown <= 0)
		comma_countdown = 3;

	/* Iterate through size string until end is reached */
	while(*ss_ptr != '\0')
	{
		/* Reset comma counter and put in a comma? */
		if(comma_countdown <= 0)
		{
			*ts_ptr++ = ',';
			comma_countdown = 3;
		}

		*ts_ptr++ = *ss_ptr++;
		comma_countdown--;
	}
	 
	/* Null terminate return string */
	*ts_ptr = '\0';

	return(ts);
#endif
}


#ifdef HAVE_LIBENDEAVOUR2
static const gchar *img_info_dlg_get_date_string(
	EDVContext *edv2_ctx,
	const gulong t
)
{
	return(edv_date_format(
		edv2_ctx,
		t
	));
}
#else
static const gchar *img_info_dlg_get_date_string(const gulong t)
{
	const time_t _t = (time_t)t;
	gchar		*sp,
			*s = (gchar *)ctime(&_t);
	if(s == NULL)
		return("");

	sp = strchr(s, '\n');
	if(sp != NULL)
		*sp = '\0';

	return(s);
}
#endif

/*
 *	Toplevel GtkWindow "delete_event" signal callback.
 */
static gint img_info_dlg_delete_event_cb(
	GtkWidget *widget, GdkEvent *event, gpointer data
)
{
	ImgInfoDlg *d = IMG_INFO_DLG(data);
	if(d == NULL)
		return(FALSE);

	img_info_dlg_ok_cb(d->ok_btn, d);
	return(TRUE);
}

/*
 *	Display GtkDrawingArea any event signal callback.
 */
static gint img_info_dlg_thumb_event_cb(
	GtkWidget *widget, GdkEvent *event, gpointer data
)
{
	ImgInfoDlg *d = IMG_INFO_DLG(data);
	if((widget == NULL) || (event == NULL) || (d == NULL))
		return(FALSE);

	switch((int)event->type)
	{
	  case GDK_CONFIGURE:
		break;
	  case GDK_EXPOSE:
		img_info_dlg_thumb_draw(d);
		break;
	}

	return(TRUE);
}

/*
 *	Comments GtkText any event signal callback.
 */
static gint img_info_dlg_comments_text_event_cb(
	GtkWidget *widget, GdkEvent *event, gpointer data
)
{
	gint status = FALSE;
	GdkEventButton *button;
	GtkText *text;
	ImgInfoDlg *d = IMG_INFO_DLG(data);
	if((widget == NULL) || (event == NULL) || (d == NULL))
		return(status);

	text = GTK_TEXT(widget);

	switch((gint)event->type)
	{
	  case GDK_BUTTON_PRESS:
		button = (GdkEventButton *)event;
		switch(button->button)
		{
		  case GDK_BUTTON4:
			/* Scroll up */
			if(text->vadj != NULL)
			{
				GtkAdjustment *adj = text->vadj;
				const gfloat inc = MAX(
					(0.25f * adj->page_size),
					adj->step_increment
				);
				gfloat v = adj->value - inc;
				if(v > (adj->upper - adj->page_size))
					v = adj->upper - adj->page_size;
				if(v < adj->lower)
					v = adj->lower;
				gtk_adjustment_set_value(adj, v);
			}
			/* Need to mark the GtkText button as 0 or else it will
			 * keep marking
			 */
			text->button = 0;
			gtk_grab_remove(widget);
			gtk_signal_emit_stop_by_name(
				GTK_OBJECT(widget), "button_press_event"
			);
			status = TRUE;
			break;

		  case GDK_BUTTON5:
			/* Scroll down */
			if(text->vadj != NULL)
			{
				GtkAdjustment *adj = text->vadj;
				const gfloat inc = MAX(
					(0.25f * adj->page_size),
					adj->step_increment
				);
				gfloat v = adj->value + inc;
				if(v > (adj->upper - adj->page_size))
					v = adj->upper - adj->page_size;
				if(v < adj->lower)
					v = adj->lower;
				gtk_adjustment_set_value(adj, v);
			}
			/* Need to mark the GtkText button as 0 or else it will
			 * keep marking
			 */
			text->button = 0;
			gtk_grab_remove(widget);
			gtk_signal_emit_stop_by_name(
				GTK_OBJECT(widget), "button_press_event"
			);
			status = TRUE;
			break;
		}
		break;

	  case GDK_BUTTON_RELEASE:
		button = (GdkEventButton *)event;
		switch(button->button)
		{
		  case GDK_BUTTON4:
			/* Need to mark the GtkText button as 0 or else it will
			 * keep marking
			 */
			text->button = 0;
			gtk_grab_remove(widget);
			gtk_signal_emit_stop_by_name(
				GTK_OBJECT(widget), "button_release_event"
			);
			status = TRUE;
			break;
		  case GDK_BUTTON5:
			/* Need to mark the GtkText button as 0 or else it will
			 * keep marking
			 */
			text->button = 0;
			gtk_grab_remove(widget);
			gtk_signal_emit_stop_by_name(
				GTK_OBJECT(widget), "button_release_event"
			);
			status = TRUE;
			break;
		}
		break;
	}

	return(status);
}

/*
 *	OK callback.
 */
static void img_info_dlg_ok_cb(GtkWidget *widget, gpointer data)
{
	ImgInfoDlg *d = IMG_INFO_DLG(data);
	if(d == NULL)
		return;

	img_info_dlg_unmap(d);
}


/*
 *	Draws the display GtkDrawingArea.
 */
static void img_info_dlg_thumb_draw(ImgInfoDlg *d)
{
	gint width, height;
	GdkGC *gc;
	GtkWidget *w = d->thumb_da;
	GtkStateType state = GTK_WIDGET_STATE(w);
	GdkWindow *window = w->window;
	GtkStyle *style = gtk_widget_get_style(w);
	GdkDrawable *drawable;
	if((window == NULL) || (style == NULL))
		return;

	drawable = (GdkDrawable *)window;
	gdk_window_get_size((GdkWindow *)drawable, &width, &height);
	gc = d->gc;
	if(gc == NULL)
		d->gc = gc = gdk_gc_new((GdkWindow *)drawable);

	state = GTK_STATE_ACTIVE;
	gdk_gc_set_foreground(gc, &style->bg[state]);
	gdk_draw_rectangle(
		drawable,
		gc,
		TRUE,				/* Fill */
		0, 0,
		width, height
	);

	/* Draw the image */
	if(d->image_pixmap != NULL)
	{
		const gint	shadow_width = 5,
				shadow_height = 5;
		gint	x, y,
			shadow_x, shadow_y,
			image_width, image_height;
		GdkPixmap *pixmap = d->image_pixmap;
		GdkBitmap *mask = d->image_mask;

		gdk_window_get_size(
			(GdkWindow *)pixmap,
			&image_width, &image_height
		);

		/* Calculate the position of the image centered on the
		 * drawable with the shadow taken into account
		 */
		x = (width - (image_width + shadow_width)) / 2;
		y = (height - (image_height + shadow_height)) / 2;
		shadow_x = x + shadow_width;
		shadow_y = y + shadow_height;

		/* Draw the shadow */
		gdk_gc_set_clip_origin(
			gc,
			shadow_x, shadow_y
		);
		gdk_gc_set_clip_mask(gc, mask);
		gdk_gc_set_foreground(gc, &style->black);
		gdk_draw_rectangle(
			drawable,
			gc,
			TRUE,				/* Fill */
			shadow_x, shadow_y,
			image_width, image_height
		);

		/* Draw the image */
		gdk_gc_set_clip_origin(
			gc,
			x, y
		);
		gdk_draw_pixmap(
			drawable,
			gc,
			(GdkDrawable *)pixmap,
			0, 0,				/* Source position */
			x, y,				/* Target position */
			image_width, image_height
		);
		gdk_gc_set_clip_mask(gc, NULL);
	}

	/* Send drawable to window if drawable is not the window */
	if(drawable != (GdkDrawable *)window)
		gdk_draw_pixmap(
			(GdkDrawable *)window,
			gc,
			drawable,
			0, 0,
			0, 0,
			width, height
		);
}


/*
 *	Creates a new Image Info Dialog.
 */
ImgInfoDlg *img_info_dlg_new(GtkWidget *ref_toplevel)
{
	const gint	bw = GUI_BUTTON_HLABEL_WIDTH_DEF,
			bh = GUI_BUTTON_HLABEL_HEIGHT_DEF,
			border_major = 5,
			border_minor = 2;
	GdkWindow *window;
	GtkAccelGroup *accelgrp;
	GtkRcStyle *rcstyle;
	GtkStyle *style;
	GtkWidget	*w,
			*parent, *parent2, *parent3, *parent4,
			*parent5,
			*toplevel,
			*main_vbox,
			*sb;
	GtkEditable *editable;
	GtkText *text;
	ImgInfoDlg *d = IMG_INFO_DLG(g_malloc0(
		sizeof(ImgInfoDlg)
	));
	if(d == NULL)
		return(NULL);

	d->toplevel = toplevel = gtk_window_new(GTK_WINDOW_DIALOG);
	d->accelgrp = accelgrp = gtk_accel_group_new();

	/* Create the toplevel GtkWindow */
	w = toplevel;
#ifdef PROG_NAME
	gtk_window_set_wmclass(
		GTK_WINDOW(w), "dialog", PROG_NAME
	);
#endif
	gtk_widget_realize(w);
	window = w->window;
	if(window != NULL)
	{
		gdk_window_set_decorations(
			window,
			GDK_DECOR_BORDER | GDK_DECOR_TITLE
		);
		gdk_window_set_functions(
			window,
			GDK_FUNC_MOVE | GDK_FUNC_CLOSE
		);
		GUISetWMIcon(
			window,
			(guint8 **)icon_info_32x32_xpm
		);
	}
	gtk_signal_connect(
		GTK_OBJECT(w), "delete_event",
		GTK_SIGNAL_FUNC(img_info_dlg_delete_event_cb), d
	);
	gtk_window_add_accel_group(GTK_WINDOW(w), accelgrp);
	if(ref_toplevel != NULL)
		gtk_window_set_transient_for(
			GTK_WINDOW(w),
			GTK_WINDOW(ref_toplevel)
		);
	style = gtk_widget_get_style(w);
	parent = w;

	d->main_vbox = main_vbox = w = gtk_vbox_new(FALSE, 0);
	gtk_container_add(GTK_CONTAINER(parent), w);
	gtk_widget_show(w);
	parent = w;

	/* Image, File, and Comments GtkVBox */
	w = gtk_vbox_new(FALSE, border_major);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent2 = w;

	/* Image GtkFrame */
	d->image_parent = w = gtk_frame_new("Image");
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent3 = w;

	w = gtk_hbox_new(FALSE, 2 * border_major);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_container_add(GTK_CONTAINER(parent3), w);
	gtk_widget_show(w);
	parent3 = w;

	/* Parameters & Values GtkVBox */
	w = gtk_vbox_new(FALSE, border_major);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent4 = w;

	w = gtk_hbox_new(FALSE, border_major);
	gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent5 = w;

	/* Parameters GtkLabel */
 	d->image_parameters_label = w = gtk_label_new("\
Width:\n\
Height:\n\
Aspect:\n\
Bytes Per Pixel:\n\
Bytes Per Line:\n\
Frames:\n\
Frame Size:\n\
Total Frames Size:\n\
Loops:\n\
Loop Play Time:\n\
Total Play Time:\n\
Background Color:\n\
Alpha Channel:\n\
X:\n\
Y:\n\
Base Width:\n\
Base Height:\n\
Has Changes:"
	);
	gtk_label_set_justify(GTK_LABEL(w), GTK_JUSTIFY_RIGHT);
	gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	/* Values GtkLabel */
	d->image_values_label = w = gtk_label_new("");
	gtk_label_set_justify(GTK_LABEL(w), GTK_JUSTIFY_LEFT);
	gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
	gtk_widget_show(w);

	/* Thumb and Display GtkVBox */
	w = gtk_vbox_new(FALSE, border_major);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent4 = w;

	/* Thumb GtkFrame */
	w = gtk_frame_new(NULL);
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_IN);
	gtk_box_pack_start(GTK_BOX(parent4), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
	parent5 = w;

	/* Display GtkDrawingArea */
	d->thumb_da = w = gtk_drawing_area_new();
	gtk_signal_connect(
		GTK_OBJECT(w), "configure_event",
		GTK_SIGNAL_FUNC(img_info_dlg_thumb_event_cb), d
	);
	gtk_signal_connect(
		GTK_OBJECT(w), "expose_event",
		GTK_SIGNAL_FUNC(img_info_dlg_thumb_event_cb), d
	);
	gtk_container_add(GTK_CONTAINER(parent5), w);
	gtk_widget_show(w);

	/* GtkHBox for the Display */
	w = gtk_hbox_new(FALSE, border_major);
	gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent5 = w;
	/* Parameters GtkLabel */
	d->display_parameters_label = w = gtk_label_new("\
Display:\n\
Visual:\n\
Depth:"
	);
	gtk_label_set_justify(GTK_LABEL(w), GTK_JUSTIFY_RIGHT);
	gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	/* Values GtkLabel */
	d->display_value_label = w = gtk_label_new("");
	gtk_label_set_justify(GTK_LABEL(w), GTK_JUSTIFY_LEFT);
	gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
	gtk_widget_show(w);


	/* File GtkFrame */
	d->file_parent = w = gtk_frame_new("File");
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent3 = w;

	w = gtk_hbox_new(FALSE, border_major);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_container_add(GTK_CONTAINER(parent3), w);
	gtk_widget_show(w);
	parent3 = w;

	/* Parameters GtkLabel */
	d->file_parameters_label = w = gtk_label_new("\
File Name:\n\
Format:\n\
Size:\n\
Last Modified:\n\
Title:\n\
Author:\n\
Creator:"
	);
	gtk_label_set_justify(GTK_LABEL(w), GTK_JUSTIFY_RIGHT);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	/* Values GtkLabel */
	d->file_values_label = w = gtk_label_new("");
	gtk_label_set_justify(GTK_LABEL(w), GTK_JUSTIFY_LEFT);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_show(w);


	/* Comments GtkFrame */
	d->comments_frame = w = gtk_frame_new("Comments");
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent3 = w;

	w = gtk_table_new(1, 2, FALSE);
	gtk_table_set_row_spacing(GTK_TABLE(w), 0, border_minor);
	gtk_table_set_col_spacing(GTK_TABLE(w), 0, border_minor);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_container_add(GTK_CONTAINER(parent3), w);
	gtk_widget_show(w);
	parent3 = w;

	/* Text GtkRcStyle */
	rcstyle = gtk_rc_style_new();
	rcstyle->font_name = g_strdup(
#if defined(PROG_LANGUAGE_POLISH)
"-adobe-courier-medium-r-normal-*-12-*-*-*-*-*-iso8859-2"
#else
"-adobe-courier-medium-r-normal-*-12-*-*-*-*-*-iso8859-1"
#endif
	);
	/* GtkText */
	d->comments_text = w = gtk_text_new(NULL, NULL);
	editable = GTK_EDITABLE(w);
	text = GTK_TEXT(w);
	gtk_widget_set_usize(w, -1, 80);
	text->default_tab_width = 8;
	gtk_text_set_editable(text, FALSE);
	gtk_text_set_word_wrap(text, TRUE);
	gtk_table_attach(
		GTK_TABLE(parent3), w,
		0, 1, 0, 1,
		GTK_EXPAND | GTK_SHRINK | GTK_FILL,
		GTK_EXPAND | GTK_SHRINK | GTK_FILL,
		0, 0
	);
	gtk_signal_connect(
		GTK_OBJECT(w), "button_press_event",
		GTK_SIGNAL_FUNC(img_info_dlg_comments_text_event_cb), d
	);
	gtk_signal_connect(
		GTK_OBJECT(w), "button_release_event",
		GTK_SIGNAL_FUNC(img_info_dlg_comments_text_event_cb), d
	);
	gtk_widget_realize(w);
	gtk_widget_modify_style(w, rcstyle);
/*        gdk_window_set_cursor(w->window, d->text_cur); */
	GUIEditableEndowPopupMenu(w, GUI_EDITABLE_POPUP_MENU_UNDO);
	gtk_widget_show(w);
	GTK_RC_STYLE_UNREF(rcstyle);
	/* Vertical GtkScrollBar */
	sb = gtk_vscrollbar_new(GTK_TEXT(w)->vadj);
	gtk_table_attach(
		GTK_TABLE(parent3), sb,
		1, 2, 0, 1,
		GTK_FILL,
		GTK_EXPAND | GTK_SHRINK | GTK_FILL,
		0, 0
	);
	gtk_widget_show(sb);
#if 0
	w = gtk_hbox_new(FALSE, 0);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_container_add(GTK_CONTAINER(parent3), w);
	gtk_widget_show(w);
	parent3 = w;

	/* Comments GtkLabel */
	d->comments_label = w = gtk_label_new("");
	gtk_label_set_justify(GTK_LABEL(w), GTK_JUSTIFY_LEFT);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
#endif

	w = gtk_hseparator_new();
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	gtk_widget_show(w);


	/* Buttons GtkHBox */
	w = gtk_hbox_new(TRUE, 0);
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, border_major);
	gtk_widget_show(w);
	parent2 = w;

	/* OK GtkButton */
	d->ok_btn = w = GUIButtonPixmapLabelH(
		(guint8 **)icon_ok_20x20_xpm,
		"OK",
		NULL
	);
	gtk_widget_set_usize(w, bw, bh);
	GTK_WIDGET_SET_FLAGS(w, GTK_CAN_DEFAULT);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_signal_connect(
		GTK_OBJECT(w), "clicked",
		GTK_SIGNAL_FUNC(img_info_dlg_ok_cb), d
	);
	gtk_accel_group_add(
		accelgrp, GDK_Escape, 0, GTK_ACCEL_VISIBLE,
		GTK_OBJECT(w), "clicked"
	);
	gtk_accel_group_add(
		accelgrp, GDK_o, 0, GTK_ACCEL_VISIBLE,
		GTK_OBJECT(w), "clicked"
	);
	GUIButtonLabelUnderline(w, GDK_o);
	gtk_widget_show(w);



	return(d);
}

/*
 *	Sets the image pixmap and mask.
 */
static void img_info_dlg_set_image(
	ImgInfoDlg *d,
	const guint8 *rgba,
	const gint width, const gint height,
	const gint bpp, const gint bpl,
	const guint8 *bg_color			/* 4 bytes RGBA */
)
{
	guint8 white_bg_color[] = { 0xff, 0xff, 0xff, 0xff };
	const gint border_major = 5;
	GtkWidget *w = d->thumb_da;

	if((rgba == NULL) || (width <= 0) || (height <= 0) ||
	   (bpp != 4) || (bpl <= 0)
	)
	{
		d->image_pixmap = GDK_PIXMAP_UNREF(d->image_pixmap);
		d->image_mask = GDK_BITMAP_UNREF(d->image_mask);
		gtk_widget_queue_draw(w);
		return;
	}

	if(bg_color == NULL)
		bg_color = white_bg_color;

	if(TRUE)
	{
		/* Render a smaller sized image with a background using
		 * the specified background color for alpha values and
		 * a shadow
		 *
		 * The thumb_width and thumb_height specifies the entire
		 * size of the thumb, while the thumb_image_width and
		 * thumb_image_height specifies the entire size minus
		 * the shadows
		 *
		 * Note that thumb_*_max is smaller than
		 * IMG_INFO_DLG_DISPLAY_MAX_* to accomidate for the trough
		 */
		const gint	thumb_bpp = 4,
				thumb_width_max = MAX(
					IMG_INFO_DLG_DISPLAY_MAX_WIDTH - (4 * border_major),
					1
				),
				thumb_height_max = MAX(
					IMG_INFO_DLG_DISPLAY_MAX_HEIGHT - (4 * border_major),
					1
				);
		gfloat aspect = (gfloat)width / (gfloat)height;
		gint	thumb_width, thumb_height,
			thumb_bpl;
		guint8	*thumb_rgba;
		GdkBitmap	*mask;
		GdkPixmap	*pixmap;
		GdkWindow	*window = w->window;

		if(aspect <= 0.0f)
			aspect = 1.0f;

		/* Is the size of the image going to be bigger than the
		 * maximum size?
		 */
		if((width > thumb_width_max) || (height > thumb_height_max))
		{
			/* Calculate the entire size of the thumb based on the
			 * specified maximum size limit of thumb_width_max and
			 * thumb_height_max while maintaining the aspect ratio
			 */
			thumb_width = width;
			thumb_height = height;
			if(thumb_width > thumb_width_max)
			{
				thumb_width = thumb_width_max;
				thumb_height = (gint)(thumb_width / aspect);
			}
			if(thumb_height > thumb_height_max)
			{
				thumb_height = thumb_height_max;
				thumb_width = (gint)(thumb_height * aspect);
			}
		}
		else
		{
			thumb_width = width;
			thumb_height = height;
		}
		gtk_widget_set_usize(
			d->thumb_da,
			thumb_width + (4 * border_major),
			thumb_height + (4 * border_major)
		);

		/* Create the thumb RGBA image data */
		thumb_bpl = thumb_width * thumb_bpp;
		thumb_rgba = (guint8 *)g_malloc(
			thumb_bpl * thumb_height
		);
		if(thumb_rgba != NULL)
		{
			/* Copy/resize the specified RGBA image data to the
			 * thumb RGBA image data
			 */
			GUIImageBufferResize(
				thumb_bpp,
				rgba,
				width, height, bpl,
				thumb_rgba,
				thumb_width, thumb_height, thumb_bpl,
				NULL, NULL
			);
			GUIImageBufferFlattenWithBG(
				thumb_bpp,
				thumb_rgba,
				thumb_width, thumb_height, thumb_bpl,
				bg_color,
				NULL, NULL
			);
		}

		/* Create the thumb GdkPixmap */
		pixmap = gdk_pixmap_new(
			window,
			thumb_width, thumb_height,
			-1
		);
		if((pixmap != NULL) && (thumb_rgba != NULL))
		{
			GdkGC *gc = gdk_gc_new((GdkWindow *)pixmap);
			if(gc != NULL)
			{
				/* Draw the thumb image */
				gdk_draw_rgb_32_image(
					pixmap,
					gc,
					0, 0,
					thumb_width, thumb_height,
					GDK_RGB_DITHER_NORMAL,
					(guchar *)thumb_rgba,
					thumb_bpl
				);
				GDK_GC_UNREF(gc);
			}
		}

#if 1
		/* Use no mask (solid) since the image was flattened with
		 * the background
		 */
		mask = NULL;
#else
		/* Create the thumb GdkBitmap mask */
		mask = GUICreateBitmapFromDataRGBA(
			thumb_width, thumb_height, -1,
			thumb_rgba,
			0x80,				/* Alpha byte threshold */
			window
		);
#endif

		/* Delete the thumb image data */
		g_free(thumb_rgba);

		/* Set the new thumb pixmap and mask */
		GDK_PIXMAP_UNREF(d->image_pixmap);
		GDK_BITMAP_UNREF(d->image_mask);
		d->image_pixmap = GDK_PIXMAP_REF(pixmap);
		d->image_mask = GDK_BITMAP_REF(mask);
		GDK_PIXMAP_UNREF(pixmap);
		GDK_BITMAP_UNREF(mask);
	}
	else
	{
		/* Render a thumb with no background and display alpha
		 * values appropriatly as transparent pixels
		 */
		const gint	thumb_bpp = 4,
				thumb_width_max = IMG_INFO_DLG_DISPLAY_MAX_WIDTH,
				thumb_height_max = IMG_INFO_DLG_DISPLAY_MAX_HEIGHT;
		gfloat aspect = (gfloat)width / (gfloat)height;
		gint	thumb_width = width,
			thumb_height = height,
			thumb_bpl;
		guint8 *thumb_rgba;
		GdkBitmap *mask;
		GdkPixmap *pixmap;
		GdkWindow *window = w->window;
		GtkStyle *style = gtk_widget_get_style(w);

		if(aspect <= 0.0f)
			aspect = 1.0f;

		if(thumb_width > thumb_width_max)
		{
			thumb_width = thumb_width_max;
			thumb_height = (gint)(thumb_width / aspect);
		}
		if(thumb_height > thumb_height_max)
		{
			thumb_height = thumb_height_max;
			thumb_width = (gint)(thumb_height * aspect);
		}
		gtk_widget_set_usize(
			d->thumb_da,
			thumb_width,
			thumb_height
		);

		thumb_bpl = thumb_width * thumb_bpp;

		/* Allocate the thumb image data */
		thumb_rgba = (guint8 *)g_malloc(thumb_bpl * thumb_height);

		/* Copy/resize the specified RGBA image data to the thumb
		 * image data
		 */
		GUIImageBufferResize(
			thumb_bpp,
			rgba,
			width, height, bpl,
			thumb_rgba,
			thumb_width, thumb_height, thumb_bpl,
			NULL, NULL
		);

		/* Create the pixmap and mask from the thumb image data*/
		pixmap = gdk_pixmap_new(
			window,
			thumb_width, thumb_height,
			-1
		);
		if((pixmap != NULL) && (thumb_rgba != NULL))
			gdk_draw_rgb_32_image(
				pixmap, style->black_gc,
				0, 0, thumb_width, thumb_height,
				GDK_RGB_DITHER_NORMAL,
				(guchar *)thumb_rgba,
				thumb_bpl
			);
		mask = GUICreateBitmapFromDataRGBA(
			thumb_width, thumb_height, -1,
			thumb_rgba,
			0x80,			/* Alpha byte threshold */
			window
		);

		/* Delete the thumb image data */
		g_free(thumb_rgba);

		/* Set the new thumb pixmap and mask */
		(void)GDK_PIXMAP_UNREF(d->image_pixmap);
		(void)GDK_BITMAP_UNREF(d->image_mask);
		d->image_pixmap = GDK_PIXMAP_REF(pixmap);
		d->image_mask = GDK_BITMAP_REF(mask);
		(void)GDK_PIXMAP_UNREF(pixmap);
		(void)GDK_BITMAP_UNREF(mask);
	}

	gtk_widget_queue_draw(w);
}

/*
 *	Sets the values.
 */
void img_info_dlg_set_values(
	ImgInfoDlg *d,
#ifdef HAVE_LIBENDEAVOUR2
	EDVContext *edv2_ctx,
#endif
	const guint8 *rgba,
	const gint width, const gint height,
	const gint bpp, const gint bpl,
	const gint nframes,
	const gint nloops,			/* -1 for infinite */
	const gulong play_time_ms,		/* Play time per loop in
						 * milliseconds */
	const gulong total_play_time_ms,	/* Total play time in
						 * milliseconds or (gulong)-1
						 * for infinite */
	const gint alpha_channel,               /* 0 = None
						 * 1 = Allocated But Uniform
						 * 2 = Allocated & Defined */
	const guint8 *bg_color,                 /* 4 bytes RGBA */
	const gint x, const gint y,             /* Offset on base */
	const gint base_width,                  /* Dimensions of base (can be 0 for no base) */
	const gint base_height,
	const gboolean has_changes,

	const gchar *file_name,
	const gchar *file_format,		/* Extension */
	const gulong file_size,                 /* File size in bytes */
	const gulong last_modified,             /* Last modified time in
						 * seconds since EPOCH */
	const gchar *title,
	const gchar *author,
	const gchar *creator,
	const gchar *comments,
	const gchar *display,                   /* X Display address */
	const gchar *visual_type,               /* X Visual type */
	const gint display_depth                /* X Depth in bits per pixel */
)
{
	const gchar	*no_val_s = "-",
			*alpha_channel_s,
			*last_modified_s;
	gchar		*s,
			*format_desc = NULL;
	const gulong	image_frame_size = (gulong)height * (gulong)bpl,
			total_image_frames_size = image_frame_size * nframes;
	GtkWidget	*w,
			*toplevel;

	if(d == NULL)
		return;

	toplevel = d->toplevel;

	/* Get the Alpha Channel String */
	switch(alpha_channel)
	{
	  case 0:
		alpha_channel_s = "No";
		break;
	  case 1:
		alpha_channel_s = "Yes (Uniform)";
		break;
	  default:
		alpha_channel_s = "Yes";
		break;
	}

	/* Get the Last Modified String */
	if(last_modified > 0l)
#ifdef HAVE_LIBENDEAVOUR2
		last_modified_s = img_info_dlg_get_date_string(edv2_ctx, last_modified);
#else
		last_modified_s = img_info_dlg_get_date_string(last_modified);
#endif
	else
		last_modified_s = no_val_s;

	/* Get format string */
	if(!STRISEMPTY(file_format))
	{
		const gchar *imgio_file_types_list[] = IMAGE_FORMAT_FILE_TYPES_LIST;
		gint i;
		for(i = 0; imgio_file_types_list[i] != NULL; i += 2)
		{
			if(strcasestr(imgio_file_types_list[i], file_format) != NULL)
			{
				g_free(format_desc);
				format_desc = STRDUP(imgio_file_types_list[i + 1]);
				break;
			}
		}
	}

	/* Title */
	w = d->toplevel;
	s = g_strdup_printf(
		IMG_INFO_DLG_DEF_TITLE ": %s",
		STRISEMPTY(file_name) ? "Untitled" : file_name
	);
	gtk_window_set_title(GTK_WINDOW(w), s);
	g_free(s);

	/* Image */
	w = d->image_values_label;
	if(w != NULL)
	{
		gchar	*width_s = STRDUP(
			img_info_get_deliminated_size_string(width)
		),
			*height_s = STRDUP(
			img_info_get_deliminated_size_string(height)
		),
			*aspect_s = (height > 0) ? g_strdup_printf(
					"%.2f",
					(gfloat)width / (gfloat)height
				) : g_strdup("-"),
			*bpp_s = STRDUP(
			img_info_get_deliminated_size_string(bpp)
		),
			*bpl_s = STRDUP(
			img_info_get_deliminated_size_string(bpl)
		),
			*nframes_s = STRDUP(
			img_info_get_deliminated_size_string(nframes)
		),
			*image_frame_size_s = STRDUP(
			img_info_get_deliminated_size_string(image_frame_size)
		),
			*total_image_frames_size_s = STRDUP(
			img_info_get_deliminated_size_string(total_image_frames_size)
		),
			*nloops_s = (nloops < 0) ?
			g_strdup("Infinite") : (nloops == 0) ? g_strdup("-") : STRDUP(
			img_info_get_deliminated_size_string(nloops)
		),
			*loop_play_time_s = (play_time_ms == (gulong)-1) ?
			g_strdup("Infinite") : (play_time_ms == 0l) ? g_strdup("-") : g_strdup_printf(
			"%.2f ms",
			(gfloat)play_time_ms / 1000.0f
		),
			*total_play_time_s = (total_play_time_ms == (gulong)-1) ?
			g_strdup("Infinite") : (total_play_time_ms == 0l) ? g_strdup("-") : g_strdup_printf(
			"%.2f ms",
			(gfloat)total_play_time_ms / 1000.0f
		),
			*bg_color_s = (bg_color != NULL) ? g_strdup_printf(
			"%.2X%.2X%.2X",
			bg_color[0],
			bg_color[1],
			bg_color[2]
		) : STRDUP(no_val_s),
			*base_width_s = (base_width > 0) ?
			STRDUP(img_info_get_deliminated_size_string(base_width)) :
			STRDUP(no_val_s),
			*base_height_s = (base_height > 0) ?
			STRDUP(img_info_get_deliminated_size_string(base_height)) :
			STRDUP(no_val_s);
		s = g_strdup_printf("\
%s\n\
%s\n\
%s\n\
%s\n\
%s\n\
%s\n\
%s %s\n\
%s %s\n\
%s\n\
%s\n\
%s\n\
%s\n\
%s\n\
%i\n\
%i\n\
%s\n\
%s\n\
%s",
			width_s,
			height_s,
			aspect_s,
			bpp_s,
			bpl_s,
			nframes_s,
			image_frame_size_s,
			(image_frame_size == 1l) ? "byte" : "bytes",
			total_image_frames_size_s,
			(total_image_frames_size == 1l) ? "byte" : "bytes",
			nloops_s,
			loop_play_time_s,
			total_play_time_s,
			bg_color_s,
			alpha_channel_s,
			x, y,
			base_width_s, base_height_s,
			(has_changes) ? "Yes" : "No"
		);
		gtk_label_set_text(GTK_LABEL(w), s);
		g_free(s);

		g_free(width_s);
		g_free(height_s);
		g_free(aspect_s);
		g_free(bpp_s);
		g_free(bpl_s);
		g_free(nframes_s);
		g_free(image_frame_size_s);
		g_free(total_image_frames_size_s);
		g_free(nloops_s);
		g_free(loop_play_time_s);
		g_free(total_play_time_s);
		g_free(bg_color_s);
		g_free(base_width_s);
		g_free(base_height_s);
	}

	/* File */
	w = d->file_parent;
	if(w != NULL)
	{
		if(!STRISEMPTY(file_name))
			gtk_widget_show(w);
		else
			gtk_widget_hide(w);
	}

	w = d->file_values_label;
	if((w != NULL) && !STRISEMPTY(file_name))
	{
		gchar	*file_size_s = STRDUP(
			img_info_get_deliminated_size_string(file_size)
		);

		s = g_strdup_printf("\
%s\n\
%s\n\
%s %s\n\
%s\n\
%s\n\
%s\n\
%s",
			file_name,
			STRISEMPTY(format_desc) ?
				(STRISEMPTY(file_format) ? no_val_s : file_format) :
				format_desc,
			file_size_s, (file_size == 1l) ? "byte" : "bytes",
			last_modified_s,
			STRISEMPTY(title) ? no_val_s : title,
			STRISEMPTY(author) ? no_val_s : author,
			STRISEMPTY(creator) ? no_val_s : creator
		);
		gtk_label_set_text(GTK_LABEL(w), s);
		g_free(s);

		g_free(file_size_s);
	}

	/* Display */
	w = d->display_value_label;
	if(w != NULL)
	{
		gchar       *display_depth_s = STRDUP(
			img_info_get_deliminated_size_string(display_depth)
		);

		s = g_strdup_printf("\
%s\n\
%s\n\
%s %s per pixel",
			STRISEMPTY(display) ? no_val_s : display,
			STRISEMPTY(visual_type) ? no_val_s : visual_type,
			display_depth_s, (display_depth == 1) ? "bit" : "bits"
		);
		gtk_label_set_text(GTK_LABEL(w), s);
		g_free(s);

		g_free(display_depth_s);
	}


	/* Comments */
	w = d->comments_frame;
	if(w != NULL)
	{
		if(!STRISEMPTY(comments))
			gtk_widget_show(w);
		else
			gtk_widget_hide(w);
	}

	w = d->comments_text;
	if((w != NULL) && !STRISEMPTY(comments))
	{
		GtkText *text = GTK_TEXT(w);
		gtk_text_freeze(text);
		gtk_text_insert(
			text,
			NULL,
			NULL, NULL,
			comments, -1
		);
		gtk_text_thaw(text);
	}


	g_free(format_desc);


	/* Thumb */
	img_info_dlg_set_image(
		d,
		rgba,
		width, height,
		bpp, bpl,
		bg_color
	);
}

/*
 *	Checks if the Image Info Dialog is mapped.
 */
gboolean img_info_dlg_is_mapped(ImgInfoDlg *d)
{
	if(d == NULL)
		return(FALSE);

	return(GTK_WIDGET_MAPPED(d->toplevel));
}

/*
 *	Maps the Image Info Dialog.
 */
void img_info_dlg_map(ImgInfoDlg *d)
{
	GtkWidget *w;

	if(d == NULL)
		return;

	w = d->toplevel;
	if(w != NULL)
		gtk_widget_show_raise(w);

	w = d->ok_btn;
	if(w != NULL)
	{
		gtk_widget_grab_default(w);
		gtk_widget_grab_focus(w);
	}
}

/*
 *	Unmaps the Image Info Dialog.
 */
void img_info_dlg_unmap(ImgInfoDlg *d)
{
	GtkWidget *w;

	if(d == NULL)
		return;

	w = d->toplevel;
	if(w == NULL)
		return;

	gtk_widget_hide(w);
}

/*
 *	Deletes the Image Info Dialog.
 */
void img_info_dlg_delete(ImgInfoDlg *d)
{
	if(d == NULL)
		return;

	GTK_WIDGET_DESTROY(d->toplevel);
	GTK_ACCEL_GROUP_UNREF(d->accelgrp);
	(void)GDK_GC_UNREF(d->gc);
	(void)GDK_PIXMAP_UNREF(d->image_pixmap);
	(void)GDK_BITMAP_UNREF(d->image_mask);

	g_free(d);
}
