#if !defined(_WIN32) && defined(HAVE_XPRINT)
#include <stdio.h>
#include <stdlib.h>

#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#ifdef HAVE_LIBENDEAVOUR2
# include <endeavour2.h>
#endif

#include "../include/string.h"

#include "guiutils.h"
#include "guirgbimg.h"
#include "statictip.h"
#include "pulist.h"
#include "imgview.h"
#include "cdialog.h"
#include "print_values.h"
#include "print.h"
#include "print_dlg.h"
#include "config.h"

#include "images/icon_print2_48x48.xpm"
#include "images/icon_print_preview_48x48.xpm"
#include "images/icon_print2_20x20.xpm"
#include "images/icon_print_preview_20x20.xpm"
#include "images/icon_cancel_20x20.xpm"
#include "images/icon_close_20x20.xpm"

#include "images/icon_minus_16x16.xpm"
#include "images/icon_plus_16x16.xpm"
#include "images/icon_brightness_16x16.xpm"
#include "images/icon_contrast_16x16.xpm"

#include "images/icon_file_20x20.xpm"
#include "images/icon_printer_tray_closed_20x20.xpm"
#include "images/icon_printer_tray_opened_20x20.xpm"


typedef struct _Preview			Preview;
#define PREVIEW(p)			((Preview *)(p))

typedef struct _PrintDlg		PrintDlg;
#define PRINT_DLG(p)			((PrintDlg *)(p))

typedef struct _PrintPreviewDlg		PrintPreviewDlg;
#define PRINT_PREVIEW_DLG(p)		((PrintPreviewDlg *)(p))


/*
 *	Preview Flags:
 */
typedef enum {
	/* Set in PreviewEventCB() to tell PreviewDraw() to draw the
	 * drag outline, PreviewDraw() clears if afterwards */
	PREVIEW_INTERACTIVE		= (1 << 0),
	PREVIEW_DRAW_DRAG_OUTLINE	= (1 << 1),
	PREVIEW_DREW_PREVIOUS_OUTLINE	= (1 << 2)
} PreviewFlags;


/*
 *	Preview Drag Flags:
 */
typedef enum {
	PREVIEW_DRAG_NONE		= 0,
	PREVIEW_DRAG_TRANSLATE		= (1 << 0),
	PREVIEW_DRAG_RESIZE		= (1 << 1)
} PreviewDragFlags;


/*
 *	Print Dialog Flags:
 */
typedef enum {
	PRINT_DLG_MAPPED		= (1 << 0),
	PRINT_DLG_REALIZED		= (1 << 1),
	PRINT_DLG_RESPONSE_OK		= (1 << 7),
} PrintDlgFlags;


/*
 *	Print Preview Dialog Flags:
 */
typedef enum {
	PRINT_PREVIEW_DLG_MAPPED	= (1 << 0),
	PRINT_PREVIEW_DLG_REALIZED	= (1 << 1)
} PrintPreviewDlgFlags;


/* Get Values */
static print_visual PrintDlgGetVisual(PrintDlg *d);
static gboolean PrintDlgGetMediumSize(
	PrintDlg *d,
	gint *width_rtn, gint *height_rtn
);
static gfloat PrintDlgGetSourceImageAspect(PrintDlg *d);
static void PrintDlgOutputSetWidgets(PrintDlg *d);
static gboolean PrintDlgTranslateOutputCoordinates(
	PrintDlg *d,
	GtkWidget *w,
	gint *x_rtn, gint *y_rtn
);

/* Revert Image */
static void PrintDlgRevertImages(
	imgview_image_struct *src_img,
	imgview_image_struct *tar_img
);

/* Connect To The X Print Server */
static void PrintDlgConnectToXPrintServer(
	PrintDlg *d,
	const gchar *xprint_server_address,
	const gboolean verbose
);

/* Update Values From The X Print Server */
static void PrintDlgUpdateXPrintServerPrinters(
	PrintDlg *d,
	const gchar *xprint_server_address,
	const gboolean verbose
);
static void PrintDlgUpdatePrinterValues(
	PrintDlg *d,
	const gchar *printer_name,
	const gboolean verbose
);
static void PrintDlgUpdateMediumsList(
	PrintDlg *d,
	const gchar *printer_name,
	const gboolean verbose
);


/* Preview Callbacks */
static void PreviewToplevelSizeAllocateCB(
	GtkWidget *widget, GtkAllocation  *allocation, gpointer data
);
static gint PreviewEventCB(
	GtkWidget *widget, GdkEvent *event, gpointer data
);

/* Preview */
static Preview *PreviewNew(
	GtkWidget *parent,
	GtkStyle *style,
	GdkColor *bg_color,
	const gboolean interactive,
	PrintDlg *print_dlg
);
static void PreviewUpdateVisual(Preview *p);
static void PreviewUpdateImages(Preview *p);
static void PreviewResize(Preview *p);
static void PreviewDraw(Preview *p);
static void PreviewQueueDraw(Preview *p);
static void PreviewDelete(Preview *p);


/* PrintDlg Callbacks */
static gint PrintDlgDeleteEventCB(
	GtkWidget *widget, GdkEvent *event, gpointer data
);
static void PrintDlgVisualAdjustmentValueChangedCB(
	GtkAdjustment *adj, gpointer data
);
static void PrintDlgVisualToggledCB(GtkWidget *widget, gpointer data);
static gint PrintDlgVisualEventCB(
	GtkWidget *widget, GdkEvent *event, gpointer data
);
static void PrintDlgXPrintServerConnectCB(
	GtkWidget *widget, gpointer data
);
static void PrintDlgPrinterChangedCB(
	GtkWidget *widget, const gint i, gpointer data
);
#if 0
static void PrintDlgTrayChangedCB(
	GtkWidget *widget, const gint i, gpointer data
);
#endif
static void PrintDlgMediumSelectRowCB(
	GtkCTree *ctree, GtkCTreeNode *node, gint column,
	gpointer data
);
static void PrintDlgOrientationChangedCB(
	GtkWidget *widget, const gint i, gpointer data
);
static void PrintDlgResolutionChangedCB(
	GtkWidget *widget, const gint i, gpointer data
);
static void PrintDlgOutputChangedCB(GtkWidget *widget, gpointer data);

static void PrintDlgPrintCB(GtkWidget *widget, gpointer data);
static void PrintDlgPreviewCB(GtkWidget *widget, gpointer data);
static void PrintDlgCancelCB(GtkWidget *widget, gpointer data);

/* PrintDlg */
static PrintDlg *PrintDlgNew(
	const gchar *path,
	GtkWidget *ref_toplevel
);
static void PrintDlgSetBusy(
	PrintDlg *d,
	const gboolean busy
);
static void PrintDlgUpdate(PrintDlg *d);
static void PrintDlgMap(PrintDlg *d);
static void PrintDlgUnmap(PrintDlg *d);
static void PrintDlgDelete(PrintDlg *d);


/* PrintPreviewDlg Callbacks */
static gint PrintPreviewDlgDeleteEventCB(
	GtkWidget *widget, GdkEvent *event, gpointer data
);
static void PrintPreviewDlgPrintCB(GtkWidget *widget, gpointer data);
static void PrintPreviewDlgCloseCB(GtkWidget *widget, gpointer data);

/* PrintPreviewDlg */
static PrintPreviewDlg *PrintPreviewDlgNew(PrintDlg *print_dlg);
static void PrintPreviewDlgUpdate(PrintPreviewDlg *d);
static void PrintPreviewDlgMap(PrintPreviewDlg *d);
static void PrintPreviewDlgUnmap(PrintPreviewDlg *d);
static void PrintPreviewDlgDelete(PrintPreviewDlg *d);


/* Front ends */
#ifdef HAVE_LIBENDEAVOUR2
gboolean PrintDlgGetResponse(
	imgview_image_struct *img,		/* Image to print */
	const gchar *path,
	print_values_struct *v,			/* Print values (will be modified) */
	const guint8 *bg_color,			/* 4 bytes RGBA */
	GtkWidget *toplevel,
	EDVContext *edv2_ctx
);
#else
gboolean PrintDlgGetResponse(
	imgview_image_struct *img,		/* Image to print */
	const gchar *path,
	print_values_struct *v,			/* Print values (will be modified) */
	const guint8 *bg_color,			/* 4 bytes RGBA */
	GtkWidget *toplevel
);
#endif


/*
 *	Preview:
 */
struct _Preview {

	GtkWidget	*toplevel;		/* Parent GtkEventBox */
	gint		freeze_count;
	GdkGC		*gc,
			*invert_gc;
	PreviewFlags	flags;

	PrintDlg	*print_dlg;

	/* Cursors */
	GdkCursor	*translate_cur,
			*resize_cur;

	GtkWidget	*da,			/* GtkDrawingArea */
			*shadow_x_da,
			*shadow_y_da;

	GdkPixmap	*pixmap;

	gint		last_set_width,		/* Last set size of the
						 * GtkDrawingArea, used in
						 * PreviewResize() */
			last_set_height;

	/* Resized ImgView image without color, brightness, or contrast
	 * applied that is suitable to be rendered and then drawn on to
	 * the GtkDrawingArea */
	imgview_image_struct	*resized_img;

	/* Rendered ImgView image (rendered from the resized_img) ready
	 * to be drawn on to the GtkDrawingArea */
	imgview_image_struct	*rendered_img;

	/* Drag */
	PreviewDragFlags	drag_modifiers,
				drag_operation;
	gint		drag_last_x,
			drag_last_y;

	/* Rectangular drag outline in window coordinates, for drawing
	 * a drag outline on the preview GtkDrawingArea
	 */
	gint		drag_outline_x[2],
			drag_outline_y[2];
	gint		prev_outline_x[2],
			prev_outline_y[2];
};

/*
 *	Print Dialog:
 */
struct _PrintDlg {

	GtkWidget	*toplevel;
	GtkAccelGroup	*accelgrp;
	GtkWidget	*ref_toplevel;
	gint		freeze_count,
			busy_count,
			main_level;
	PrintDlgFlags	flags;

	GdkCursor	*busy_cur;

	/* Path to the image file being printed or NULL for untitled */
	gchar		*path;

	/* Source Image */
	imgview_image_struct	*src_img;
	gint			src_frame_num;

	/* Address */
	GtkWidget	*xprint_server_address_entry;

	/* Printer */
	GtkWidget	*printer_pulistbox;

	/* Visual */
	GtkWidget	*visual_bw_radio,
			*visual_greyscale_radio,
			*visual_color_radio;
	GtkWidget	*brightness_scale;
	GtkAdjustment	*brightness_adj;
	GtkWidget	*contrast_scale;
	GtkAdjustment	*contrast_adj;

	/* Medium */
	GtkWidget	*medium_ctree;
	GdkPixmap	*tray_closed_icon_pm;
	GdkBitmap	*tray_closed_icon_mask;
	GdkPixmap	*tray_opened_icon_pm;
	GdkBitmap	*tray_opened_icon_mask;
	GdkPixmap	*medium_icon_pm;
	GdkBitmap	*medium_icon_mask;

	/* Orientation */
	GtkWidget	*orientation_pulistbox;

	/* Resolution */
	GtkWidget	*resolution_pulistbox;

	/* Output */
	GtkWidget	*output_x_spin,
			*output_y_spin,
			*output_width_spin,
			*output_height_spin,
			*output_frame_spin,
			*output_maintain_aspect_check,
			*output_ncoppies_spin;

	/* Preview */
	Preview		*preview;

	/* GtkButtons */
	GtkWidget	*print_btn,
			*preview_btn,
			*cancel_btn,

			*menu;			/* Preview Right Click GtkMenu */

	/* Medium Size */
	GdkRectangle	medium_size;		/* x and y ignored */

	/* Output Image Geometry */
	GdkRectangle	output_img_geometry;
	gboolean	output_maintain_aspect;

	guint8		bg_color[4];		/* 4 bytes RGBA */

	/* Print Preview Dialog */
	PrintPreviewDlg	*print_preview_dlg;
};

/*
 *	Print Preview Dialog:
 */
struct _PrintPreviewDlg {

	GtkWidget	*toplevel;
	GtkAccelGroup	*accelgrp;
	gint		freeze_count;
	PrintPreviewDlgFlags	flags;

	PrintDlg	*print_dlg;

	/* Preview */
	Preview		*preview;

	GtkWidget	*print_btn,
			*close_btn;
};


#define PRINT_DLG_TITLE			"Print"

#define PRINT_DLG_WIDTH			640
#define PRINT_DLG_HEIGHT		480

#define PRINT_DLG_DEF_MEDIUM_WIDTH	240
#define PRINT_DLG_DEF_MEDIUM_HEIGHT	320

/* Nominal size of the PrintDlg's preview GtkDrawingArea, this size must
 * be square in shape
 */
#define PRINT_DLG_PREVIEW_NOMINAL_WIDTH		200
#define PRINT_DLG_PREVIEW_NOMINAL_HEIGHT	200


#define PRINT_PREVIEW_DLG_TITLE		"Print Preview"

#define PRINT_PREVIEW_DLG_WIDTH		480
#define PRINT_PREVIEW_DLG_HEIGHT	640


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0) 
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f) 
#define STRDUP(s)	(((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define ABSOLUTE(x)	(((x) < 0) ? ((x) * -1) : (x))

#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : TRUE)


/*
 *	Gets the visual.
 */
static print_visual PrintDlgGetVisual(PrintDlg *d)
{
	if(GTK_TOGGLE_BUTTON_GET_ACTIVE(
	    d->visual_bw_radio
	))
	    return(PRINT_VISUAL_BW);
	else if(GTK_TOGGLE_BUTTON_GET_ACTIVE(
	    d->visual_greyscale_radio
	))
	    return(PRINT_VISUAL_GREYSCALE);
	else if(GTK_TOGGLE_BUTTON_GET_ACTIVE(
	    d->visual_color_radio
	))
	    return(PRINT_VISUAL_RGB);
	else
	    return(PRINT_VISUAL_BW);
}

/*
 *	Gets the paper size based on the current selected printer,
 *	orientation, and resolution.
 *
 *	The width_rtn and height_rtn specifies the return values for
 *	the paper size in pixels.
 *
 *	Returns TRUE if the paper size was obtained or FALSE on error.
 */
static gboolean PrintDlgGetMediumSize(
	PrintDlg *d,
	gint *width_rtn, gint *height_rtn
)
{
	gint		x_dpi, y_dpi;
	const gchar	*xprint_server_address = gtk_entry_get_text(
	    GTK_ENTRY(d->xprint_server_address_entry)
			),
			*printer_name,
			*tray_name,
			*medium_name,
			*orientation_name;
	GList *glist;
	GtkCList *clist;
	GtkCTreeNode *node;
	GtkCTree *ctree;
	IVPrintXPrintPrinter *p;
	IVPrintXPrintMedium *m;
	IVPrintXPrintResolution *r;

	/* Get the selected printer */
	p = IV_PRINT_XPRINT_PRINTER(PUListBoxGetSelectedData(
	    d->printer_pulistbox
	));
	if(p != NULL)
	    printer_name = p->name;
	else
	    printer_name = NULL;

	/* Get the selected medium */
	clist = GTK_CLIST(d->medium_ctree);
	ctree = GTK_CTREE(clist);
	glist = clist->selection_end;
	node = (glist != NULL) ?
	    (GtkCTreeNode *)glist->data : NULL;
	m = IV_PRINT_XPRINT_MEDIUM(gtk_ctree_node_get_row_data(
	    ctree,
	    node
	));
	if(m != NULL)
	{
	    tray_name = m->tray;
	    medium_name = m->name;
	}
	else
	{
	    tray_name = NULL;
	    medium_name = NULL;
	}

	/* Get the selected orientation */
	orientation_name = (const gchar *)PUListBoxGetSelectedData(
	    d->orientation_pulistbox
	);

	/* Get the selected resolution */
	r = IV_PRINT_XPRINT_RESOLUTION(PUListBoxGetSelectedData(
	    d->resolution_pulistbox
	));
	if(r != NULL)
	{
	    x_dpi = r->x_dpi;
	    y_dpi = r->y_dpi;
	}
	else
	{
	    x_dpi = 0;
	    y_dpi = 0;
	}

	/* Get the medium size from the X Print server based on the
	 * currently selected printer, tray, medium, orientation, and
	 * resolution
	 */
	if(iv_print_xprint_get_medium_size(
	    xprint_server_address,
	    printer_name,
	    tray_name,
	    medium_name,
	    orientation_name,
	    x_dpi, y_dpi,
	    width_rtn, height_rtn
	))
	    return(FALSE);
	else
	    return(TRUE);
}

/*
 *	Returns the Print Window's source image aspect.
 */
static gfloat PrintDlgGetSourceImageAspect(PrintDlg *d)
{
	const imgview_image_struct *img = d->src_img;
	if(img == NULL)
	    return(0.0f);

	if(img->height <= 0)
	    return(0.0f);
	else
	    return((gfloat)img->width / (gfloat)img->height);
}

/*
 *	Sets the output GtkWidget values to match the current output
 *	values.
 *
 *	The preview will not be updated or drawn.
 */
static void PrintDlgOutputSetWidgets(PrintDlg *d)
{
	GdkRectangle *output = &d->output_img_geometry;

	d->freeze_count++;

	gtk_spin_button_set_value(
	    GTK_SPIN_BUTTON(d->output_x_spin),
	    (gfloat)output->x
	);
	gtk_spin_button_set_value(
	    GTK_SPIN_BUTTON(d->output_y_spin),
	    (gfloat)output->y
	);
	gtk_spin_button_set_value(
	    GTK_SPIN_BUTTON(d->output_width_spin),
	    (gfloat)output->width
	);
	gtk_spin_button_set_value(
	    GTK_SPIN_BUTTON(d->output_height_spin),
	    (gfloat)output->height
	);
	gtk_spin_button_set_value(
	    GTK_SPIN_BUTTON(d->output_frame_spin),
	    (gfloat)(d->src_frame_num + 1)
	);
	gtk_toggle_button_set_active(
	    GTK_TOGGLE_BUTTON(d->output_maintain_aspect_check),
	    d->output_maintain_aspect
	);

	d->freeze_count--;
}

/*
 *	Calculates the position of the image relative to the GtkWidget
 *	based on the PrintDlg's current medium and output geometry.
 *
 *	The d specifies the PrintDlg who's medium and output geometry
 *	will be used to calculate the coordinates.
 *
 *	The w specifies the GtkWidget to calculate the coordinates
 *	relative to. This GtkWidget must have its own GdkWindow.
 *
 *	The x_rtn and y_rtn specifies the return values for the
 *	calculated coordinates relative to the upper left corner of the
 *	specified GtkWidget.
 *
 *	Returns TRUE if the coordinates were calculated or FALSE on
 *	error.
 */
static gboolean PrintDlgTranslateOutputCoordinates(
	PrintDlg *d,
	GtkWidget *w,
	gint *x_rtn, gint *y_rtn
)
{
	gint		ww, wh;
	gfloat otow_coeff;
	GdkRectangle	*medium = &d->medium_size,
			*output = &d->output_img_geometry;
	GdkWindow *window = w->window;

	if((window == NULL) || (medium->width <= 0) || (medium->height <= 0))
	{
	    if(x_rtn != NULL)
		*x_rtn = 0;
	    if(y_rtn != NULL)
		*y_rtn = 0;
	    return(TRUE);
	}

	/* Get the size of the specified GtkWidget */
	gdk_window_get_size(window, &ww, &wh);

	/* Calculate the coordinates */
	if(x_rtn != NULL)
	{
	    otow_coeff = (gfloat)ww / (gfloat)medium->width;
	    *x_rtn = output->x * otow_coeff;
	}
	if(y_rtn != NULL)
	{
	    otow_coeff = (gfloat)wh / (gfloat)medium->height;
	    *y_rtn = output->y * otow_coeff;
	}

	return(TRUE);
}


/*
 *	Copies the image data from src_img to tar_img.
 *
 *	Both images must have the same size and depth.
 */
static void PrintDlgRevertImages(
	imgview_image_struct *src_img,
	imgview_image_struct *tar_img
)
{
	gint		bpp,
			src_bpl, tar_bpl,
			min_bpl,
			height;
	const guint8	*src_data,
			*src_line,
			*src_line_end;
	guint8		*tar_data,
			*tar_line;
	imgview_frame_struct	*src_frame,
				*tar_frame;

	if((src_img == NULL) || (tar_img == NULL))
	    return;

	/* Source image geometry and bpp must match the target image */
	if((src_img->bpp != tar_img->bpp) ||
	   (src_img->width != tar_img->width) ||
	   (src_img->height != tar_img->height)
	)
	{
	    g_printerr(
"PrintDlgRevertImages(): BPP or geometry mismatch, cannot copy.\n"
	    );
	    return;
	}

	bpp = src_img->bpp;
	height = src_img->height;

	src_bpl = src_img->bpl;
	tar_bpl = tar_img->bpl;
	min_bpl = src_img->width * bpp;

	src_frame = ImgViewImageGetFrame(src_img, 0);
	tar_frame = ImgViewImageGetFrame(tar_img, 0);

	/* Get pointers to both image's buffers, both must be allocated */
	src_data = src_frame->data;
	tar_data = tar_frame->data;
	if((src_data == NULL) || (tar_data == NULL))
	    return;

	for(src_line = src_data,
	    src_line_end = src_line + (src_bpl * height),
	    tar_line = tar_data;
	    src_line < src_line_end;
	    src_line += src_bpl,
	    tar_line += tar_bpl
	)
	    memcpy(tar_line, src_line, min_bpl);
}


/*
 *	Connects to the X Print server and gets the listing of
 *	printers from it. The first printer will be selected and
 *	subsequently the values for the first printer. The preview
 *	will be updated.
 *
 *	The xprint_server_address specifies the address to the X
 *	Print server. If xprint_server_address is NULL then the default
 *	address will be used instead.
 */
static void PrintDlgConnectToXPrintServer(
	PrintDlg *d,
	const gchar *xprint_server_address,
	const gboolean verbose
)
{
	gchar *xprint_server_address_d = STRDUP(xprint_server_address);
	const gchar *printer_name;
	GdkRectangle *geo;
	IVPrintXPrintPrinter *p;

	/* Update the X Print server address entry */
	if(xprint_server_address_d != NULL)
	    gtk_entry_set_text(
		GTK_ENTRY(d->xprint_server_address_entry),
		xprint_server_address_d
	    );

	/* Get the list of printers */
	PrintDlgUpdateXPrintServerPrinters(
	    d,
	    xprint_server_address_d,
	    TRUE				/* Verbose */
	);

	/* Get values for the selected printer, updating the
	 * mediums list, orientations list, and resolutions list
	 */
	p = IV_PRINT_XPRINT_PRINTER(PUListBoxGetSelectedData(
	    d->printer_pulistbox
	));
	if(p != NULL)
	    printer_name = p->name;
	else
	    printer_name = NULL;
	PrintDlgUpdatePrinterValues(
	    d,
	    printer_name,
	    FALSE				/* Not verbose */
	);

	/* Update the output values and the preview */
	geo = &d->output_img_geometry;
	geo->x = 0;
	geo->y = 0;
	PrintDlgOutputSetWidgets(d);
	PrintDlgUpdate(d);
	PreviewResize(d->preview);

	g_free(xprint_server_address_d);
}

/*
 *	Updates the list of printers from the X Print server.
 *
 *	The xprint_server_address specifies the address to the X
 *	Print server. If xprint_server_address is NULL then the default
 *	address will be used instead.
 */
static void PrintDlgUpdateXPrintServerPrinters(
	PrintDlg *d,
	const gchar *xprint_server_address,
	const gboolean verbose
)
{
	GList *printers_list;
	GtkWidget	*pulistbox,
			*pulist;

	/* Get the printers list */
	pulistbox = d->printer_pulistbox;
	pulist = PUListBoxGetPUList(pulistbox);
	PUListClear(pulist);
	PUListBoxSetLinesVisible(pulistbox, 1);
	printers_list = iv_print_xprint_get_printers_list(
	    xprint_server_address
	);
	if(printers_list != NULL)
	{
	    gint i = 0;
	    gchar *s;
	    GList *glist;
	    IVPrintXPrintPrinter *printer;

	    for(glist = printers_list;
		glist != NULL;
		glist = g_list_next(glist)
	    )
	    {
		printer = IV_PRINT_XPRINT_PRINTER(glist->data);
		if(printer == NULL)
		    continue;

		if(STRISEMPTY(printer->name))
		    continue;

		if(printer->description != NULL)
		    s = g_strconcat(
			printer->name,
			" - ",
			printer->description,
			NULL
		    );
		else
		    s = g_strdup(printer->name);
		i = PUListAddItem(pulist, s);
		g_free(s);

		PUListSetItemDataFull(
		    pulist,
		    i,
		    iv_print_xprint_printer_copy(printer),
		    (GtkDestroyNotify)iv_print_xprint_printer_delete
		);
	    }

	    PUListBoxSetLinesVisible(pulistbox, i + 1);
	    PUListBoxSelect(pulistbox, 0);

	    iv_print_xprint_printers_list_delete(printers_list);
	}
	else if(verbose)
	{
	    gchar *msg = g_strdup_printf(
"No printers found or unable to connect to the X Print server at:\n\
\n\
    %s",
		xprint_server_address
	    );
	    CDialogSetTransientFor(d->toplevel);
	    CDialogGetResponse(
		"No Printers Found",
		msg,
		NULL,
		CDIALOG_ICON_WARNING,
		CDIALOG_BTNFLAG_OK,
		CDIALOG_BTNFLAG_OK
	    );
	    CDialogSetTransientFor(NULL);
	    g_free(msg);
	}
}

/*
 *	Updates the printer's tray, medium, orientation, and resolution
 *	values from the X Print server.
 *
 *	The tray, medium, orientation, and resolution lists will be
 *	updated and the first value in each will be selected.
 *
 *	The printer_name specifies the printer. If printer_name is
 *	NULL then the lists will be cleared.
 *
 *	If verbose is TRUE then any errors or problems that are
 *	encountered will be displayed.
 */
static void PrintDlgUpdatePrinterValues(
	PrintDlg *d,
	const gchar *printer_name,
	const gboolean verbose
)
{
	const gchar *xprint_server_address = gtk_entry_get_text(
	    GTK_ENTRY(d->xprint_server_address_entry)
	);
	GList *glist;
	GtkWidget	*pulistbox,
			*pulist;

	/* Update the mediums list from the X Print server */
	PrintDlgUpdateMediumsList(
	    d,
	    printer_name,			/* Printer name */
	    verbose				/* Verbose */
	);

	/* Orientations */
	pulistbox = d->orientation_pulistbox;
	pulist = PUListBoxGetPUList(d->orientation_pulistbox);
	PUListClear(pulist);
	PUListBoxSetLinesVisible(pulistbox, 1);
	glist = iv_print_xprint_get_printer_orientations(
	    xprint_server_address,
	    printer_name
	);
	if(glist != NULL)
	{
	    gint i = 0;
	    GList *glist2 = glist;
	    while(glist2 != NULL)
	    {
		i = PUListAddItem(pulist, (const gchar *)glist2->data);
		PUListSetItemDataFull(
		    pulist, i,
		    STRDUP(glist2->data),
		    (GtkDestroyNotify)g_free
		);
		glist2 = g_list_next(glist2);
	    }
	    PUListBoxSetLinesVisible(pulistbox, i + 1);
	    PUListBoxSelect(pulistbox, 0);
	    g_list_foreach(glist, (GFunc)g_free, NULL);
	    g_list_free(glist);
	}

	/* Resolutions */
	pulistbox = d->resolution_pulistbox;
	pulist = PUListBoxGetPUList(pulistbox);
	PUListClear(pulist);
	PUListBoxSetLinesVisible(pulistbox, 1);
	glist = iv_print_xprint_get_printer_resolutions(
	    xprint_server_address,
	    printer_name
	);
	if(glist != NULL)
	{
	    gint i = 0;
	    gchar *s;
	    IVPrintXPrintResolution	*r,
					*def_r = iv_print_xprint_get_default_resolution(
		xprint_server_address,
		printer_name
	    );
	    GList *glist2 = glist;
	    while(glist2 != NULL)
	    {
		r = IV_PRINT_XPRINT_RESOLUTION(glist2->data);
		if(r != NULL)
		{
		    if(STRISEMPTY(r->name))
			s = g_strdup_printf(
			    "%i x %i dpi",
			    r->x_dpi, r->y_dpi
			);
		    else
			s = g_strdup_printf(
			    "%s (%i x %i dpi)",
			    r->name,
			    r->x_dpi,
			    r->y_dpi
			);
		    i = PUListAddItem(pulist, s);
		    g_free(s);
		    if(i > -1)
		    {
			PUListSetItemDataFull(
			    pulist,
			    i,
			    iv_print_xprint_resolution_copy(r),
			    (GtkDestroyNotify)iv_print_xprint_resolution_delete
			);
			/* Select the default resolution */
			if(def_r != NULL)
			{
			    if((def_r->x_dpi == r->x_dpi) &&
			       (def_r->y_dpi == r->y_dpi)
			    )
				PUListBoxSelect(pulistbox, i);
			}
		    }
		}
		glist2 = g_list_next(glist2);
	    }
	    PUListBoxSetLinesVisible(pulistbox, i + 1);
	    iv_print_xprint_resolutions_list_delete(glist);
	    iv_print_xprint_resolution_delete(def_r);
	}
}

/*
 *	Updates the printer's mediums list from the X Print server.
 *
 *	The printer_name specifies the printer.
 *
 *	The tray_name specifies the tray. If tray_name is NULL then
 *	the mediums from all the trays will be added to the list.
 *	the default tray will be used.
 *
 *	If verbose is TRUE then any errors or problems that are
 *	encountered will be displayed.
 */
static void PrintDlgUpdateMediumsList(
	PrintDlg *d,
	const gchar *printer_name,
	const gboolean verbose
)
{
	const gchar *xprint_server_address = gtk_entry_get_text(
	    GTK_ENTRY(d->xprint_server_address_entry)
	);
	GList *mediums_list;
	GtkCList *clist = GTK_CLIST(d->medium_ctree);
	GtkCTree *ctree = GTK_CTREE(clist);

	gtk_clist_freeze(clist);

	/* Delete all the existing nodes */
	gtk_clist_clear(clist);
	gtk_clist_columns_autosize(clist);

	/* Get the new mediums list */
	mediums_list = iv_print_xprint_get_printer_mediums(
	    xprint_server_address,
	    printer_name
	);
	if(mediums_list != NULL)
	{
	    const gint ncolumns = MAX(clist->columns, 1);
	    gint i;
	    gchar	*s,
			**strv = (gchar **)g_malloc(ncolumns * sizeof(gchar *));
	    GList *glist;
	    GtkCTreeNode	*node,
				*parent_node,
				*first_medium_node;
	    IVPrintXPrintMedium *m;

	    for(i = 0; i < ncolumns; i++)
		strv[i] = "";

	    /* Add the mediums to the mediums GtkCTree */
	    parent_node = NULL;
	    first_medium_node = NULL;
	    for(glist = mediums_list;
		glist != NULL;
		glist = g_list_next(glist)
	    )
	    {
		m = IV_PRINT_XPRINT_MEDIUM(glist->data);
		if(m == NULL)
		    continue;

		/* If this medium has a specified tray then find the
		 * parent node who's tray matches this medium's tray
		 */
		if(!STRISEMPTY(m->tray))
		{
		    IVPrintXPrintMedium *m_tray;

		    parent_node = NULL;		/* Clear before searching */

		    for(node = gtk_ctree_node_nth(ctree, 0);
			node != NULL;
			node = GTK_CTREE_NODE_SIBLING(node)
		    )
		    {
			m_tray = IV_PRINT_XPRINT_MEDIUM(gtk_ctree_node_get_row_data(
			    ctree,
			    node
			));
			if(m_tray == NULL)
			    continue;

			if(STRISEMPTY(m_tray->tray))
			    continue;

			if(!g_strcasecmp(m->tray, m_tray->tray))
			{
			    parent_node = node;
			    break;
			}
		    }
		}

		/* If no toplevel tray node has been created for this
		 * this tray then create a new one
		 */
		if(parent_node == NULL)
		{
		    /* Count the number of existing trays */
		    gint ntrays = 0;
		    for(node = gtk_ctree_node_nth(ctree, 0);
			node != NULL;
			node = GTK_CTREE_NODE_SIBLING(node)
		    )
			ntrays++;

		    /* Set the new tray's name */
		    if(!STRISEMPTY(m->tray))
			s = STRDUP(m->tray);
		    else if(ntrays > 0)
			s = g_strdup_printf(
			    "Tray #%i",
			    ntrays + 1
			);
		    else
			s = STRDUP("Default Tray");
		    if(ncolumns > 0)
			strv[0] = s;
		    parent_node = node = gtk_ctree_insert_node(
			ctree,
			NULL,			/* Toplevel */
			NULL,			/* Append */
			strv,
			2,			/* Spacing */
			d->tray_closed_icon_pm,
			d->tray_closed_icon_mask,
			d->tray_opened_icon_pm,
			d->tray_opened_icon_mask,
			FALSE,			/* Not a leaf */
			(ntrays > 0) ? FALSE : TRUE	/* Expanded */
		    );
		    g_free(s);
		    if(node != NULL)
		    {
			gtk_ctree_node_set_selectable(
			    ctree,
			    node,
			    FALSE
			);
			gtk_ctree_node_set_row_data_full(
			    ctree,
			    node,
			    iv_print_xprint_medium_copy(m),
			    (GtkDestroyNotify)iv_print_xprint_medium_delete
			);
		    }
		}

		/* Append a node to the current parent tray for its
		 * medium
		 */
		if(TRUE)
		{
		    /* Units in inches */
#define MM_TO_INCHES	iv_print_xprint_medium_mm_to_inches
		    if(m->name != NULL)
			s = g_strdup_printf(
			    "%s (%.2fin x %.2fin)",
			    m->name,
			    MM_TO_INCHES(m->width_mm + m->x_mm),
			    MM_TO_INCHES(m->height_mm + m->y_mm)
			);
		    else
			s = g_strdup_printf(
			    "%.2fin x %.2fin",
			    MM_TO_INCHES(m->width_mm + m->x_mm),
			    MM_TO_INCHES(m->height_mm + m->y_mm)
			);
#undef MM_TO_INCHES
		}
		else
		{
		    /* Units in millimeters */
		    if(m->name != NULL)
			s = g_strdup_printf(
			    "%s (%.0fmm x %.0fmm)",
			    m->name,
			    m->width_mm + m->x_mm,
			    m->height_mm + m->y_mm
			);
		    else
			s = g_strdup_printf(
			    "%.0fmm x %.0fmm",
			    m->width_mm + m->x_mm,
			    m->height_mm + m->y_mm
			);
		}
		if(ncolumns > 0)
		    strv[0] = s;
		node = gtk_ctree_insert_node(
		    ctree,
		    parent_node,
		    NULL,			/* Append */
		    strv,
		    2,				/* Spacing */
		    d->medium_icon_pm,
		    d->medium_icon_mask,
		    d->medium_icon_pm,
		    d->medium_icon_mask,
		    TRUE,			/* Is a leaf */
		    FALSE			/* Not expanded */
		);
		g_free(s);
		if(node != NULL)
		{
		    gtk_ctree_node_set_row_data_full(
			ctree,
			node,
			iv_print_xprint_medium_copy(m),
			(GtkDestroyNotify)iv_print_xprint_medium_delete
		    );

		    /* Record the first medium node */
		    if(first_medium_node == NULL)
			first_medium_node = node;
		}
	    }
	    gtk_clist_columns_autosize(clist);
	    if(first_medium_node != NULL)
		gtk_ctree_select(ctree, first_medium_node);
	    iv_print_xprint_mediums_list_delete(mediums_list);
	    g_free(strv);
	}
	else if(verbose)
	{
	    gchar *msg = g_strdup_printf(
"No mediums found on printer \"%s\".",
		printer_name
	    );
	    CDialogSetTransientFor(d->toplevel);
	    CDialogGetResponse(
		"No Printer Mediums Found",
		msg,
		NULL,
		CDIALOG_ICON_WARNING,
		CDIALOG_BTNFLAG_OK,
		CDIALOG_BTNFLAG_OK
	    );
	    CDialogSetTransientFor(NULL);
	    g_free(msg);
	}
	gtk_clist_thaw(clist);
}


/*
 *	Preview toplevel GtkEventBox "size_allocate" signal callback.
 */
static void PreviewToplevelSizeAllocateCB(
	GtkWidget *widget, GtkAllocation  *allocation, gpointer data
)
{
	Preview *p = PREVIEW(data);
	if((widget == NULL) || (allocation == NULL) || (p == NULL))
	    return;

	PreviewResize(p);
}

/*
 *	Preview GtkDrawingArea event signal callback.
 */
static gint PreviewEventCB(
	GtkWidget *widget, GdkEvent *event, gpointer data
)
{
	gint status = FALSE;
	gboolean	need_grab_pointer = FALSE,
			press;
	guint keyval;
	GdkEventConfigure *configure;
	GdkEventCrossing *crossing;
	GdkEventKey *key;
	GdkEventButton *button;
	GdkEventMotion *motion;
	Preview *p = PREVIEW(data);
	if((widget == NULL) || (event == NULL) || (p == NULL))
	    return(status);

	switch((gint)event->type)
	{
	  case GDK_CONFIGURE:
	    configure = (GdkEventConfigure *)event;
	    if((configure->width > 0) && (configure->height > 0))
	    {
		GDK_PIXMAP_UNREF(p->pixmap);
		p->pixmap = gdk_pixmap_new(
		    configure->window,
		    configure->width, configure->height,
		    -1
		);
	    }
	    PreviewUpdateImages(p);
	    PreviewQueueDraw(p);
	    status = TRUE;
	    break;

	  case GDK_EXPOSE:
	    PreviewDraw(p);
	    status = TRUE;
	    break;

	  case GDK_ENTER_NOTIFY:
	    crossing = (GdkEventCrossing *)event;
	    if(!crossing->send_event)
	    {
#if 0
		/* Reset all modifiers */
		if(p->drag_modifiers != 0x00000000)
		{
		    p->drag_modifiers = 0x00000000;
		    gdk_window_set_cursor(widget->window, NULL);
		}
#endif
		/* Set up widget to grab focus so "key_press_event" and
		 * and "key_release_event" events will be sent to it
		 */
		gtk_widget_grab_focus(widget);

		/* Force tooltip to show, this will send a synthetic
		 * GDK_ENTER_NOTIFY event which we will ignore
		 */
		GUIShowTipsNow(widget);
		status = TRUE;
	    }
	    break;

	  case GDK_LEAVE_NOTIFY:
	    crossing = (GdkEventCrossing *)event;
	    status = TRUE;
	    break;

	  case GDK_KEY_PRESS:
	  case GDK_KEY_RELEASE:
	    if((p->freeze_count > 0) ||
	       !(p->flags & PREVIEW_INTERACTIVE)
	    )
		return(FALSE);
	    key = (GdkEventKey *)event;
	    keyval = key->keyval;
	    press = (key->type == GDK_KEY_PRESS) ? TRUE : FALSE;
	    /* Handle by keyval */
	    /* ALT? */
	    if((keyval == GDK_Alt_L) || (keyval == GDK_Alt_R))
	    {
		if(press)
		    p->drag_modifiers |= PREVIEW_DRAG_RESIZE;
		else
		    p->drag_modifiers &= ~PREVIEW_DRAG_RESIZE;
		status = TRUE;
	    }
	    /* CTRL? */
	    else if((keyval == GDK_Control_L) || (keyval == GDK_Control_R))
	    {
		if(press)
		    p->drag_modifiers |= PREVIEW_DRAG_TRANSLATE;
		else
		    p->drag_modifiers &= ~PREVIEW_DRAG_TRANSLATE;
		status = TRUE;
	    }
	    /* SHIFT? */
	    else if((keyval == GDK_Shift_L) || (keyval == GDK_Shift_R))
	    {
		if(press)
		    p->drag_modifiers |= PREVIEW_DRAG_RESIZE;
		else
		    p->drag_modifiers &= ~PREVIEW_DRAG_RESIZE;
		status = TRUE;
	    }
	    if(status)
	    {
		/* Is there currently no drag operation in progress? */
		if(p->drag_operation == PREVIEW_DRAG_NONE)
		{
		    GdkCursor *cursor = NULL;

		    if(p->drag_modifiers & PREVIEW_DRAG_TRANSLATE)
			cursor = p->translate_cur;
		    if(p->drag_modifiers & PREVIEW_DRAG_RESIZE)
			cursor = p->resize_cur;

		    gdk_window_set_cursor(widget->window, cursor);
		}
	    }
	    break;

	  case GDK_BUTTON_PRESS:
	    if((p->freeze_count > 0) ||
	       !(p->flags & PREVIEW_INTERACTIVE)
	    )
		return(FALSE);
	    button = (GdkEventButton *)event;
	    switch(button->button)
	    {
	      case GDK_BUTTON1:		/* Translate */
		if(p->drag_modifiers & PREVIEW_DRAG_RESIZE)
		{
		    /* Resize */
		    need_grab_pointer = TRUE;

		    p->drag_operation |= PREVIEW_DRAG_RESIZE;
		    p->drag_last_x = (gint)button->x;
		    p->drag_last_y = (gint)button->y;

		    gdk_window_set_cursor(widget->window, p->resize_cur);

		    if(PrintDlgTranslateOutputCoordinates(
			p->print_dlg,
			widget,			/* Calculate relative to the
						 * preview GtkDrawingArea */
			&p->drag_outline_x[0], &p->drag_outline_y[0]
		    ))
		    {
			imgview_image_struct *img = p->rendered_img;
			if(img != NULL)
			{
			    p->drag_outline_x[1] = p->drag_outline_x[0] + img->width;
			    p->drag_outline_y[1] = p->drag_outline_y[0] + img->height;
			}
		    }
		    else
		    {
			p->drag_outline_x[0] = 0;
			p->drag_outline_y[0] = 0;
		    }
		    p->flags &= ~PREVIEW_DREW_PREVIOUS_OUTLINE;
		    p->flags |= PREVIEW_DRAW_DRAG_OUTLINE;
		    PreviewQueueDraw(p);
		}
		else
		{
		    /* Translate */
		    need_grab_pointer = TRUE;
		    p->drag_operation |= PREVIEW_DRAG_TRANSLATE;
		    p->drag_last_x = (gint)button->x;
		    p->drag_last_y = (gint)button->y;

		    gdk_window_set_cursor(widget->window, p->translate_cur);

		    if(PrintDlgTranslateOutputCoordinates(
		        p->print_dlg,
			widget,			/* Calculate relative to the
						 * preview GtkDrawingArea */
			&p->drag_outline_x[0], &p->drag_outline_y[0]
		    ))
		    {
			imgview_image_struct *img = p->rendered_img;
			if(img != NULL)
			{
			    p->drag_outline_x[1] = p->drag_outline_x[0] + img->width;
			    p->drag_outline_y[1] = p->drag_outline_y[0] + img->height;
			}
		    }
		    else
		    {
		        p->drag_outline_x[0] = 0;
		        p->drag_outline_y[0] = 0;
		    }
		    p->flags &= ~PREVIEW_DREW_PREVIOUS_OUTLINE;
		    p->flags |= PREVIEW_DRAW_DRAG_OUTLINE;
		    PreviewQueueDraw(p);
		}
		status = TRUE;
		break;

	      case GDK_BUTTON2:		/* Resize */
		need_grab_pointer = TRUE;

		p->drag_operation |= PREVIEW_DRAG_RESIZE;
		p->drag_last_x = (gint)button->x;
		p->drag_last_y = (gint)button->y;

		gdk_window_set_cursor(widget->window, p->resize_cur);

		if(PrintDlgTranslateOutputCoordinates(
		    p->print_dlg,
		    widget,			/* Calculate relative to the
						 * preview GtkDrawingArea */
		    &p->drag_outline_x[0], &p->drag_outline_y[0]
		))
		{
		    imgview_image_struct *img = p->rendered_img;
		    if(img != NULL)
		    {
			p->drag_outline_x[1] = p->drag_outline_x[0] + img->width;
			p->drag_outline_y[1] = p->drag_outline_y[0] + img->height;
		    }
		}
		else
		{
		    p->drag_outline_x[0] = 0;
		    p->drag_outline_y[0] = 0;
		}
		/* Draw the outline */
		p->flags &= ~PREVIEW_DREW_PREVIOUS_OUTLINE;
		p->flags |= PREVIEW_DRAW_DRAG_OUTLINE;
		PreviewQueueDraw(p);
		status = TRUE;
		break;

	      case GDK_BUTTON4:		/* Shrink */
		if(p->drag_modifiers == 0x00000000)
		{
		    PrintDlg *d = p->print_dlg;
		    GtkAdjustment *adj = gtk_spin_button_get_adjustment(
			GTK_SPIN_BUTTON(d->output_width_spin)
		    );
		    GdkRectangle *geo = &d->output_img_geometry;
		    if(adj != NULL)
		    {
			const gfloat aspect = (geo->height > 0) ?
			    ((gfloat)geo->width / (gfloat)geo->height) : 1.0f;
			if(aspect > 0.0f)
			{
			    geo->width -= MAX(
				adj->page_increment,
				adj->step_increment
			    );
			    if(geo->width < 1)
				geo->width = 1;

			    geo->height = MAX(geo->width / aspect, 1);

			    PrintDlgOutputSetWidgets(d);
			    PreviewUpdateImages(p);
			    PreviewQueueDraw(p);
			}
		    }
		}
		status = TRUE;
		break;

	      case GDK_BUTTON5:		/* Increase size  */
		if(p->drag_modifiers == 0x00000000)
		{
		    PrintDlg *d = p->print_dlg;
		    GtkAdjustment *adj = gtk_spin_button_get_adjustment(
			GTK_SPIN_BUTTON(d->output_width_spin)
		    );
		    GdkRectangle *geo = &d->output_img_geometry;
		    if(adj != NULL)
		    {
			const gfloat aspect = (geo->height > 0) ?
			    ((gfloat)geo->width / (gfloat)geo->height) : 1.0f;
			if(aspect > 0.0f)
			{
			    geo->width += MAX(
				adj->page_increment,
				adj->step_increment
			    );
			    if(geo->width < 1)
				geo->width = 1;

			    geo->height = MAX(geo->width / aspect, 1);

			    PrintDlgOutputSetWidgets(d);
			    PreviewUpdateImages(p);
			    PreviewQueueDraw(p);
			}
		    }
		}
		status = TRUE;
		break;
	    }

	    /* Grab pointer? */
	    if(need_grab_pointer)
	    {
		gdk_pointer_grab(
		    widget->window,
		    FALSE,
		    GDK_BUTTON_RELEASE_MASK | GDK_POINTER_MOTION_MASK,
		    NULL,
		    NULL,
		    button->time
		);
		need_grab_pointer = FALSE;
	    }
	    break;

	  case GDK_BUTTON_RELEASE:
	    if((p->freeze_count > 0) ||
	       !(p->flags & PREVIEW_INTERACTIVE)
	    )
		return(FALSE);

	    button = (GdkEventButton *)event;

	    /* Ungrab the pointer */
	    if(gdk_pointer_is_grabbed())
		gdk_pointer_ungrab(button->time);

	    /* Translate */
	    if(p->drag_operation & PREVIEW_DRAG_TRANSLATE)
	    {
		PrintDlg *d = p->print_dlg;
		const gint	ox = MIN(
		    p->drag_outline_x[0], p->drag_outline_x[1]
		),
				oy = MIN(
		    p->drag_outline_y[0], p->drag_outline_y[1]
		);
		gint ww, wh;
		gfloat wtoo_coeff;
		GdkRectangle	*medium = &d->medium_size,
				*output = &d->output_img_geometry;

		gdk_window_get_size(widget->window, &ww, &wh);
		if(ww > 0)
		    wtoo_coeff = (gfloat)medium->width / (gfloat)ww;
		else
		    wtoo_coeff = 0.0f;
		output->x = ox * wtoo_coeff;

		if(wh > 0)
		    wtoo_coeff = (gfloat)medium->height / (gfloat)wh;
		else
		    wtoo_coeff = 0.0f;
		output->y = oy * wtoo_coeff;

		p->drag_operation &= ~PREVIEW_DRAG_TRANSLATE;
		PrintDlgOutputSetWidgets(d);
		p->flags &= ~PREVIEW_DREW_PREVIOUS_OUTLINE;
		p->flags &= ~PREVIEW_DRAW_DRAG_OUTLINE;
		PreviewQueueDraw(p);
		status = TRUE;
	    }
	    /* Resize */
	    if(p->drag_operation & PREVIEW_DRAG_RESIZE)
	    {
		PrintDlg *d = p->print_dlg;
		const gint	ow = ABSOLUTE(
		    p->drag_outline_x[1] - p->drag_outline_x[0]
		),
				oh = ABSOLUTE(
		    p->drag_outline_y[1] - p->drag_outline_y[0]
		);
		gint ww, wh;
		gfloat wtoo_coeff;
		GdkRectangle	*medium = &d->medium_size,
				*geo = &d->output_img_geometry;

		gdk_window_get_size(widget->window, &ww, &wh);

		/* Calculate the output size */
		wtoo_coeff = (ww > 0) ?
		    ((gfloat)medium->width / (gfloat)ww) : 0.0f;
		if((ow > 0) && (wtoo_coeff > 0.0f))
		    geo->width = ow * wtoo_coeff;

		wtoo_coeff = (wh > 0) ?
		    ((gfloat)medium->height / (gfloat)wh) : 0.0f;
		if((oh > 0) && (wtoo_coeff > 0.0f))
		    geo->height = oh * wtoo_coeff;

		p->drag_operation &= ~PREVIEW_DRAG_RESIZE;
		PrintDlgOutputSetWidgets(d);
		PreviewUpdateImages(p);
		p->flags &= ~PREVIEW_DREW_PREVIOUS_OUTLINE;
		p->flags &= ~PREVIEW_DRAW_DRAG_OUTLINE;
		PreviewQueueDraw(p);
		status = TRUE;
	    }
	    /* If there are no drag modifiers then reset the cursor */
	    if(p->drag_modifiers == PREVIEW_DRAG_NONE)
	    {
		gdk_window_set_cursor(widget->window, NULL);
	    }
	    break;

	  case GDK_MOTION_NOTIFY:
	    if((p->freeze_count > 0) ||
	       !(p->flags & PREVIEW_INTERACTIVE)
	    )
		return(FALSE);

	    motion = (GdkEventMotion *)event;
	    if(p->drag_operation != PREVIEW_DRAG_NONE)
	    {
		gint	x, y,
			dx, dy;
		GdkModifierType mask;

		if(motion->is_hint)
		{
		    gdk_window_get_pointer(
			motion->window,
			&x, &y,
			&mask
		    );
		}
		else
		{
		    x = (gint)motion->x;
		    y = (gint)motion->y;
		    mask = motion->state;
		}

		/* Translate */
		if(p->drag_operation & PREVIEW_DRAG_TRANSLATE)
		{
		    dx = x - p->drag_last_x;
		    dy = y - p->drag_last_y;

		    p->drag_outline_x[0] += dx;
		    p->drag_outline_y[0] += dy;
		    p->drag_outline_x[1] += dx;
		    p->drag_outline_y[1] += dy;

		    /* Draw the drag outline */
		    p->flags |= PREVIEW_DRAW_DRAG_OUTLINE;
		    PreviewQueueDraw(p);

		    p->drag_last_x = x;
		    p->drag_last_y = y;

		    status = TRUE;
		}
		/* Resize */
		if(p->drag_operation & PREVIEW_DRAG_RESIZE)
		{
		    PrintDlg *d = p->print_dlg;

		    dx = x - p->drag_last_x;
		    dy = y - p->drag_last_y;

		    /* Apply the delta values to resize drag outline
		     * based on if maintaining aspect or not
		     */
		    if(d->output_maintain_aspect)
		    {
			/* Maintaining aspect, so get aspect and resize
			 * drag output while maintaining aspect
			 */
			gint len;
			gfloat aspect = PrintDlgGetSourceImageAspect(d);
			if(aspect <= 0.0f)
			    aspect = 1.0f;

			/* Use greater of the two compoent movements */
			if(ABSOLUTE(dx) > ABSOLUTE(dy))
			{
			    p->drag_outline_x[1] += dx;
			    len = MAX(
				(p->drag_outline_x[1] - 
				    p->drag_outline_x[0]),
				0
			    );
			    p->drag_outline_y[1] = p->drag_outline_y[0] +
				(len / aspect);
			}
			else
			{
			    p->drag_outline_y[1] += dy;
			    len = MAX(
				(p->drag_outline_y[1] - 
				    p->drag_outline_y[0]),
				0 
			    );
			    p->drag_outline_x[1] = p->drag_outline_x[0] +
				(len * aspect);
			}
		    }
		    else
		    {
			p->drag_outline_x[1] += dx;
			p->drag_outline_y[1] += dy;
		    }
		    /* Make sure that the drag outline does not get
		     * smaller than its origin
		     */
		    if(p->drag_outline_x[1] < p->drag_outline_x[0])
			p->drag_outline_x[1] = p->drag_outline_x[0];
		    if(p->drag_outline_y[1] < p->drag_outline_y[0])
			p->drag_outline_y[1] = p->drag_outline_y[0];

		    /* Draw the drag outline */
		    p->flags |= PREVIEW_DRAW_DRAG_OUTLINE;
		    PreviewQueueDraw(p);

		    p->drag_last_x = x;
		    p->drag_last_y = y;

		    status = TRUE;
		}
	    }
	    break;
	}

	return(status);
}


/*
 *	Create a new Preview.
 */
static Preview *PreviewNew(
	GtkWidget *parent,
	GtkStyle *style,
	GdkColor *bg_color,
	const gboolean interactive,
	PrintDlg *print_dlg
)
{
	const gint border_major = 5;
	GdkWindow *window;
	GtkWidget	*w,
			*parent2, *parent3;

	Preview *p = PREVIEW(g_malloc0(sizeof(Preview)));
	if(p == NULL)
	    return(NULL);

/*	p->freeze_count = 0; */
	p->flags |= (interactive ? PREVIEW_INTERACTIVE : 0);

	p->translate_cur = gdk_cursor_new(GDK_FLEUR);
	p->resize_cur = gdk_cursor_new(GDK_SIZING);

	p->print_dlg = print_dlg;

	p->freeze_count++;

	/* Toplevel GtkEventBox */
	p->toplevel = w = gtk_event_box_new();
	gtk_signal_connect(
	    GTK_OBJECT(w), "size_allocate",
	    GTK_SIGNAL_FUNC(PreviewToplevelSizeAllocateCB), p
	);
	gtk_container_add(GTK_CONTAINER(parent), w);
	gtk_widget_show(w);
	parent2 = w;

	/* Set the background color for the toplevel GtkEventBox (not
	 * the image's background color)
	 */
	if(bg_color != NULL)
	{
	    GtkRcStyle *rcstyle = gtk_rc_style_new();
	    rcstyle->color_flags[GTK_STATE_NORMAL] = GTK_RC_BG;
	    memcpy(
		&rcstyle->bg[GTK_STATE_NORMAL],
		bg_color,
		sizeof(GdkColor)
	    );
	    gtk_widget_modify_style(w, rcstyle);
	    GTK_RC_STYLE_UNREF(rcstyle);
	}

	w = gtk_vbox_new(TRUE, 0);
	gtk_container_add(GTK_CONTAINER(parent2), w);
	gtk_widget_show(w);
	parent2 = w;

	w = gtk_hbox_new(TRUE, 0);
	gtk_box_pack_start(GTK_BOX(parent2), w, TRUE, FALSE, 0);
	gtk_widget_show(w);
	parent2 = w;

	w = gtk_table_new(3, 3, FALSE);
	gtk_table_set_col_spacings(GTK_TABLE(w), 0);
	gtk_table_set_row_spacings(GTK_TABLE(w), 0);
	gtk_box_pack_start(GTK_BOX(parent2), w, TRUE, FALSE, 0);
	gtk_widget_show(w);
	parent3 = w;

	/* Preview GtkDrawingArea */
	p->da = w = gtk_drawing_area_new();
	GTK_WIDGET_SET_FLAGS(w, GTK_CAN_FOCUS | GTK_CAN_DEFAULT);
	gtk_widget_add_events(
	    w,
	    GDK_STRUCTURE_MASK | GDK_EXPOSURE_MASK |
	    GDK_KEY_PRESS_MASK | GDK_KEY_RELEASE_MASK |
	    GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK |
	    GDK_POINTER_MOTION_MASK | GDK_POINTER_MOTION_HINT_MASK |
	    GDK_ENTER_NOTIFY_MASK | GDK_LEAVE_NOTIFY_MASK
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "configure_event",
	    GTK_SIGNAL_FUNC(PreviewEventCB), p
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "expose_event",
	    GTK_SIGNAL_FUNC(PreviewEventCB), p
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "key_press_event",
	    GTK_SIGNAL_FUNC(PreviewEventCB), p
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "key_release_event",
	    GTK_SIGNAL_FUNC(PreviewEventCB), p
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "button_press_event",
	    GTK_SIGNAL_FUNC(PreviewEventCB), p
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "button_release_event",
	    GTK_SIGNAL_FUNC(PreviewEventCB), p
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "motion_notify_event",
	    GTK_SIGNAL_FUNC(PreviewEventCB), p
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "enter_notify_event",
	    GTK_SIGNAL_FUNC(PreviewEventCB), p
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "leave_notify_event",
	    GTK_SIGNAL_FUNC(PreviewEventCB), p
	);
	gtk_table_attach(
	    GTK_TABLE(parent3),
	    w,
	    0, 2,
	    0, 2,
	    GTK_FILL,
	    GTK_FILL,
	    0, 0
	);
	gtk_widget_realize(w);
	window = w->window;
	if(interactive)
	    GUISetWidgetTip(
		w,
"Press and hold Button1 to adjust the position,\
 press and hold SHIFT + Button1 (or Button2) to\
 adjust the size"
	    );
	gtk_widget_show(w);

	/* Create the GdkGCs */
	if(window != NULL)
	{
	    GdkGCValues gcv;
	    GdkGCValuesMask gcv_mask;

	    p->gc = gdk_gc_new(window);

	    gcv_mask = GDK_GC_FUNCTION |
		GDK_GC_LINE_WIDTH | GDK_GC_LINE_STYLE |
		GDK_GC_CAP_STYLE | GDK_GC_JOIN_STYLE;
	    gcv.function = GDK_INVERT;
	    gcv.line_width = 1;
	    gcv.line_style = GDK_LINE_SOLID;
	    gcv.cap_style = GDK_CAP_NOT_LAST;
	    gcv.join_style = GDK_JOIN_MITER;
	    p->invert_gc = gdk_gc_new_with_values(
		window,
		&gcv,
		gcv_mask
	    );
	}

	/* Shadow GtkDrawingAreas */
	w = gtk_drawing_area_new();
	gtk_widget_set_usize(w, border_major, border_major);
	gtk_table_attach(
	    GTK_TABLE(parent3),
	    w,
	    2, 3,
	    0, 1,
	    0,
	    0,
	    0, 0
	);
	if(bg_color != NULL)
	{
	    gtk_widget_realize(w);
	    gdk_window_set_background(w->window, bg_color);
	}
	gtk_widget_show(w);

	p->shadow_y_da = w = gtk_drawing_area_new();
	gtk_widget_set_usize(w, border_major, -1);
	gtk_table_attach(
	    GTK_TABLE(parent3),
	    w,
	    2, 3,
	    1, 2,
	    0,
	    0,
	    0, 0
	);
	gtk_widget_realize(w);
	gdk_window_set_background(w->window, &style->black);
	gtk_widget_show(w);

	w = gtk_drawing_area_new();
	gtk_widget_set_usize(w, border_major, border_major);
	gtk_table_attach(
	    GTK_TABLE(parent3),
	    w,
	    0, 1,
	    2, 3,
	    0,
	    0,
	    0, 0
	);
	if(bg_color != NULL)
	{
	    gtk_widget_realize(w);
	    gdk_window_set_background(w->window, bg_color);
	}
	gtk_widget_show(w);

	p->shadow_x_da = w = gtk_drawing_area_new();
	gtk_widget_set_usize(w, -1, border_major);
	gtk_table_attach(
	    GTK_TABLE(parent3),
	    w,
	    1, 2,
	    2, 3,
	    0,
	    0,
	    0, 0
	);
	gtk_widget_realize(w);
	gdk_window_set_background(w->window, &style->black);
	gtk_widget_show(w);

	w = gtk_drawing_area_new();
	gtk_widget_set_usize(w, border_major, border_major);
	gtk_table_attach(
	    GTK_TABLE(parent3),
	    w,
	    2, 3,
	    2, 3,
	    0,
	    0,
	    0, 0
	);
	gtk_widget_realize(w);
	gdk_window_set_background(w->window, &style->black);
	gtk_widget_show(w);

	p->freeze_count--;

	return(p);
}

/*
 *	Rerenders the Preview's rendered_img from its resized_img
 *	with the current values on the PrintDlg.
 */
static void PreviewUpdateVisual(Preview *p)
{
	PrintDlg *d = p->print_dlg;
	print_values_struct *pv = PrintValuesNew();
	if(pv == NULL)
	    return;

	/* Revert the rendered_img to that of the resized_img */
	PrintDlgRevertImages(
	    p->resized_img,			/* Source */
	    p->rendered_img			/* Target */
	);

	/* Set up the print values from the PrintDlg to render the
	 * rendered_img
	 */
	pv->flags = PRINT_VALUE_BRIGHTNESS | PRINT_VALUE_CONTRAST |
	    PRINT_VALUE_VISUAL;
	pv->brightness = GTK_ADJUSTMENT_GET_VALUE(d->brightness_adj);
	pv->contrast = GTK_ADJUSTMENT_GET_VALUE(d->contrast_adj);
	pv->visual = PrintDlgGetVisual(d);

	/* Render the rendered_img, modify its color, brightness, and
	 * contrast
	 */
	iv_print_adjust_image_color(
	    p->rendered_img,
	    pv,
	    NULL, NULL
	);

	PrintValuesDelete(pv);
}

/*
 *	Recreates the Preview images resized_img and rendered_img.
 */
static void PreviewUpdateImages(Preview *p)
{
	gint		i,
			bpp,
			width, height,
			new_width, new_height;
	gfloat          x_coeff, y_coeff;
	GdkWindow *window;
	GtkWidget *w = p->da;
	PrintDlg *print_dlg = p->print_dlg;
	const gint src_frame_num = print_dlg->src_frame_num;
	GdkRectangle    *medium = &print_dlg->medium_size,
			*geo = &print_dlg->output_img_geometry;
	imgview_frame_struct    *src_frame,
				*tar_frame;
	imgview_image_struct    *src_img = print_dlg->src_img,
				*resized_img,
				*rendered_img;

	/* Delete the existing images */
	ImgViewImageDelete(p->resized_img);
	p->resized_img = NULL;
	ImgViewImageDelete(p->rendered_img);
	p->rendered_img = NULL;

	window = w->window;
	if(window == NULL)
	    return;

	gdk_window_get_size(window, &width, &height);

	/* Get the source image */
	src_frame = ImgViewImageGetFrame(src_img, src_frame_num);
	if(src_frame == NULL)
	    return;
	if(src_frame->data == NULL)
	    return;

	bpp = src_img->bpp;

	if((medium->width <= 0) || (medium->height <= 0) ||
	   (geo->width <= 0) || (geo->height <= 0)
	)
	    return;

	/* Calculate the coefficients that represent the amount of
	 * the output image visible on the paper
	 */
	x_coeff = (gfloat)geo->width / (gfloat)medium->width;
	y_coeff = (gfloat)geo->height / (gfloat)medium->height;

	/* Calculate the size for the resized and rendered images */
	new_width = (gint)(width * x_coeff);
	new_height = (gint)(height * y_coeff);

	if((new_width <= 0) || (new_height <= 0))
	    return;

	/* Create the resized image */
	p->resized_img = resized_img = ImgViewImageNew(
	    new_width, new_height, bpp, 0
	);
	if(resized_img == NULL)
	    return;

	i = ImgViewImageAppendFrameNew(resized_img);
	tar_frame = ImgViewImageGetFrame(resized_img, i);
	if((tar_frame != NULL) ? (tar_frame->data != NULL) : FALSE)
	{
	    /* Copy/resize the source image to the target image */
	    GUIImageBufferResize(
		bpp,
		src_frame->data,
		src_img->width, src_img->height, src_img->bpl,
		tar_frame->data,
		resized_img->width, resized_img->height, resized_img->bpl,
		NULL, NULL
	    );

	    /* Apply the background to the target image */
	    GUIImageBufferFlattenWithBG(
		bpp,
		tar_frame->data,
		resized_img->width, resized_img->height, resized_img->bpl,
		print_dlg->bg_color,
		NULL, NULL
	    );
	}

	/* Create the rendered image */
	p->rendered_img = rendered_img = ImgViewImageNew(
	    new_width, new_height, bpp, 0
	);
	if(rendered_img == NULL)
	{
	    ImgViewImageDelete(resized_img);
	    p->resized_img = NULL;
	    return;
	}

	i = ImgViewImageAppendFrameNew(rendered_img);

	/* Render/apply the color, contrast, and brightness to the
	 * rendered image
	 */
	PreviewUpdateVisual(p);
}

/*
 *	Resizes the GtkDrawingArea based on the current size of the
 *	medium and the size of the toplevel GtkEventBox.
 *
 *	The GtkDrawingArea will be resized as needed and a
 *      "configure_event" signal will be emitted, in which case
 *	PreviewUpdateImages() will be called to recreated the
 *      preview images.
 */
static void PreviewResize(Preview *p)
{
	const gint border_major = 5;
	gint		pv_width, pv_height,
			new_width, new_height;
	gfloat zoom;
	GdkWindow *window;
	GtkWidget *w;
	PrintDlg *print_dlg = p->print_dlg;
	GdkRectangle *medium = &print_dlg->medium_size;

	if((medium->width <= 0) || (medium->height <= 0))
	    return;

	w = p->toplevel;
	window = w->window;
	if(window == NULL)
	    return;

	/* Get the current size of the toplevel GtkEventBox
	 * minus sufficient border
	 */
	gdk_window_get_size(window, &pv_width, &pv_height);
	pv_width -= (4 * border_major);
	pv_height -= (4 * border_major);
	if((pv_width <= 0) || (pv_height <= 0))
	    return;

	/* Calculate the zoom to fit the display of the medium
	 * within the preview GtkDrawingArea
	 */
	zoom = (gfloat)pv_width / (gfloat)medium->width;
	if(zoom <= 0.0f)
	    return;

	if(pv_height < (medium->height * zoom))
	    zoom = (gfloat)pv_height / (gfloat)medium->height;

	/* Calculate the new size of the preview GtkDrawingArea */
	new_width = medium->width * zoom;
	new_height = medium->height * zoom;

	/* Get the current size of the preview GtkDrawingArea and
	 * check if we need to set the new size
	 */
	if((p->last_set_width != new_width) ||
	   (p->last_set_height != new_height)
	)
	{
	    /* Record the new size */
	    p->last_set_width = new_width;
	    p->last_set_height = new_height;

	    /* Set the new size of preview GtkDrawingArea and its shadows */
	    gtk_widget_set_usize(
		p->da,
		new_width, new_height
	    );
	    gtk_widget_set_usize(
		p->shadow_x_da,
		new_width, border_major
	    );
	    gtk_widget_set_usize(
		p->shadow_y_da,
		border_major, new_height
	    );

	    /* Get the preview GtkDrawingArea and its shadows to line
	     * up properly
	     */
	    gtk_container_check_resize(GTK_CONTAINER(p->toplevel));

	    /* Queue the toplevel GtkWindow to resize, this will cause
	     * the preview_parent to emit another "size_allocate"
	     * signal but this function will not queue again if the
	     * calculated size of the preview_da does not change
	     */
	    w = gtk_widget_get_toplevel(p->toplevel);
	    if(w != NULL)
		gtk_widget_queue_resize(w);
	}
}

/*
 *	Draws the Preview.
 */
static void PreviewDraw(Preview *p)
{
	gboolean only_draw_outline;
	gint ww, wh;
	GdkWindow *window;
	GdkDrawable *drawable;
	GtkStateType state;
	GtkStyle *style;
	GtkWidget *w = p->da;

	/* Get marker for if we only want to draw the drag outlines */
	only_draw_outline = (p->flags & PREVIEW_DRAW_DRAG_OUTLINE) ? TRUE : FALSE;

	/* Clear the draw drag outline flag so that it does not get
	 * drawn again
	 */
	p->flags &= ~PREVIEW_DRAW_DRAG_OUTLINE;

	window = w->window;
	state = GTK_WIDGET_STATE(w);
	style = gtk_widget_get_style(w);
	if(window == NULL)
	    return;

	drawable = (GdkDrawable *)((p->pixmap != NULL) ?
	    p->pixmap : window
	);

	gdk_window_get_size((GdkWindow *)drawable, &ww, &wh);
	if((ww <= 0) || (wh <= 0))
	    return;

	/* Only draw the outline? */
	if(only_draw_outline)
	{
	    GdkGC *gc = p->invert_gc;

	    /* Clear the previous drag outline if it was drawn */
	    if(p->flags & PREVIEW_DREW_PREVIOUS_OUTLINE)
	    {
		gint    x = MIN(p->prev_outline_x[0], p->prev_outline_x[1]),
			y = MIN(p->prev_outline_y[0], p->prev_outline_y[1]),
			width = ABSOLUTE(
		p->prev_outline_x[1] - p->prev_outline_x[0]
			 ),
			 height = ABSOLUTE(
		p->prev_outline_y[1] - p->prev_outline_y[0]
			 );

		if((width > 0) && (height > 0))
		    gdk_draw_rectangle(
			drawable,
			gc, FALSE, x, y, width, height
		    );

		/* Clear the have previous outline flag */
		p->flags &= ~PREVIEW_DREW_PREVIOUS_OUTLINE;
	    }

	    /* Draw new drag marks? */
	    if(p->drag_operation != PREVIEW_DRAG_NONE)
	    {
		gint    x = MIN(p->drag_outline_x[0], p->drag_outline_x[1]),
			y = MIN(p->drag_outline_y[0], p->drag_outline_y[1]),
			width = ABSOLUTE(
		p->drag_outline_x[1] - p->drag_outline_x[0]
			),
			height = ABSOLUTE(
		p->drag_outline_y[1] - p->drag_outline_y[0]
			);

		if((width > 0) && (height > 0))
		{
		    gdk_draw_rectangle(
			drawable,
			gc,
			FALSE,                  /* Outline */
			x, y,
			width, height
		    );

		    /* Record the newly drawn drag outline marks */
		    p->flags |= PREVIEW_DREW_PREVIOUS_OUTLINE;
		    p->prev_outline_x[0] = p->drag_outline_x[0];
		    p->prev_outline_y[0] = p->drag_outline_y[0];
		    p->prev_outline_x[1] = p->drag_outline_x[1];
		    p->prev_outline_y[1] = p->drag_outline_y[1];
		}
	    }
	}
	else
	{
	    /* Draw the preview normally */
	    GdkGC *gc = p->gc;
	    imgview_image_struct *img = p->rendered_img;
	    imgview_frame_struct *frame = ImgViewImageGetFrame(img, 0);

	    /* Background */
	    gdk_draw_rectangle(
		drawable,
		style->white_gc,
		TRUE,
		0, 0,
		ww, wh
	    );

	    /* Outline */
	    gdk_draw_rectangle(
		drawable,
		style->black_gc,
		FALSE,                          /* Outline */
		0, 0,
		ww - 1, wh - 1
	    );

	    /* Draw the preview image */
	    if((frame != NULL) ? (frame->data != NULL) : FALSE)
	    {
		gint x, y;

		(void)PrintDlgTranslateOutputCoordinates(
		    p->print_dlg,
		    w,                          /* Calculate relative to the
						 * preview GtkDrawingArea */
		    &x, &y
		);

		switch(img->bpp)
		{
		  case 4:
		    gdk_draw_rgb_32_image(
			drawable,
			gc,
			x, y,
			img->width, img->height,
			GDK_RGB_DITHER_NONE,
			(guchar *)frame->data,
			img->bpl
		    );
		    break;
		  case 3:
		    gdk_draw_rgb_image(
			drawable,
			gc,
			x, y,
			img->width, img->height,
			GDK_RGB_DITHER_NONE,
			(guchar *)frame->data,
			img->bpl
		    );
		    break;
		  case 1:
		    gdk_draw_gray_image(
			drawable,
			gc,
			x, y,
			img->width, img->height,
			GDK_RGB_DITHER_NONE,
			(guchar *)frame->data,
			img->bpl
		    );
		    break;
		}
	    }
	    else if(style->font != NULL)
	    {
		const gchar *s = "No Image Data";
		GdkFont *font = style->font;
		const gint font_height = font->ascent + font->descent;
		GdkTextBounds b;
		gdk_string_bounds(font, s, &b);
		gdk_draw_string(
		    drawable,
		    font,
		    style->text_gc[state],
		    ((ww - b.width) / 2) - b.lbearing,
		    ((wh - font_height) / 2) + font->ascent,
		    s
		);
	    }
	}

	/* Send the drawable to the window if it is not the window */
	if(drawable != (GdkDrawable *)window)
	    gdk_draw_pixmap(
		window,
		p->gc,
		drawable,
		0, 0,
		0, 0,
		ww, wh
	    );
}

/*
 *	Queues the Preview to draw.
 */
static void PreviewQueueDraw(Preview *p)
{
	gtk_widget_queue_draw(p->da);
}

/*
 *	Deletes the Preview.
 */
static void PreviewDelete(Preview *p)
{
	if(p == NULL)
	    return;

	p->freeze_count++;

	/* Delete the preview images */
	gdk_flush();
	ImgViewImageDelete(p->rendered_img);
	p->rendered_img = NULL;
	ImgViewImageDelete(p->resized_img);
	p->resized_img = NULL;

	GTK_WIDGET_DESTROY(p->toplevel);

	GDK_GC_UNREF(p->gc);
	GDK_GC_UNREF(p->invert_gc);

	GDK_PIXMAP_UNREF(p->pixmap);

	GDK_CURSOR_DESTROY(p->translate_cur);
	GDK_CURSOR_DESTROY(p->resize_cur);

	p->freeze_count--;

	g_free(p);
}


/*
 *	Toplevel GtkWindow "delete_event" signal callback.
 */
static gint PrintDlgDeleteEventCB(GtkWidget *widget, GdkEvent *event, gpointer data)
{
	PrintDlg *d = PRINT_DLG(data);
	if((widget == NULL) || (d == NULL))
	    return(FALSE);

	PrintDlgCancelCB(d->cancel_btn, d);

	return(TRUE);
}

/*
 *	PrintDlg GtkAdjustment "value_changed" signal callback.
 */
static void PrintDlgVisualAdjustmentValueChangedCB(
	GtkAdjustment *adj, gpointer data
)
{
	PrintDlg *d = PRINT_DLG(data);
	if((adj == NULL) || (d == NULL))
	    return;

	if(d->freeze_count > 0)
	    return;

	d->freeze_count++;

	/* Rerender the preview image, apply the color, contrast, and
	 * brightness to it
	 */
	PreviewUpdateVisual(d->preview);
	PreviewQueueDraw(d->preview);

	d->freeze_count--;
}

/*
 *	Visual GtkToggleButton "toggled" signal callback.
 */
static void PrintDlgVisualToggledCB(GtkWidget *widget, gpointer data)
{
	PrintDlg *d = PRINT_DLG(data);
	if((widget == NULL) || (d == NULL))
	    return;

	if(d->freeze_count > 0)
	    return;

	/* Only handle toggled to TRUE signals */
	if(GTK_IS_TOGGLE_BUTTON(widget) ?
	    !GTK_TOGGLE_BUTTON_GET_ACTIVE(widget) : TRUE
	)
	    return;

	d->freeze_count++;

	/* Visual Black & White, Greyscale, or Color GtkRadioButton? */
	if((widget == d->visual_bw_radio) ||
	   (widget == d->visual_greyscale_radio) ||
	   (widget == d->visual_color_radio)
	)
	{
	    /* Rerender the preview's image */
	    PreviewUpdateVisual(d->preview);
	    PreviewQueueDraw(d->preview);
	}

	d->freeze_count--;
}

/*
 *	Any visual GtkWidget event signal callback.
 */
static gint PrintDlgVisualEventCB(
	GtkWidget *widget, GdkEvent *event, gpointer data
)
{
	gint status = FALSE;
	PrintDlg *d = PRINT_DLG(data);
	if((widget == NULL) || (event == NULL) || (d == NULL))
	    return(status);

#define SET_STATIC_TIP(_s_)	{	\
 StaticTipSet(				\
  widget,				\
  (_s_),				\
  STATIC_TIP_ALIGN_WIDGET_CENTER,	\
  STATIC_TIP_ALIGN_WIDGET_MIN,		\
  0, 0					\
 );					\
}

	switch((gint)event->type)
	{
	  case GDK_BUTTON_PRESS:
	    if(GTK_IS_RANGE(widget))
	    {
		GtkRange *range = GTK_RANGE(widget);
		GtkAdjustment *adj = gtk_range_get_adjustment(range);
		if(adj != NULL)
		{
		    gchar *s = g_strdup_printf(
			"%.2f",
			adj->value
		    );
		    SET_STATIC_TIP(s);
		    g_free(s);
		}
	    }
	    break;

	  case GDK_BUTTON_RELEASE:
	    SET_STATIC_TIP(NULL);
	    break;

	  case GDK_MOTION_NOTIFY:
	    if(GTK_IS_RANGE(widget))
	    {
		GtkRange *range = GTK_RANGE(widget);
		GtkAdjustment *adj = gtk_range_get_adjustment(range);
		if(adj != NULL)
		{
		    gchar *s = g_strdup_printf(
			"%.2f",
			adj->value
		    );
		    SET_STATIC_TIP(s);
		    g_free(s);
		}
	    }
	    break;

	  case GDK_ENTER_NOTIFY:
	    break;

	  case GDK_LEAVE_NOTIFY:
	    SET_STATIC_TIP(NULL);
	    break;
	}

#undef SET_STATIC_TIP

	return(status);
}

/*
 *	PrintDlg connect to X Print server callback.
 */
static void PrintDlgXPrintServerConnectCB(
	GtkWidget *widget, gpointer data
)
{
	PrintDlg *d = PRINT_DLG(data);
	if(d == NULL)
	    return;

	if(d->freeze_count > 0)
	    return;

	d->freeze_count++;
	PrintDlgSetBusy(d, TRUE);

	PrintDlgConnectToXPrintServer(
	    d,
	    gtk_entry_get_text(
		GTK_ENTRY(d->xprint_server_address_entry)
	    ),
	    TRUE			/* Verbose */
	);

	PrintDlgSetBusy(d, FALSE);
	d->freeze_count--;
}

/*
 *	Printer popup list box "changed" signal callback.
 */
static void PrintDlgPrinterChangedCB(
	GtkWidget *widget, const gint i, gpointer data
)
{
	gint medium_width, medium_height;
	GdkRectangle *geo;
	GtkWidget *pulist;
	IVPrintXPrintPrinter *p;
	PrintDlg *d = PRINT_DLG(data);
	if((widget == NULL) || (d == NULL))
	    return;

	if(d->freeze_count > 0)
	    return;

	d->freeze_count++;
	PrintDlgSetBusy(d, TRUE);

	pulist = PUListBoxGetPUList(widget);
	p = IV_PRINT_XPRINT_PRINTER(PUListGetItemData(
	    pulist,
	    i
	));
	if(p != NULL)
	    PrintDlgUpdatePrinterValues(
		d,
		p->name,
		TRUE				/* Verbose */
	    );

	/* Update the medium size */
	if(PrintDlgGetMediumSize(
	    d,
	    &medium_width, &medium_height
	))
	{
	    GdkRectangle *medium = &d->medium_size;
	    medium->width = medium_width;
	    medium->height = medium_height;
	}
	else
	{
	    GdkRectangle *medium = &d->medium_size;
	    medium->width = PRINT_DLG_DEF_MEDIUM_WIDTH;
	    medium->height = PRINT_DLG_DEF_MEDIUM_HEIGHT;
	}

	/* Reset the output geometry */
	geo = &d->output_img_geometry;
	geo->x = 0;
	geo->y = 0;
	PrintDlgOutputSetWidgets(d);

	PreviewResize(d->preview);
	PreviewUpdateImages(d->preview);
	PreviewQueueDraw(d->preview);

	PrintDlgSetBusy(d, FALSE);
	d->freeze_count--;
}

#if 0
/*
 *	Tray Popup List Box "changed" signal callback.
 */
static void PrintDlgTrayChangedCB(
	GtkWidget *widget, const gint i, gpointer data
)
{
	gint medium_width, medium_height;
	GdkRectangle *geo;
	IVPrintXPrintPrinter *p;
	PrintDlg *d = PRINT_DLG(data);
	if((widget == NULL) || (d == NULL))
	    return;

	if(d->freeze_count > 0)
	    return;

	d->freeze_count++;
	PrintDlgSetBusy(d, TRUE);

	/* Update the medium list */
	p = IV_PRINT_XPRINT_PRINTER(PUListBoxGetSelectedData(
	    d->printer_pulistbox
	));
	if(p != NULL)
	    PrintDlgUpdateMediumsList(
		d,
		p->name,			/* Printer name */
		TRUE				/* Verbose */
	    );

	/* Update the current medium size */
	if(PrintDlgGetMediumSize(
	    d,
	    &medium_width, &medium_height
	))
	{
	    GdkRectangle *medium = &d->medium_size;
	    medium->width = medium_width;
	    medium->height = medium_height;
	}
	else
	{
	    GdkRectangle *medium = &d->medium_size;
	    medium->width = PRINT_DLG_DEF_MEDIUM_WIDTH;
	    medium->height = PRINT_DLG_DEF_MEDIUM_HEIGHT;
	}

	/* Reset the output geometry */
	geo = &d->output_img_geometry;
	geo->x = 0;
	geo->y = 0;
	PrintDlgOutputSetWidgets(d);

	PreviewResize(d->preview);
	PreviewUpdateImages(d->preview);
	PreviewQueueDraw(d->preview);

	PrintDlgSetBusy(d, FALSE);
	d->freeze_count--;
}
#endif

/*
 *	Medium GtkCTree "tree_select_row" signal callback.
 */
static void PrintDlgMediumSelectRowCB(
	GtkCTree *ctree, GtkCTreeNode *node, gint column,
	gpointer data
)
{
	gint medium_width, medium_height;
	GdkRectangle *geo;
	PrintDlg *d = PRINT_DLG(data);
	if((ctree == NULL) || (node == NULL) || (d == NULL))
	    return;

	if(d->freeze_count > 0)
	    return;

	d->freeze_count++;
	PrintDlgSetBusy(d, TRUE);

	if(gtk_ctree_node_is_visible(ctree, node)
	    != GTK_VISIBILITY_FULL
	)
	    gtk_ctree_node_moveto(
		ctree,
		node, -1,
		0.5f, 0.0f			/* Row align, column align */
	    );

	/* Update the current medium size */
	if(PrintDlgGetMediumSize(
	    d,
	    &medium_width, &medium_height
	))
	{
	    GdkRectangle *medium = &d->medium_size;
	    medium->width = medium_width;
	    medium->height = medium_height;
	}
	else
	{
	    GdkRectangle *medium = &d->medium_size;
	    medium->width = PRINT_DLG_DEF_MEDIUM_WIDTH;
	    medium->height = PRINT_DLG_DEF_MEDIUM_HEIGHT;

	    if(TRUE)
	    {
		IVPrintXPrintMedium *m = IV_PRINT_XPRINT_MEDIUM(
		    gtk_ctree_node_get_row_data(ctree, node)
		);
		const gchar *medium_name = (m != NULL) ? m->name : NULL;
		gchar *msg = g_strdup_printf(
"Unable to get the size of the medium \"%s\".",
		    medium_name
		);
		CDialogSetTransientFor(d->toplevel);
		CDialogGetResponse(
		    "Get Medium Size Failed",
		    msg,
		    NULL,
		    CDIALOG_ICON_WARNING,
		    CDIALOG_BTNFLAG_OK,
		    CDIALOG_BTNFLAG_OK
		);
		CDialogSetTransientFor(NULL);
		g_free(msg);
	    }
	}

	/* Reset the output geometry */
	geo = &d->output_img_geometry;
	geo->x = 0;
	geo->y = 0;
	PrintDlgOutputSetWidgets(d);

	PreviewResize(d->preview);
	PreviewUpdateImages(d->preview);
	PreviewQueueDraw(d->preview);

	PrintDlgSetBusy(d, FALSE);
	d->freeze_count--;
}

/*
 *	Orientation popup list box "changed" signal callback.
 */
static void PrintDlgOrientationChangedCB(
	GtkWidget *widget, const gint i, gpointer data
)
{
	gint medium_width, medium_height;
	GdkRectangle *geo;
	PrintDlg *d = PRINT_DLG(data);
	if((widget == NULL) || (d == NULL))
	    return;

	if(d->freeze_count > 0)
	    return;

	d->freeze_count++;
	PrintDlgSetBusy(d, TRUE);

	/* Update the medium size */
	if(PrintDlgGetMediumSize(
	    d,
	    &medium_width, &medium_height
	))
	{
	    GdkRectangle *medium = &d->medium_size;
	    medium->width = medium_width;
	    medium->height = medium_height;
	}
	else
	{
	    GdkRectangle *medium = &d->medium_size;
	    medium->width = PRINT_DLG_DEF_MEDIUM_WIDTH;
	    medium->height = PRINT_DLG_DEF_MEDIUM_HEIGHT;
	}

	/* Reset the output geometry */
	geo = &d->output_img_geometry;
	geo->x = 0;
	geo->y = 0;
	PrintDlgOutputSetWidgets(d);

	PreviewResize(d->preview);
	PreviewUpdateImages(d->preview);
	PreviewQueueDraw(d->preview);

	PrintDlgSetBusy(d, FALSE);
	d->freeze_count--;
}

/*
 *	Resolution popup list box "changed" signal callback.
 */
static void PrintDlgResolutionChangedCB(
	GtkWidget *widget, const gint i, gpointer data
)
{
	gint medium_width, medium_height;
	GdkRectangle *geo;
	PrintDlg *d = PRINT_DLG(data);
	if((widget == NULL) || (d == NULL))
	    return;

	if(d->freeze_count > 0)
	    return;

	d->freeze_count++;
	PrintDlgSetBusy(d, TRUE);

	/* Update the medium size */
	if(PrintDlgGetMediumSize(
	    d,
	    &medium_width, &medium_height
	))
	{
	    GdkRectangle *medium = &d->medium_size;
	    medium->width = medium_width;
	    medium->height = medium_height;
	}
	else
	{
	    GdkRectangle *medium = &d->medium_size;
	    medium->width = PRINT_DLG_DEF_MEDIUM_WIDTH;
	    medium->height = PRINT_DLG_DEF_MEDIUM_HEIGHT;
	}

	/* Reset the output geometry */
	geo = &d->output_img_geometry;
	geo->x = 0;
	geo->y = 0;
	PrintDlgOutputSetWidgets(d);

	PreviewResize(d->preview);
	PreviewUpdateImages(d->preview);
	PreviewQueueDraw(d->preview);

	PrintDlgSetBusy(d, FALSE);
	d->freeze_count--;
}

/*
 *	PrintDlg Output GtkWidget "changed" or "toggled" signal
 *	callback.
 */
static void PrintDlgOutputChangedCB(GtkWidget *widget, gpointer data)
{
	GdkRectangle *output;
	PrintDlg *d = PRINT_DLG(data);
	if((widget == NULL) || (d == NULL))
	    return;

	if(d->freeze_count > 0)
	    return;

	d->freeze_count++;

	output = &d->output_img_geometry;

	if(widget == d->output_x_spin)
	{
	    output->x = GTK_ENTRY_GET_VALUEI(widget);
	    PreviewQueueDraw(d->preview);
	}
	else if(widget == d->output_y_spin)
	{
	    output->y = GTK_ENTRY_GET_VALUEI(widget);
	    PreviewQueueDraw(d->preview);
	}
	else if(widget == d->output_width_spin)
	{
	    output->width = GTK_ENTRY_GET_VALUEI(widget);
	    if(output->width > 0)
	    {
		/* Maintain aspect by updating the height? */
		if(d->output_maintain_aspect)
		{ 
		    const gfloat aspect = PrintDlgGetSourceImageAspect(d);
		    if(aspect > 0.0f)
		    {
			output->height = output->width / aspect;
			gtk_spin_button_set_value(
			    GTK_SPIN_BUTTON(d->output_height_spin),
			    (gfloat)output->height
			);
		    }
		}
		PreviewUpdateImages(d->preview);
		PreviewQueueDraw(d->preview);
	    }
	}
	else if(widget == d->output_height_spin)
	{
	    output->height = GTK_ENTRY_GET_VALUEI(widget);
	    if(output->height > 0)
	    {
		/* Maintain aspect by updating the width? */
		if(d->output_maintain_aspect)
		{
		    const gfloat aspect = PrintDlgGetSourceImageAspect(d);
		    if(aspect > 0.0f)
		    {
			output->width = output->height * aspect;
			gtk_spin_button_set_value(
			    GTK_SPIN_BUTTON(d->output_width_spin),
			    (gfloat)output->width
			);
		    }
		}
		PreviewUpdateImages(d->preview);
		PreviewQueueDraw(d->preview);
	    }
	}
	else if(widget == d->output_frame_spin)
	{
	    d->src_frame_num = GTK_ENTRY_GET_VALUEI(widget) - 1;
	    PreviewUpdateImages(d->preview);
	    PreviewQueueDraw(d->preview);
	}
	else if(widget == d->output_maintain_aspect_check)
	{
	    d->output_maintain_aspect = GTK_TOGGLE_BUTTON_GET_ACTIVE(widget);
	    if(d->output_maintain_aspect)
	    {
		const gfloat aspect = PrintDlgGetSourceImageAspect(d);
		if(aspect > 0.0f)
		{
		    /* Update the width and height values */
		    output->height = (gint)((gfloat)output->width / aspect);
		    gtk_spin_button_set_value(
			GTK_SPIN_BUTTON(d->output_width_spin),
			(gfloat)output->width
		    );
		    gtk_spin_button_set_value(
			GTK_SPIN_BUTTON(d->output_height_spin),
			(gfloat)output->height
		    );
		    PreviewUpdateImages(d->preview);
		    PreviewQueueDraw(d->preview);
		}
	    }
	}

	d->freeze_count--;
}

/*
 *	Print callback.
 */
static void PrintDlgPrintCB(GtkWidget *widget, gpointer data)
{
	PrintDlg *d = PRINT_DLG(data);
	if(d == NULL)
	    return;

	if(d->freeze_count > 0)
	    return;

	/* Set the user's response to OK */
	d->flags |= PRINT_DLG_RESPONSE_OK;

	if(d->main_level > 0)
	{
	    d->main_level--;
	    gtk_main_quit();
	}
}

/*
 *	Preview callback.
 */
static void PrintDlgPreviewCB(GtkWidget *widget, gpointer data)
{
	PrintPreviewDlg *ppd;
	PrintDlg *d = PRINT_DLG(data);
	if(d == NULL)
	    return;

	if(d->freeze_count > 0)
	    return;

	ppd = d->print_preview_dlg;
	if(ppd == NULL)
	    d->print_preview_dlg = ppd = PrintPreviewDlgNew(d);

	PreviewResize(ppd->preview);

	/* Need to recreate the preview image because a "configure_event"
	 * is not always generated in this case since the size of the
	 * PrintPreviewDlg may not have changed
	 */
	PreviewUpdateImages(ppd->preview);

	PrintPreviewDlgUpdate(ppd);
	PrintPreviewDlgMap(ppd);
}

/*
 *	Cancel callback.
 */
static void PrintDlgCancelCB(GtkWidget *widget, gpointer data)
{
	PrintDlg *d = PRINT_DLG(data);
	if(d == NULL)
	    return;

	if(d->freeze_count > 0)
	    return;

	/* Set the user's response to cancel */
	d->flags &= ~PRINT_DLG_RESPONSE_OK;

	if(d->main_level > 0)
	{
	    d->main_level--;
	    gtk_main_quit();
	}
}


/*
 *	Creates a new PrintDlg.
 */
static PrintDlg *PrintDlgNew(
	const gchar *path,
	GtkWidget *ref_toplevel
)
{
	const gint	bw = GUI_BUTTON_HLABEL_WIDTH_DEF,
			bh = GUI_BUTTON_HLABEL_HEIGHT_DEF,
			border_major = 5,
			border_minor = 2;
	gchar *s;
	GSList *gslist;
	GdkWindow *window;
	GtkAccelGroup *accelgrp;
	GtkAdjustment *adj;
	GtkStyle *style;
	GtkWidget	*w,
			*parent, *parent2, *parent3, *parent4,
			*parent5, *parent6,
			*toplevel,
			*main_vbox,
			*scroll_parent;
	GtkCList *clist;
	GtkCTree *ctree;
	PrintDlg *d = PRINT_DLG(
	    g_malloc0(sizeof(PrintDlg))
	);
	if(d == NULL)
	    return(NULL);

	gtk_widget_push_visual(gdk_rgb_get_visual());
	gtk_widget_push_colormap(gdk_rgb_get_cmap());
	d->toplevel = toplevel = gtk_window_new(GTK_WINDOW_DIALOG);
	gtk_widget_pop_visual();
	gtk_widget_pop_colormap();
	d->accelgrp = accelgrp = gtk_accel_group_new();
	d->ref_toplevel = ref_toplevel;
#if 0
	d->freeze_count = 0;
	d->busy_count = 0;
	d->main_level = 0;
	d->flags = 0;
#endif
	d->path = STRDUP(path);
#if 0
	d->src_img = NULL;
	d->preview_img = NULL;
	d->orig_preview_img = NULL;
#endif
	d->freeze_count++;

	/* Toplevel GtkWindow */
	w = toplevel;
#ifdef PROG_NAME
	gtk_window_set_wmclass(
	    GTK_WINDOW(w), "dialog", PROG_NAME
	);
#endif
	gtk_window_set_policy(GTK_WINDOW(w), FALSE, FALSE, FALSE);
	s = g_strdup_printf(
	    PRINT_DLG_TITLE ": %s",
	    STRISEMPTY(path) ? "Untitled" : path
	);
	gtk_window_set_title(GTK_WINDOW(w), s);
	g_free(s);
	gtk_widget_set_usize(
	    w,
	    PRINT_DLG_WIDTH, PRINT_DLG_HEIGHT
	);
	gtk_widget_add_events(
	    w,
	    GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "delete_event",
	    GTK_SIGNAL_FUNC(PrintDlgDeleteEventCB), d
	);
	gtk_widget_realize(w);
	d->flags |= PRINT_DLG_REALIZED;
	window = w->window;
	style = gtk_widget_get_style(w);
	if(window != NULL)
	{
	    gdk_window_set_decorations(
		window,
		GDK_DECOR_BORDER | GDK_DECOR_TITLE
	    );
	    gdk_window_set_functions(
		window,
		GDK_FUNC_MOVE | GDK_FUNC_CLOSE
	    );
	    GUISetWMIcon(
		window,
		(guint8 **)icon_print2_48x48_xpm
	    );
	}
	gtk_accel_group_attach(accelgrp, GTK_OBJECT(w));
	parent = w;

	d->tray_closed_icon_pm = gdk_pixmap_create_from_xpm_d(
	    window,
	    &d->tray_closed_icon_mask,
	    NULL,
	    (gchar **)icon_printer_tray_closed_20x20_xpm
	);
	d->tray_opened_icon_pm = gdk_pixmap_create_from_xpm_d(
	    window,
	    &d->tray_opened_icon_mask,
	    NULL,
	    (gchar **)icon_printer_tray_opened_20x20_xpm
	);
	d->medium_icon_pm = gdk_pixmap_create_from_xpm_d(
	    window,
	    &d->medium_icon_mask,
	    NULL,
	    (gchar **)icon_file_20x20_xpm
	);



	/* Main GtkVBox */
	main_vbox = w = gtk_vbox_new(FALSE, 0);
	gtk_container_add(GTK_CONTAINER(parent), w);
	gtk_widget_show(w);
	parent = w;

	/* GtkVBox */
	w = gtk_vbox_new(FALSE, border_major);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_box_pack_start(GTK_BOX(main_vbox), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
	parent = w;


	/* Print To GtkFrame */
	w = gtk_frame_new("Print To");
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent2 = w;

	w = gtk_vbox_new(FALSE, border_major);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_container_add(GTK_CONTAINER(parent2), w);
	gtk_widget_show(w);
	parent2 = w;

	/* X Print Server Address */
	w = gtk_hbox_new(FALSE, border_minor);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent3 = w;

	w = gtk_label_new("X Print Server Address:");
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_show(w);

	d->xprint_server_address_entry = w = gtk_entry_new();
	gtk_box_pack_start(GTK_BOX(parent3), w, TRUE, TRUE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "activate",
	    GTK_SIGNAL_FUNC(PrintDlgXPrintServerConnectCB), d
	);
	GUIEditableEndowPopupMenu(w, 0);
	GUISetWidgetTip(
	    w,
"Enter the address to the X Print Server in the format\
 \"hostname:displaynumber\" or just \":displaynumber\" and\
 then press ENTER"
	);
	gtk_widget_show(w);

	/* Printer */
	w = gtk_hbox_new(FALSE, border_minor);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent3 = w;

	w = gtk_label_new("Printer:");
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_show(w);

	d->printer_pulistbox = w = PUListBoxNew(
	    -1, -1
	);
	gtk_box_pack_start(GTK_BOX(parent3), w, TRUE, TRUE, 0);
	PUListBoxSetChangedCB(
	    w,
	    PrintDlgPrinterChangedCB, d
	);
	gtk_widget_show(w);


	/* GtkHBox to divide into two columns */
	w = gtk_hbox_new(FALSE, border_major);
	gtk_box_pack_start(GTK_BOX(parent), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
	parent2 = w;


	/* Left column GtkVBox */
	w = gtk_vbox_new(FALSE, border_major);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent3 = w;


	/* Visual GtkFrame */
	w = gtk_frame_new("Visual");
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent4 = w;

	/* Visual GtkVBox */
	w = gtk_vbox_new(FALSE, border_major);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_container_add(GTK_CONTAINER(parent4), w);
	gtk_widget_show(w);
	parent4 = w;

	/* Black & White, Greyscale, and Color GtkHBox */
	w = gtk_hbox_new(FALSE, border_minor);
	gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent5 = w;
	/* Black & White GtkRadioButton */
	gslist = NULL;
	d->visual_bw_radio = w =
	    gtk_radio_button_new_with_label(gslist, "B&W");
	gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "toggled",
	    GTK_SIGNAL_FUNC(PrintDlgVisualToggledCB), d
	);
	gtk_widget_show(w);
	gslist = gtk_radio_button_group(GTK_RADIO_BUTTON(w));
	/* Greyscale GtkRadioButton */
	d->visual_greyscale_radio = w =
	    gtk_radio_button_new_with_label(gslist, "GreyScale");
	gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "toggled",
	    GTK_SIGNAL_FUNC(PrintDlgVisualToggledCB), d
	);
	gtk_widget_show(w);
	gslist = gtk_radio_button_group(GTK_RADIO_BUTTON(w));
	/* Color GtkRadioButton */
	d->visual_color_radio = w =
	    gtk_radio_button_new_with_label(gslist, "Color");
	gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "toggled",
	    GTK_SIGNAL_FUNC(PrintDlgVisualToggledCB), d
	);
	gtk_widget_show(w);

	/* Brightness GtkVBox */
	w = gtk_vbox_new(FALSE, border_minor);
	gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent5 = w;
	/* Hbox for ticks */
	w = gtk_hbox_new(FALSE, 0);
	gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent6 = w;
	w = gtk_pixmap_new_from_xpm_d(
	    window, NULL,
	    (guint8 **)icon_minus_16x16_xpm
	);
	gtk_widget_set_usize(w, 16, 16);
	gtk_box_pack_start(GTK_BOX(parent6), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	w = gtk_pixmap_new_from_xpm_d(
	    window, NULL,
	    (guint8 **)icon_brightness_16x16_xpm
	);
	gtk_widget_set_usize(w, 16, 16);
	gtk_box_pack_start(GTK_BOX(parent6), w, TRUE, FALSE, 0);
	gtk_widget_show(w);
	w = gtk_pixmap_new_from_xpm_d(
	    window, NULL,
	    (guint8 **)icon_plus_16x16_xpm
	);
	gtk_widget_set_usize(w, 16, 16);
	gtk_box_pack_start(GTK_BOX(parent6), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	/* Brightness GtkAdjustment */
	adj = (GtkAdjustment *)gtk_adjustment_new(
	    0.0f,		/* Initial value */
	    -1.0f,		/* Minimum */
	    1.0f,		/* Maximum */
	    0.1f,		/* Step inc */
	    0.1f,		/* Page inc */
	    0.0f		/* Page size */
	);
	d->brightness_adj = adj = (GtkAdjustment *)GTK_OBJECT_REF(GTK_OBJECT(adj));
	gtk_signal_connect(
	    GTK_OBJECT(adj), "value_changed",
	    GTK_SIGNAL_FUNC(PrintDlgVisualAdjustmentValueChangedCB), d
	);
	/* Brightness GtkHScale */
	d->brightness_scale = w = gtk_hscale_new(adj);
#if 0
	gtk_widget_add_events(
	    w,
	    GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK |
	    GDK_POINTER_MOTION_MASK | GDK_ENTER_NOTIFY_MASK |
	    GDK_LEAVE_NOTIFY_MASK
	);
#endif
	gtk_signal_connect(
	    GTK_OBJECT(w), "button_press_event",
	    GTK_SIGNAL_FUNC(PrintDlgVisualEventCB), d
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "button_release_event",
	    GTK_SIGNAL_FUNC(PrintDlgVisualEventCB), d
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "motion_notify_event",
	    GTK_SIGNAL_FUNC(PrintDlgVisualEventCB), d
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "enter_notify_event",
	    GTK_SIGNAL_FUNC(PrintDlgVisualEventCB), d
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "leave_notify_event",
	    GTK_SIGNAL_FUNC(PrintDlgVisualEventCB), d
	);
	gtk_scale_set_draw_value(GTK_SCALE(w), FALSE);
	gtk_scale_set_value_pos(GTK_SCALE(w), GTK_POS_RIGHT);
	gtk_scale_set_digits(GTK_SCALE(w), 2);
	gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
	GUISetWidgetTip(w, "Brightness");
	gtk_widget_show(w);

	/* Contrast GtkVBox */
	w = gtk_vbox_new(FALSE, border_minor);
	gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent5 = w;
	/* Hbox for ticks */
	w = gtk_hbox_new(FALSE, 0);
	gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent6 = w;
	w = gtk_pixmap_new_from_xpm_d(
	    window, NULL,
	    (guint8 **)icon_minus_16x16_xpm
	);
	gtk_widget_set_usize(w, 16, 16);
	gtk_box_pack_start(GTK_BOX(parent6), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	w = gtk_pixmap_new_from_xpm_d(
	    window, NULL,
	    (guint8 **)icon_contrast_16x16_xpm
	);
	gtk_widget_set_usize(w, 16, 16);
	gtk_box_pack_start(GTK_BOX(parent6), w, TRUE, FALSE, 0);
	gtk_widget_show(w);
	w = gtk_pixmap_new_from_xpm_d(
	    window, NULL,
	    (guint8 **)icon_plus_16x16_xpm
	);
	gtk_widget_set_usize(w, 16, 16);
	gtk_box_pack_start(GTK_BOX(parent6), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	/* Contrast GtkAdjustment */
	adj = (GtkAdjustment *)gtk_adjustment_new(
	    0.0f,		/* Initial value */
	    -1.0f,		/* Minimum */
	    1.0f,		/* Maximum */
	    0.1f,		/* Step inc */
	    0.1f,		/* Page inc */
	    0.0f		/* Page size */
	);
	d->contrast_adj = adj = (GtkAdjustment *)GTK_OBJECT_REF(GTK_OBJECT(adj));
	gtk_signal_connect(
	    GTK_OBJECT(adj), "value_changed",
	    GTK_SIGNAL_FUNC(PrintDlgVisualAdjustmentValueChangedCB), d
	);
	/* Contrast GtkHScale */
	d->contrast_scale = w = gtk_hscale_new(adj);
#if 0
	gtk_widget_add_events(
	    w,
	    GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK |
	    GDK_POINTER_MOTION_MASK | GDK_ENTER_NOTIFY_MASK |
	    GDK_LEAVE_NOTIFY_MASK
	);
#endif
	gtk_signal_connect(
	    GTK_OBJECT(w), "button_press_event",
	    GTK_SIGNAL_FUNC(PrintDlgVisualEventCB), d
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "button_release_event",
	    GTK_SIGNAL_FUNC(PrintDlgVisualEventCB), d
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "motion_notify_event",
	    GTK_SIGNAL_FUNC(PrintDlgVisualEventCB), d
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "enter_notify_event",
	    GTK_SIGNAL_FUNC(PrintDlgVisualEventCB), d
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "leave_notify_event",
	    GTK_SIGNAL_FUNC(PrintDlgVisualEventCB), d
	);
	gtk_scale_set_draw_value(GTK_SCALE(w), FALSE);
	gtk_scale_set_value_pos(GTK_SCALE(w), GTK_POS_RIGHT);
	gtk_scale_set_digits(GTK_SCALE(w), 2);
	gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
	GUISetWidgetTip(w, "Contrast");
	gtk_widget_show(w);


	/* Medium GtkFrame */
	w = gtk_frame_new("Medium");
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
	gtk_box_pack_start(GTK_BOX(parent3), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
	parent4 = w;

	w = gtk_vbox_new(FALSE, border_major);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_container_add(GTK_CONTAINER(parent4), w);
	gtk_widget_show(w);
	parent4 = w;


	/* Medium GtkScrolledWindow */
	w = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(
	    GTK_SCROLLED_WINDOW(w),
	    GTK_POLICY_AUTOMATIC,
	    GTK_POLICY_AUTOMATIC
	);
	gtk_box_pack_start(GTK_BOX(parent4), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
	scroll_parent = w;

	/* Medium GtkCTree */
	d->medium_ctree = w = gtk_ctree_new(1, 0);
	clist = GTK_CLIST(w);
	ctree = GTK_CTREE(w);
	gtk_widget_set_usize(w, 250, -1);
	gtk_container_add(GTK_CONTAINER(scroll_parent), w);
	gtk_widget_realize(w);
	gtk_clist_column_titles_hide(clist);
	gtk_clist_column_titles_passive(clist);
	gtk_clist_set_selection_mode(clist, GTK_SELECTION_BROWSE);
#ifdef IV_CLIST_DEF_ROW_SPACING
	gtk_clist_set_row_height(clist, IV_CLIST_DEF_ROW_SPACING);
#else
	gtk_clist_set_row_height(clist, 20);
#endif
	gtk_clist_set_column_justification(
	    clist, 0, GTK_JUSTIFY_LEFT
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "tree_select_row",
	    GTK_SIGNAL_FUNC(PrintDlgMediumSelectRowCB), d
	);
	gtk_widget_show(w);


	/* Orientation GtkHBox */
	w = gtk_hbox_new(FALSE, border_minor);
	gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent5 = w;

	w = gtk_label_new("Orientation:");
	gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
	gtk_widget_show(w);

	d->orientation_pulistbox = w = PUListBoxNew(
	    -1, -1
	);
	gtk_box_pack_start(GTK_BOX(parent5), w, TRUE, TRUE, 0);
	PUListBoxSetChangedCB(
	    w,
	    PrintDlgOrientationChangedCB, d
	);
	gtk_widget_show(w);


	/* Resolution GtkHBox */
	w = gtk_hbox_new(FALSE, border_minor);
	gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent5 = w;

	w = gtk_label_new("Resolution:");
	gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
	gtk_widget_show(w);

	d->resolution_pulistbox = w = PUListBoxNew(
	    -1, -1
	);
	gtk_box_pack_start(GTK_BOX(parent5), w, TRUE, TRUE, 0);
	PUListBoxSetChangedCB(
	    w,
	    PrintDlgResolutionChangedCB, d
	);
	gtk_widget_show(w);


	/* Right column GtkVBox */
	w = gtk_vbox_new(FALSE, border_major);
	gtk_box_pack_start(GTK_BOX(parent2), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
	parent3 = w;


	/* Output geometry */
	w = gtk_frame_new("Output");
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
	gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent4 = w;
	/* Vbox */
	w = gtk_vbox_new(FALSE, border_major);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_container_add(GTK_CONTAINER(parent4), w);
	gtk_widget_show(w);
	parent4 = w;
	/* Hbox */
	w = gtk_hbox_new(FALSE, border_major);
	gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent5 = w;
	/* Output X */
	w = gtk_hbox_new(FALSE, border_minor);
	gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent6 = w;
	w = gtk_label_new("X:");
	gtk_box_pack_start(GTK_BOX(parent6), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	adj = (GtkAdjustment *)gtk_adjustment_new(
	    0.0f,
	    -(gfloat)((guint16)-1 / 2),
	    (gfloat)((guint16)-1 / 2),
	    1.0f,
	    10.0f,
	    0.0f
	);
	d->output_x_spin = w = gtk_spin_button_new(
	    adj, 1.0, 0
	);
	gtk_widget_set_usize(w, 65, -1);
	gtk_box_pack_start(GTK_BOX(parent6), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "changed",
	    GTK_SIGNAL_FUNC(PrintDlgOutputChangedCB), d
	);
	GUIEditableEndowPopupMenu(w, 0);
	gtk_widget_show(w);
	/* Output Y */
	w = gtk_hbox_new(FALSE, border_minor);
	gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent6 = w;
	w = gtk_label_new("Y:");
	gtk_box_pack_start(GTK_BOX(parent6), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	adj = (GtkAdjustment *)gtk_adjustment_new(
	    0.0f,
	    -(gfloat)((guint16)-1 / 2),
	    (gfloat)((guint16)-1 / 2),
	    1.0f,
	    10.0f,
	    0.0f  
	);
	d->output_y_spin = w = gtk_spin_button_new(
	    adj, 1.0, 0
	);
	gtk_widget_set_usize(w, 65, -1);
	gtk_box_pack_start(GTK_BOX(parent6), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "changed",
	    GTK_SIGNAL_FUNC(PrintDlgOutputChangedCB), d
	);
	GUIEditableEndowPopupMenu(w, 0);
	gtk_widget_show(w);
	/* Output Width */
	w = gtk_hbox_new(FALSE, border_minor);
	gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent6 = w;
	w = gtk_label_new("W:");
	gtk_box_pack_start(GTK_BOX(parent6), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	adj = (GtkAdjustment *)gtk_adjustment_new(
	    0.0f,
	    0.0f,
	    (gfloat)((guint16)-1),
	    1.0f,
	    10.0f,
	    0.0f  
	);
	d->output_width_spin = w = gtk_spin_button_new(
	    adj, 1.0, 0
	);
	gtk_widget_set_usize(w, 65, -1);
	gtk_box_pack_start(GTK_BOX(parent6), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "changed",
	    GTK_SIGNAL_FUNC(PrintDlgOutputChangedCB), d
	);
	GUIEditableEndowPopupMenu(w, 0);
	gtk_widget_show(w);
	/* Output Height */
	w = gtk_hbox_new(FALSE, border_minor);
	gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent6 = w;
	w = gtk_label_new("H:");
	gtk_box_pack_start(GTK_BOX(parent6), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	adj = (GtkAdjustment *)gtk_adjustment_new(
	    0.0f,
	    0.0f,
	    (gfloat)((guint16)-1),
	    1.0f,
	    10.0f,
	    0.0f  
	);
	d->output_height_spin = w = gtk_spin_button_new(
	    adj, 1.0, 0
	);
	gtk_widget_set_usize(w, 65, -1);
	gtk_box_pack_start(GTK_BOX(parent6), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "changed",
	    GTK_SIGNAL_FUNC(PrintDlgOutputChangedCB), d
	);
	GUIEditableEndowPopupMenu(w, 0);
	gtk_widget_show(w);

	/* Hbox */
	w = gtk_hbox_new(FALSE, border_major);
	gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent5 = w;

	/* Output Frame */
	w = gtk_hbox_new(FALSE, border_minor);
	gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent6 = w;
	w = gtk_label_new("Frame:");
	gtk_box_pack_start(GTK_BOX(parent6), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	adj = (GtkAdjustment *)gtk_adjustment_new(
	    1.0f,
	    1.0f,
	    (gfloat)((guint32)-1),
	    1.0f,
	    5.0f,
	    0.0f  
	);
	d->output_frame_spin = w = gtk_spin_button_new(
	    adj, 1.0, 0
	);
	gtk_widget_set_usize(w, 50, -1);
	gtk_box_pack_start(GTK_BOX(parent6), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "changed",
	    GTK_SIGNAL_FUNC(PrintDlgOutputChangedCB), d
	);
	GUIEditableEndowPopupMenu(w, 0);
	gtk_widget_show(w);

	/* Maintain Aspect */
	d->output_maintain_aspect_check = w = gtk_check_button_new_with_label(
	    "Maintain Aspect"
	);
	gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "toggled",
	    GTK_SIGNAL_FUNC(PrintDlgOutputChangedCB), d
	);
	gtk_widget_show(w);

	/* Output Number Of Coppies */
	w = gtk_hbox_new(FALSE, border_minor);
	gtk_box_pack_end(GTK_BOX(parent5), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent6 = w;
	w = gtk_label_new("Coppies:");
	gtk_box_pack_start(GTK_BOX(parent6), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	adj = (GtkAdjustment *)gtk_adjustment_new(
	    1.0f,
	    1.0f,
	    (gfloat)((guint32)-1),
	    1.0f,
	    5.0f,
	    0.0f  
	);
	d->output_ncoppies_spin = w = gtk_spin_button_new(
	    adj,
	    1.0,
	    0
	);
	gtk_widget_set_usize(w, 50, -1);
	gtk_box_pack_start(GTK_BOX(parent6), w, FALSE, FALSE, 0);
	GUIEditableEndowPopupMenu(w, 0);
	gtk_widget_show(w);


	/* Preview GtkFrame */
	w = gtk_frame_new("Preview");
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
	gtk_box_pack_start(GTK_BOX(parent3), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
	parent4 = w;

	/* Preview */
	d->preview = PreviewNew(
	    parent4,
	    style,
	    NULL,				/* No background color */
	    TRUE,				/* Interactive */
	    d
	);


	parent = main_vbox;

	/* GtkSeparator */
	w = gtk_hseparator_new();
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	gtk_widget_show(w);


	/* Buttons GtkHBox */
	w = gtk_hbox_new(TRUE, 0);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, border_major);
	gtk_widget_show(w);
	parent2 = w;

	/* Print GtkButton */
	d->print_btn = w = GUIButtonPixmapLabelH(
	    (guint8 **)icon_print2_20x20_xpm,
	    "Print",
	    NULL
	);
	gtk_widget_set_usize(w, bw, bh);
	GTK_WIDGET_SET_FLAGS(w, GTK_CAN_DEFAULT);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "clicked",
	    GTK_SIGNAL_FUNC(PrintDlgPrintCB), d
	);
	gtk_accel_group_add(
	    accelgrp,
	    GDK_p, GDK_CONTROL_MASK,
	    GTK_ACCEL_VISIBLE,
	    GTK_OBJECT(w),
	    "clicked"
	);
	GUIButtonLabelUnderline(w, GDK_p);
	gtk_widget_show(w);

	/* Preview GtkButton */
	d->preview_btn = w = GUIButtonPixmapLabelH(
	    (guint8 **)icon_print_preview_20x20_xpm,
	    "Preview",
	    NULL
	);
	gtk_widget_set_usize(w, bw, bh);
	GTK_WIDGET_SET_FLAGS(w, GTK_CAN_DEFAULT);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "clicked",
	    GTK_SIGNAL_FUNC(PrintDlgPreviewCB), d
	);
	gtk_accel_group_add(
	    accelgrp,
	    GDK_v, GDK_CONTROL_MASK,
	    GTK_ACCEL_VISIBLE,
	    GTK_OBJECT(w),
	    "clicked"
	);
	GUIButtonLabelUnderline(w, GDK_v);
	gtk_widget_show(w);

	/* Cancel GtkButton */
	d->cancel_btn = w = GUIButtonPixmapLabelH(
	    (guint8 **)icon_cancel_20x20_xpm,
#if defined(PROG_LANGUAGE_SPANISH)
	    "Cancele", 
#elif defined(PROG_LANGUAGE_FRENCH)
	    "Annuler",
#elif defined(PROG_LANGUAGE_GERMAN)
	    "Heben",
#elif defined(PROG_LANGUAGE_ITALIAN)
	    "Annullar",
#elif defined(PROG_LANGUAGE_DUTCH)
	    "Annuler",
#elif defined(PROG_LANGUAGE_PORTUGUESE)
	    "Cancela",
#elif defined(PROG_LANGUAGE_NORWEGIAN)
	    "Kanseler",
#else
	    "Cancel",
#endif
	    NULL
	);
	gtk_widget_set_usize(w, bw, bh);
	GTK_WIDGET_SET_FLAGS(w, GTK_CAN_DEFAULT);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "clicked",
	    GTK_SIGNAL_FUNC(PrintDlgCancelCB), d
	);
	gtk_accel_group_add(
	    accelgrp,
	    GDK_Escape, 0,
	    GTK_ACCEL_VISIBLE,
	    GTK_OBJECT(w),
	    "clicked"
	);
	gtk_accel_group_add(
	    accelgrp,
	    GDK_c, GDK_CONTROL_MASK,
	    GTK_ACCEL_VISIBLE,
	    GTK_OBJECT(w),
	    "clicked"                             
	);
	GUIButtonLabelUnderline(w, GDK_c);
	gtk_widget_show(w);


	d->freeze_count--;


	return(d);
}

/*
 *	Sets the PrintDlg busy or ready.
 */
static void PrintDlgSetBusy(
	PrintDlg *d,
	const gboolean busy
)
{
	GdkCursor *cur;
	GtkWidget *w = d->toplevel;
	if(busy)
	{
	    d->busy_count++;
	    if(d->busy_count > 1)
		return;
	    cur = d->busy_cur;
	}
	else
	{   
	    d->busy_count--;
	    if(d->busy_count < 0)
		d->busy_count = 0;
	    if(d->busy_count > 0)
		return;
	    cur = NULL;			/* Use default cursor */
	}
	if(w->window != NULL)
	{
	    gdk_window_set_cursor(w->window, cur);
	    gdk_flush();
	}
}

/*
 *	Updates the PrintDlg widgets to reflect current values.
 */
static void PrintDlgUpdate(PrintDlg *d)
{
	gboolean sensitive;
	GtkWidget	*pulistbox,
			*pulist;

	pulistbox = d->printer_pulistbox;
	pulist = PUListBoxGetPUList(pulistbox);
	sensitive = (PUListGetTotalItems(pulist) > 0) ? TRUE : FALSE;

	gtk_widget_set_sensitive(d->printer_pulistbox, sensitive);
	gtk_widget_set_sensitive(d->visual_bw_radio, sensitive);
	gtk_widget_set_sensitive(d->visual_greyscale_radio, sensitive);
	gtk_widget_set_sensitive(d->visual_color_radio, sensitive);
	gtk_widget_set_sensitive(d->brightness_scale, sensitive);
	gtk_widget_set_sensitive(d->contrast_scale, sensitive);
	gtk_widget_set_sensitive(d->medium_ctree, sensitive);
	gtk_widget_set_sensitive(d->orientation_pulistbox, sensitive);
	gtk_widget_set_sensitive(d->resolution_pulistbox, sensitive);
	gtk_widget_set_sensitive(d->output_x_spin, sensitive);
	gtk_widget_set_sensitive(d->output_y_spin, sensitive);
	gtk_widget_set_sensitive(d->output_width_spin, sensitive);
	gtk_widget_set_sensitive(d->output_height_spin, sensitive);
	gtk_widget_set_sensitive(d->output_frame_spin, sensitive);
	gtk_widget_set_sensitive(d->output_maintain_aspect_check, sensitive);
	gtk_widget_set_sensitive(d->output_ncoppies_spin, sensitive);
	if(d->preview != NULL)
	    gtk_widget_set_sensitive(d->preview->toplevel, sensitive);
	gtk_widget_set_sensitive(d->print_btn, sensitive);
	gtk_widget_set_sensitive(d->preview_btn, sensitive);
}

/*
 *	Maps the PrintDlg.
 */
static void PrintDlgMap(PrintDlg *d)
{
	GtkWidget *w = d->print_btn;
	gtk_widget_grab_focus(w);
	gtk_widget_grab_default(w);
	w = d->toplevel;
	gtk_window_set_modal(GTK_WINDOW(w), TRUE);
	gtk_window_set_transient_for(
	    GTK_WINDOW(w),
	    GTK_WINDOW(d->ref_toplevel)
	);
	gtk_widget_show(d->toplevel);
	d->flags |= PRINT_DLG_MAPPED;
}

/*
 *	Unmaps the PrintDlg.
 */
static void PrintDlgUnmap(PrintDlg *d)
{
	GtkWidget *w = d->toplevel;
	gtk_widget_hide(d->toplevel);
	d->flags &= ~PRINT_DLG_MAPPED;
	gtk_window_set_modal(GTK_WINDOW(w), FALSE);
	gtk_window_set_transient_for(GTK_WINDOW(w), NULL);
}

/*
 *	Deletes the PrintDlg.
 */
static void PrintDlgDelete(PrintDlg *d)
{
	if(d == NULL)
	    return;

	/* Delete the Print Preview Dialog */
	PrintPreviewDlgDelete(d->print_preview_dlg);
	d->print_preview_dlg = NULL;

	/* Unmap the Print Dialog */
	PrintDlgUnmap(d);

	d->src_img = NULL;

	/* Delete the existing medium nodes */
	gtk_clist_clear(GTK_CLIST(d->medium_ctree));

	d->freeze_count++;

	/* Destroy all the GtkWidgets */
	GTK_WIDGET_DESTROY(d->menu);
	PreviewDelete(d->preview);
	GTK_WIDGET_DESTROY(d->toplevel);

	GTK_ACCEL_GROUP_UNREF(d->accelgrp);

	(void)GTK_OBJECT_UNREF(GTK_OBJECT(d->brightness_adj));
	(void)GTK_OBJECT_UNREF(GTK_OBJECT(d->contrast_adj));

	(void)GDK_PIXMAP_UNREF(d->tray_closed_icon_pm);
	(void)GDK_BITMAP_UNREF(d->tray_closed_icon_mask);
	(void)GDK_PIXMAP_UNREF(d->tray_opened_icon_pm);
	(void)GDK_BITMAP_UNREF(d->tray_opened_icon_mask);
	(void)GDK_PIXMAP_UNREF(d->medium_icon_pm);
	(void)GDK_BITMAP_UNREF(d->medium_icon_mask);

	GDK_CURSOR_DESTROY(d->busy_cur);

	g_free(d->path);

	d->freeze_count--;

	g_free(d);
}


/*
 *	PrintPreviewDlg toplevel GtkWindow "delete_event" signal
 *	callback.
 */
static gint PrintPreviewDlgDeleteEventCB(
	GtkWidget *widget, GdkEvent *event, gpointer data
)
{
	PrintPreviewDlg *d = PRINT_PREVIEW_DLG(data);
	if(d == NULL)
	    return(FALSE);

	if(d->freeze_count > 0)
	    return(FALSE);

	PrintPreviewDlgUnmap(d);

	return(TRUE);
}


/*
 *	PrintPreviewDlg close callback.
 */
static void PrintPreviewDlgPrintCB(GtkWidget *widget, gpointer data)
{
	PrintPreviewDlg *d = PRINT_PREVIEW_DLG(data);
	if(d == NULL)
	    return;

	if(d->freeze_count > 0)
	    return;

	PrintPreviewDlgUnmap(d);

	/* Call the Print Dialog's Print callback */
	PrintDlgPrintCB(
	    NULL,
	    d->print_dlg
	);
}

/*
 *	PrintPreviewDlg close callback.
 */
static void PrintPreviewDlgCloseCB(GtkWidget *widget, gpointer data)
{
	PrintPreviewDlg *d = PRINT_PREVIEW_DLG(data);
	if(d == NULL)
	    return;

	if(d->freeze_count > 0)
	    return;

	PrintPreviewDlgUnmap(d);
}


/*
 *	Creates a new PrintPreviewDlg.
 */
static PrintPreviewDlg *PrintPreviewDlgNew(PrintDlg *print_dlg)
{
	gint    bw = GUI_BUTTON_HLABEL_WIDTH_DEF,
		bh = GUI_BUTTON_HLABEL_HEIGHT_DEF,
		border_major = 5;
	gchar *s;
	GdkWindow *window;
	GtkAccelGroup *accelgrp;
	GtkStyle *style;
	GtkWidget       *w,
			*parent, *parent2,
			*toplevel,
			*main_vbox;
	PrintPreviewDlg *d = PRINT_PREVIEW_DLG(
	    g_malloc0(sizeof(PrintPreviewDlg))
	);
	if(d == NULL)
	    return(NULL);

	gtk_widget_push_visual(gdk_rgb_get_visual());
	gtk_widget_push_colormap(gdk_rgb_get_cmap());
	d->toplevel = toplevel = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_widget_pop_visual();
	gtk_widget_pop_colormap();
	d->accelgrp = accelgrp = gtk_accel_group_new();
/*	d->flags = 0; */
	d->print_dlg = print_dlg;

	/* Toplevel GtkWindow */
	w = toplevel;
#ifdef PROG_NAME
	gtk_window_set_wmclass(
	    GTK_WINDOW(w), "dialog", PROG_NAME
	);
#endif
	gtk_window_set_policy(GTK_WINDOW(w), TRUE, TRUE, FALSE);
	s = g_strdup_printf(
	    PRINT_PREVIEW_DLG_TITLE ": %s",
	    STRISEMPTY(print_dlg->path) ? "Untitled" : print_dlg->path
	);
	gtk_window_set_title(GTK_WINDOW(w), s);
	g_free(s);
	gtk_widget_set_usize(
	    w,
	    PRINT_PREVIEW_DLG_WIDTH, PRINT_PREVIEW_DLG_HEIGHT
	);
	gtk_widget_add_events(
	    w,
	    GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK
	);
	gtk_signal_connect(
	    GTK_OBJECT(w), "delete_event",
	    GTK_SIGNAL_FUNC(PrintPreviewDlgDeleteEventCB), d
	);
	gtk_widget_realize(w);
	d->flags |= PRINT_PREVIEW_DLG_REALIZED;
	window = w->window;
	if(window != NULL)
	{
	    gdk_window_set_decorations(
		window,
		GDK_DECOR_BORDER | GDK_DECOR_RESIZEH |
		    GDK_DECOR_TITLE | GDK_DECOR_MAXIMIZE
	    );
	    gdk_window_set_functions(
		window,
		GDK_FUNC_RESIZE | GDK_FUNC_MOVE | GDK_FUNC_MAXIMIZE |
		    GDK_FUNC_CLOSE
	    );
	    GUISetWMIcon(
		window,
		(guint8 **)icon_print_preview_48x48_xpm
	    );
	}
	style = gtk_widget_get_style(w);
	gtk_accel_group_attach(accelgrp, GTK_OBJECT(w));
	parent = w;


	/* Main GtkVBox */
	main_vbox = w = gtk_vbox_new(FALSE, 0);
	gtk_container_add(GTK_CONTAINER(parent), w);
	gtk_widget_show(w);
	parent = w;

	/* Preview GtkFrame */
	w = gtk_frame_new(NULL);
	gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_IN);
	gtk_box_pack_start(GTK_BOX(parent), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
	parent2 = w;

	/* Preview */
	d->preview = PreviewNew(
	    parent2,
	    style,
	    &style->bg[GTK_STATE_ACTIVE],	/* Background color */
	    FALSE,				/* Not interactive */
	    print_dlg
	);


	/* Buttons GtkHBox */
	w = gtk_hbox_new(TRUE, 0);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	gtk_widget_show(w);
	parent2 = w;

	/* Print GtkButton */
	d->print_btn = w = GUIButtonPixmapLabelH(
	    (guint8 **)icon_print2_20x20_xpm,
	    "Print",
	    NULL
	);
	gtk_widget_set_usize(w, bw, bh);
	GTK_WIDGET_SET_FLAGS(w, GTK_CAN_DEFAULT);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "clicked",
	    GTK_SIGNAL_FUNC(PrintPreviewDlgPrintCB), d
	);
	gtk_accel_group_add(
	    accelgrp,
	    GDK_p, GDK_CONTROL_MASK,
	    GTK_ACCEL_VISIBLE,
	    GTK_OBJECT(w),
	    "clicked"
	);
	GUIButtonLabelUnderline(w, GDK_p);
	gtk_widget_show(w);

	/* Close GtkButton */
	d->close_btn = w = GUIButtonPixmapLabelH(
	    (guint8 **)icon_close_20x20_xpm,
	    "Close",
	    NULL
	);
	gtk_widget_set_usize(w, bw, bh);
	GTK_WIDGET_SET_FLAGS(w, GTK_CAN_DEFAULT);
	gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "clicked",
	    GTK_SIGNAL_FUNC(PrintPreviewDlgCloseCB), d
	);
	gtk_accel_group_add(
	    accelgrp,
	    GDK_Escape, 0,
	    GTK_ACCEL_VISIBLE,
	    GTK_OBJECT(w),
	    "clicked"
	);
	gtk_accel_group_add(
	    accelgrp,
	    GDK_c, GDK_CONTROL_MASK,
	    GTK_ACCEL_VISIBLE,
	    GTK_OBJECT(w),
	    "clicked"                             
	);
	GUIButtonLabelUnderline(w, GDK_c);
	gtk_widget_show(w);



	return(d);
}

/*
 *	Updates the PrintPreviewDlg's widgets to reflect current values.
 */
static void PrintPreviewDlgUpdate(PrintPreviewDlg *d)
{
	if(d == NULL)
	    return;

}

/*
 *	Maps the PrintPreviewDlg.
 */
static void PrintPreviewDlgMap(PrintPreviewDlg *d)
{
	GtkWidget *w = d->print_btn;
	gtk_widget_grab_focus(w);
	gtk_widget_grab_default(w);
	w = d->toplevel;
	gtk_window_set_modal(GTK_WINDOW(w), TRUE);
	gtk_window_set_transient_for(
	    GTK_WINDOW(w),
	    GTK_WINDOW(d->print_dlg->toplevel)
	);
	gtk_widget_show(w);
	d->flags |= PRINT_PREVIEW_DLG_MAPPED;
}

/*
 *	Unmaps the PrintPreviewDlg.
 */
static void PrintPreviewDlgUnmap(PrintPreviewDlg *d)
{
	GtkWidget *w = d->toplevel;
	gtk_widget_hide(d->toplevel);
	d->flags &= ~PRINT_PREVIEW_DLG_MAPPED;
	gtk_window_set_modal(GTK_WINDOW(w), FALSE);
	gtk_window_set_transient_for(GTK_WINDOW(w), NULL);
}

/*
 *	Deletes the PrintPreviewDlg.
 */
static void PrintPreviewDlgDelete(PrintPreviewDlg *d)
{
	if(d == NULL)
	    return;

	d->freeze_count++;

	PreviewDelete(d->preview);
	GTK_WIDGET_DESTROY(d->toplevel);
	GTK_ACCEL_GROUP_UNREF(d->accelgrp);

	d->freeze_count--;

	g_free(d);
}


/*
 *	Queries user to print the specified ImgView image.
 *
 *	Creates a new print dialog and blocks output until the user
 *	responds.
 *
 *	The specified values *v will be used to initialize the print
 *	values and *v will be modified if TRUE is returned, otherwise
 *	*v should be considered garbage after this call.
 */
#ifdef HAVE_LIBENDEAVOUR2
gboolean PrintDlgGetResponse(
	imgview_image_struct *img,		/* Image to print */
	const gchar *path,
	print_values_struct *v,			/* Print values (will be modified) */
	const guint8 *bg_color,			/* 4 bytes RGBA */
	GtkWidget *toplevel,
	EDVContext *edv2_ctx
)
#else
gboolean PrintDlgGetResponse(
	imgview_image_struct *img,		/* Image to print */
	const gchar *path,
	print_values_struct *v,			/* Print values (will be modified) */
	const guint8 *bg_color,			/* 4 bytes RGBA */
	GtkWidget *toplevel
)
#endif
{
	gboolean got_response;
	gchar *xprint_server_address = NULL;
	GdkRectangle *output;
	GtkWidget *w;
	print_value_flags vflags;
	PrintDlg *d;

	if((v == NULL) || (img == NULL))
	    return(FALSE);

	/* Create a new print dialog */
	d = PrintDlgNew(path, toplevel);
	if(d == NULL)
	    return(FALSE);

	d->freeze_count++;

	/* Begin setting values */
	vflags = v->flags;

	/* Source ImgView image that user wants to print */
	d->src_img = img;
	if(vflags & PRINT_VALUE_FRAME)
	{
	    GtkAdjustment *adj = gtk_spin_button_get_adjustment(
		GTK_SPIN_BUTTON(d->output_frame_spin)
	    );
	    d->src_frame_num = v->frame;
	    adj->lower = 1.0f;
	    adj->upper = (gfloat)img->nframes;
	    adj->value = (gfloat)(v->frame + 1);
	    gtk_adjustment_changed(adj);
	    gtk_adjustment_value_changed(adj);
	}
	else
	{
	    d->src_frame_num = 0;
	}

	/* Background color */
	if(bg_color != NULL)
	    memcpy(&d->bg_color, bg_color, 4);
	else
	    memset(&d->bg_color, 0xffffffff, 4);

	/* X Print Server Address */
	if(vflags & PRINT_VALUE_XPRINT_SERVER_ADDRESS)
	{
	    g_free(xprint_server_address);
	    xprint_server_address = STRDUP(v->xprint_server_address);

	    if(xprint_server_address != NULL)
		gtk_entry_set_text(
		    GTK_ENTRY(d->xprint_server_address_entry),
		    xprint_server_address
		);		
	}

	/* Update the printers list */
	if(TRUE)
	{
	    PrintDlgUpdateXPrintServerPrinters(
		d,
		xprint_server_address,
		FALSE				/* Not verbose */
	    );
	}

	/* Printer */
	if(vflags & PRINT_VALUE_PRINTER)
	{
	    GtkWidget	*pulistbox = d->printer_pulistbox,
			*pulist = PUListBoxGetPUList(pulistbox);
	    const gint m = PUListGetTotalItems(pulist);
	    IVPrintXPrintPrinter *p;

	    if(v->printer != NULL)
	    {
		const gchar *printer_name = v->printer;
		gint i;

		for(i = 0; i < m; i++)
		{
		    p = IV_PRINT_XPRINT_PRINTER(PUListGetItemData(
			pulist,
			i
		    ));
		    if(p == NULL)
			continue;

		    if(STRISEMPTY(p->name))
			continue;

		    if(!g_strcasecmp(printer_name, p->name))
		    {
			PUListBoxSelect(pulistbox, i);
			break;
		    }
		}
	    }
	    else
	    {
		PUListBoxSelect(pulistbox, 0);
	    }

	    /* Get the selected printer values */
	    p = IV_PRINT_XPRINT_PRINTER(PUListBoxGetSelectedData(
		pulistbox
	    ));
	    if(p != NULL)
		PrintDlgUpdatePrinterValues(
		    d,
		    p->name,
		    FALSE			/* Not verbose */
		);
	}

	/* Medium */
	if((vflags & PRINT_VALUE_INPUT_TRAY) &&
	   (vflags & PRINT_VALUE_MEDIUM)
	)
	{
	    gboolean selected = FALSE;
	    GtkCList *clist = GTK_CLIST(d->medium_ctree);
	    GtkCTree *ctree = GTK_CTREE(clist);

	    gtk_clist_freeze(clist);

	    /* Tray and medium specified? */
	    if(!STRISEMPTY(v->input_tray) && !STRISEMPTY(v->medium))
	    {
		const gchar	*tray_name = v->input_tray,
				*medium_name = v->medium;
		GtkCTreeNode	*node,
				*parent_node;
		IVPrintXPrintMedium *m;

		/* Select the first medium who's tray and medium
		 * matches the specified tray and medium
		 */
		for(parent_node = gtk_ctree_node_nth(ctree, 0);
		    parent_node != NULL;
		    parent_node = GTK_CTREE_NODE_SIBLING(parent_node)
		)
		{
		    for(node = GTK_CTREE_NODE_CHILD(parent_node);
			node != NULL;
			node = GTK_CTREE_NODE_SIBLING(node)
		    )
		    {
			m = IV_PRINT_XPRINT_MEDIUM(gtk_ctree_node_get_row_data(
			    ctree,
			    node
			));
			if(m == NULL)
			    continue;

			if(STRISEMPTY(m->name) || STRISEMPTY(m->tray))
			    continue;

			if(!g_strcasecmp(m->tray, tray_name) &&
			   !g_strcasecmp(m->name, medium_name)
			)
			{
			    gtk_clist_unselect_all(clist);
			    gtk_ctree_select(ctree, node);
			    selected = TRUE;
			    break;
			}
		    }
		}
	    }
	    /* Only medium specified? */
	    else if(!STRISEMPTY(v->medium))
	    {
		const gchar *medium_name = v->medium;
		GtkCTreeNode	*node,
				*parent_node;
		IVPrintXPrintMedium *m;

		/* No tray specified, select the first medium that
		 * matches the specified medium
		 */
		for(parent_node = gtk_ctree_node_nth(ctree, 0);
		    parent_node != NULL;
		    parent_node = GTK_CTREE_NODE_SIBLING(parent_node)
		)
		{
		    for(node = GTK_CTREE_NODE_CHILD(parent_node);
			node != NULL;
			node = GTK_CTREE_NODE_SIBLING(node)
		    )
		    {
			m = IV_PRINT_XPRINT_MEDIUM(gtk_ctree_node_get_row_data(
			    ctree,
			    node
			));
			if(m == NULL)
			    continue;

			if(STRISEMPTY(m->name))
			    continue;

			if(!g_strcasecmp(m->name, medium_name))
			{
			    gtk_clist_unselect_all(clist);
			    gtk_ctree_select(ctree, node);
			    selected = TRUE;
			    break;
			}
		    }
		}
	    }

	    gtk_clist_thaw(clist);
	}

	/* Orientation */
	if(vflags & PRINT_VALUE_ORIENTATION)
	{
	    GtkWidget	*pulistbox = d->orientation_pulistbox,
			*pulist = PUListBoxGetPUList(pulistbox);
	    const gint m = PUListGetTotalItems(pulist);

	    if(v->orientation != NULL)
	    {
		const gchar *orientation_name = v->orientation;
		gint i;
		const gchar *s;
		for(i = 0; i < m; i++)
		{
		    s = (const gchar *)PUListGetItemData(pulist, i);
		    if(s == NULL)
			continue;

		    if(!g_strcasecmp(orientation_name, s))
		    {
			PUListBoxSelect(pulistbox, i);
			break;
		    }
		}
	    }
	    else
	    {
		PUListBoxSelect(pulistbox, 0);
	    }
	}

	/* Resolution */
	if(vflags & PRINT_VALUE_RESOLUTION)
	{
	    GtkWidget	*pulistbox = d->resolution_pulistbox,
			*pulist = PUListBoxGetPUList(pulistbox);
	    const gint m = PUListGetTotalItems(pulist);

	    if((v->resolution_x_dpi > 0) && (v->resolution_y_dpi > 0))
	    {
		gint i;
		IVPrintXPrintResolution *r;

		for(i = 0; i < m; i++)
		{
		    r = IV_PRINT_XPRINT_RESOLUTION(PUListGetItemData(
			pulist,
			i
		    ));
		    if(r == NULL)
			continue;

		    if((r->x_dpi == v->resolution_x_dpi) &&
		       (r->y_dpi == v->resolution_y_dpi)
		    )
		    {
			PUListBoxSelect(pulistbox, i);
			break;
		    }
		}
	    }
	}

	/* Brightness */
	if(vflags & PRINT_VALUE_BRIGHTNESS)
	{
	    GtkAdjustment *adj = d->brightness_adj;
	    if(adj != NULL)
	    {
		adj->value = v->brightness;
		if(adj->value < adj->lower)
		    adj->value = adj->lower;
		if(adj->value > adj->upper)
		    adj->value = adj->upper;
		gtk_signal_emit_by_name(GTK_OBJECT(adj), "changed");
	    }
	}

	/* Contrast */
	if(vflags & PRINT_VALUE_CONTRAST)
	{
	    GtkAdjustment *adj = d->contrast_adj;
	    if(adj != NULL)
	    {
		adj->value = v->contrast;
		if(adj->value < adj->lower)
		    adj->value = adj->lower;
		if(adj->value > adj->upper)
		    adj->value = adj->upper;
		gtk_signal_emit_by_name(GTK_OBJECT(adj), "changed");
	    }
	}

	/* Visual */
	if(vflags & PRINT_VALUE_VISUAL)
	{
	    w = d->visual_bw_radio;
	    if((w != NULL) ? GTK_IS_TOGGLE_BUTTON(w) : FALSE)
		GTK_TOGGLE_BUTTON(w)->active = FALSE;
	    w = d->visual_greyscale_radio;
	    if((w != NULL) ? GTK_IS_TOGGLE_BUTTON(w) : FALSE)
		GTK_TOGGLE_BUTTON(w)->active = FALSE;
	    w = d->visual_color_radio;
	    if((w != NULL) ? GTK_IS_TOGGLE_BUTTON(w) : FALSE)
		GTK_TOGGLE_BUTTON(w)->active = FALSE;

	    switch(v->visual)
	    {
	      case PRINT_VISUAL_BW:
		w = d->visual_bw_radio;
		break;
	      case PRINT_VISUAL_GREYSCALE:
		w = d->visual_greyscale_radio;
		break;
	      case PRINT_VISUAL_RGB:
		w = d->visual_color_radio;
		break;
	    }
	    if((w != NULL) ? GTK_IS_TOGGLE_BUTTON(w) : FALSE)
		GTK_TOGGLE_BUTTON(w)->active = TRUE;
	}

	/* Number Of Coppies */
	if(vflags & PRINT_VALUE_OUTPUT_NCOPPIES)
	    gtk_spin_button_set_value(
		GTK_SPIN_BUTTON(d->output_ncoppies_spin),
		(gfloat)v->output_ncoppies
	    );

	/* Output Image Geometry */
	output = &d->output_img_geometry;
	if(vflags & PRINT_VALUE_OUTPUT_X)
	    output->x = v->output_x;
	if(vflags & PRINT_VALUE_OUTPUT_Y)
	    output->y = v->output_y;
	if(vflags & PRINT_VALUE_OUTPUT_WIDTH)
	    output->width = v->output_width;
	if(vflags & PRINT_VALUE_OUTPUT_HEIGHT)
	    output->height = v->output_height;
	if(vflags & PRINT_VALUE_OUTPUT_MAINTAIN_ASPECT)
	    d->output_maintain_aspect = v->output_maintain_aspect;
	PrintDlgOutputSetWidgets(d);

	PrintDlgUpdate(d);

	/* Get the initial medium size */
	if(TRUE)
	{
	    gint medium_width, medium_height;
	    if(PrintDlgGetMediumSize(
		d,
		&medium_width, &medium_height
	    ))
	    {
		GdkRectangle *medium = &d->medium_size;
		medium->width = medium_width;
		medium->height = medium_height;
	    }
	    else
	    {
		GdkRectangle *medium = &d->medium_size;
		medium->width = PRINT_DLG_DEF_MEDIUM_WIDTH;
		medium->height = PRINT_DLG_DEF_MEDIUM_HEIGHT;
	    }
	    PreviewResize(d->preview);
	}


	g_free(xprint_server_address);
	xprint_server_address = NULL;

	d->freeze_count--;

	/* Map the print dialog */
	PrintDlgMap(d);

	/* Enter main loop and wait for user response */
	d->main_level++;
	gtk_main();

	/* Break out of any additional main loop levels */
	while(d->main_level > 0)
	{
	    d->main_level--;
	    gtk_main_quit();
	}


	/* Got response? */
	got_response = (d->flags & PRINT_DLG_RESPONSE_OK) ? TRUE : FALSE;
	if(got_response)
	{
	    /* X Print Server Address */
	    if(vflags & PRINT_VALUE_XPRINT_SERVER_ADDRESS)
	    {
		g_free(v->xprint_server_address);
		v->xprint_server_address = STRDUP(gtk_entry_get_text(
		    GTK_ENTRY(d->xprint_server_address_entry)
		));
	    }
	    /* Printer */
	    if(vflags & PRINT_VALUE_PRINTER)
	    {
		IVPrintXPrintPrinter *p = IV_PRINT_XPRINT_PRINTER(
		    PUListBoxGetSelectedData(d->printer_pulistbox)
		);
		if(p != NULL)
		{
		    g_free(v->printer);
		    v->printer = STRDUP(p->name);
		}
	    }
	    /* Input Tray & Medium */
	    if((vflags & PRINT_VALUE_INPUT_TRAY) &&
	       (vflags & PRINT_VALUE_MEDIUM)
	    )
	    {
		GtkCList *clist = GTK_CLIST(d->medium_ctree);
		GList *glist = clist->selection_end;
		GtkCTreeNode *node = (glist != NULL) ?
		    (GtkCTreeNode *)glist->data : NULL;
		IVPrintXPrintMedium *m = IV_PRINT_XPRINT_MEDIUM(
		    gtk_ctree_node_get_row_data(GTK_CTREE(clist), node)
		);
		if(m != NULL)
		{
		    g_free(v->input_tray);
		    v->input_tray = STRDUP(m->tray);
		    g_free(v->medium);
		    v->medium = STRDUP(m->name);
		}
	    }
	    /* Orientation */
	    if(vflags & PRINT_VALUE_ORIENTATION)
	    {
		const gchar *orientation = (const gchar *)PUListBoxGetSelectedData(
		    d->orientation_pulistbox
		);
		g_free(v->orientation);
		v->orientation = STRDUP(orientation);
	    }
	    /* Resolution */
	    if(vflags & PRINT_VALUE_RESOLUTION)
	    {
		IVPrintXPrintResolution *r = IV_PRINT_XPRINT_RESOLUTION(PUListBoxGetSelectedData(
		    d->resolution_pulistbox
		));
		if(r != NULL)
		{
		    v->resolution_x_dpi = r->x_dpi;
		    v->resolution_y_dpi = r->y_dpi;
		}
	    }
	    /* Brightness */
	    if(vflags & PRINT_VALUE_BRIGHTNESS)
	    {
		GtkAdjustment *adj = d->brightness_adj;
		if(adj != NULL)
		{
		    v->brightness = adj->value;
		}
	    }
	    /* Contrast */
	    if(vflags & PRINT_VALUE_CONTRAST)
	    {
		GtkAdjustment *adj = d->contrast_adj;
		if(adj != NULL)
		{
		    v->contrast = adj->value;
		}
	    }
	    /* Visual */
	    if(vflags & PRINT_VALUE_VISUAL)
	    {
		if(GTK_TOGGLE_BUTTON_GET_ACTIVE(
		    d->visual_bw_radio
		))
		    v->visual = PRINT_VISUAL_BW;
		else if(GTK_TOGGLE_BUTTON_GET_ACTIVE(   
		    d->visual_greyscale_radio 
		)) 
		    v->visual = PRINT_VISUAL_GREYSCALE; 
		else if(GTK_TOGGLE_BUTTON_GET_ACTIVE(
		    d->visual_color_radio 
		)) 
		    v->visual = PRINT_VISUAL_RGB;
	    }

	    /* Frame */
	    if(vflags & PRINT_VALUE_FRAME)
		v->frame = d->src_frame_num;

	    /* Output image geometry */
	    output = &d->output_img_geometry;
	    if(vflags & PRINT_VALUE_OUTPUT_X)
		v->output_x = output->x;
	    if(vflags & PRINT_VALUE_OUTPUT_Y)
		v->output_y = output->y;
	    if(vflags & PRINT_VALUE_OUTPUT_WIDTH)
		v->output_width = output->width;
	    if(vflags & PRINT_VALUE_OUTPUT_HEIGHT)
		v->output_height = output->height;

	    if(vflags & PRINT_VALUE_OUTPUT_MAINTAIN_ASPECT)
		v->output_maintain_aspect = d->output_maintain_aspect;

	    if(vflags & PRINT_VALUE_OUTPUT_NCOPPIES)
	    {
		w = d->output_ncoppies_spin;
		if((w != NULL) ? GTK_IS_SPIN_BUTTON(w) : FALSE)
		    v->output_ncoppies = gtk_spin_button_get_value_as_int(
			GTK_SPIN_BUTTON(w)
		    );
	    }
	}

	/* Delete the print dialog */
	PrintDlgDelete(d);

	return(got_response);
#undef PRINT_COMMAND_MAX
}

#endif	/* !_WIN32 && HAVE_XPRINT */

