#include <gtk/gtk.h>
#include <gdk/gdkprivate.h>
#if defined(_WIN32)
# include <gdk/gdkwin32.h>
#else
# include <gdk/gdkx.h>
# define HAVE_X
#endif

#include "guiutils.h"
#include "guirgbimg.h"
#include "screenshot.h"


/*
 *	Screenshot Targets:
 */
typedef enum {
	SCREENSHOT_TARGET_UNDETERMINED,		/* Not yet determined */
	SCREENSHOT_TARGET_WINDOW,		/* Window */
	SCREENSHOT_TARGET_RECTANGULAR,		/* Rectangular area */
	SCREENSHOT_TARGET_ABORT			/* User aborted */
} ScreenshotTarget;


/* Pointer to the last screenshot error message */
static const gchar	*last_screenshot_error = NULL;


typedef struct _ScreenshotData		ScreenshotData;
#define SCREENSHOT_DATA(p)		((ScreenshotData *)(p))


#if defined(HAVE_X)
static void draw_X_rectangular_area(
	ScreenshotData *data,
	const Bool draw_new
);
static gint grab_X_window(
	ScreenshotData *data,
	imgview_struct *iv,
	const gboolean include_decorations,
	imgview_image_struct **img_rtn,
	gint *x_rtn, gint *y_rtn,
	gint *base_width_rtn, gint *base_height_rtn
);
static gint grab_X_rectangular(
	ScreenshotData *data,
	imgview_struct *iv,
	const gboolean include_decorations,
	imgview_image_struct **img_rtn,
	gint *x_rtn, gint *y_rtn,
	gint *base_width_rtn, gint *base_height_rtn
);
#endif

/* Callbacks */
static gint screenshot_window_destroy_cb(
	GtkWidget *widget, GdkEvent *event, gpointer data
);
static gint screenshot_delay_timeout_cb(gpointer data);
static gint screenshot_button_cb(
	GtkWidget *widget, GdkEventButton *button, gpointer data
);
static gint screenshot_motion_cb(
	GtkWidget *widget, GdkEventMotion *motion, gpointer data
);

/* Screenshot */
const gchar *iv_screenshot_get_error(void);
gint iv_screenshot_start(
	imgview_struct *iv,
	const gulong delay_ms,
	const gboolean include_decorations,
	GtkWidget *grab_window,
	GdkCursor *grab_cursor,
	imgview_image_struct **img_rtn,
	gint *x_rtn, gint *y_rtn,
	gint *base_width_rtn, gint *base_height_rtn
);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define ABSOLUTE_VALUE(x)	(((x) < 0) ? ((x) * -1) : (x))
#define STRLEN(s)	(((s) != NULL) ? (gint)strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : TRUE)


/*
 *	Screenshot Data:
 */
struct _ScreenshotData {

	ScreenshotTarget	target;
	GdkWindow	*root_win;
	gboolean	active;			/* Indicates when a grab
											 * becomes active, while this
											 * is FALSE no grab stuff
											 * should be processed on the
											 * signal callbacks */
	gint		nmain_levels;		/* Number of GTK main levels */

	gint		grab_button_number;	/* Pointer button that is
											 * currently pressed, can be
											 * 0 for no button pressed */

	/* Coordinates for the grab, x[0] and y[0] represent starting
	 * coordinates for targets SCREENSHOT_TARGET_RECTANGULAR or
	 * SCREENSHOT_TARGET_WINDOW
	 *
	 * The x[1] and y[1] are only defined if the target is
	 * SCREENSHOT_TARGET_RECTANGULAR
	 *
	 * All these values unare defined if the target is
	 * SCREENSHOT_TARGET_UNDETERMINED
	 */
	gint		x[2], y[2];

#if defined(HAVE_X)
	/* X resourcse */
	Display		*xdisplay;
	int		xscreen_num;
	Window		xrootwin;
	GC		gc_inverse;
	/* Last drawn rectangle position and size */
	Bool		have_last_rect;
	int		last_rect_x,
					last_rect_y;
	unsigned int	last_rect_width,
					last_rect_height;
#endif

};


#if defined(HAVE_X)
/*
 *	Draws an X grab rectangle.
 *
 *	This will use the inverse GC on the screenshot data to; first
 *	reset the bits drawn by a previous call to this function
 *	(if there was one, check grab_data->have_last_rect == True) by
 *	drawing the rectangle bits using the previous geometry values on
 *	the grab_data structure, then draw the new rectangle bits if
 *	draw_new is True (using the same inverse GC).
 *
 *	Note that if draw_new is False then grab_data->have_last_rect will
 *	be set to False, indicating for the next call to this function
 *	there would have been no previous rectangle drawn.
 */
static void draw_X_rectangular_area(
	ScreenshotData *data,
	const Bool draw_new
)
{
	int		x, y,
					width, height;
	Display *dpy;
	Window xrootwin;
	Drawable d;
	GC gc;

	if(data == NULL)
		return;

	dpy = data->xdisplay;
	xrootwin = data->xrootwin;
	gc = data->gc_inverse;
	if((dpy == NULL) || (xrootwin == None) || (gc == None))
		return;

	/* Get the root window as the drawable to draw to */
	d = (Drawable)xrootwin;

	/* Get new rectangle bits geometry values */
	x = MIN(data->x[0], data->x[1]);
	y = MIN(data->y[0], data->y[1]);
	width = ABSOLUTE_VALUE(data->x[1] - data->x[0]);
	height = ABSOLUTE_VALUE(data->y[1] - data->y[0]);

	/* Draw over last rectangle?
	 *
	 * This is to effectivly reset the bits drawn by a prior call
	 * to this function
	 */
	if(data->have_last_rect)
		XDrawRectangle(
			dpy, d, gc,
			data->last_rect_x,
			data->last_rect_y,
			data->last_rect_width,
			data->last_rect_height
		);

	/* Draw the new rectangle bits and the new rectangle has
	 * valid size?
	 */
	if(draw_new && (width > 0) && (height > 0))
	{
		XDrawRectangle(dpy, d, gc, x, y, width, height);

		/* Since the new rectangle bits have been drawn, its geometry
		 * values should be recorded so that the next call to this
		 * function resets the bits just drawn
		 */
		data->have_last_rect = True;
		data->last_rect_x = x;
		data->last_rect_y = y;
		data->last_rect_width = width;
		data->last_rect_height = height;
	}
	else
	{
		/* Do not draw new rectangle bits */

		/* Need to mark that there would be no prior rectangle
		 * drawn for the next call to this function
		 */
		data->have_last_rect = False;
	}
}
#endif	/* defined(HAVE_X) */

#if defined(HAVE_X)
/*
 *	Grabs an X window.
 *
 *	The pointer will be warped (moved) to the coordinates specified
 *	the screenshot data and then the pointer will be queried to see
 *	which Window the pointer is over. That Window (or next child if
 *	include_decorations is TRUE) will have its geometry obtained
 *	and a rectangular area will be grabbed from the root Window that
 *	matches the geometry of the Window in question.
 *
 *	Returns 0 on success or non-zero on error.
 */
static gint grab_X_window(
	ScreenshotData *data,
	imgview_struct *iv,
	const gboolean include_decorations,
	imgview_image_struct **img_rtn,
	gint *x_rtn, gint *y_rtn,
	gint *base_width_rtn, gint *base_height_rtn
)
{
	int		x, y,
			x_return, y_return,
			xr2, yr2;
	Display *dpy;
	Window		xrootwin,
			xwin,
			xrootrtn;		/* Dummy root for fetches */
	GdkWindowPrivate *private;
	unsigned int	width_return, height_return,
			border_width_return,
			depth_return;
	unsigned int mask;
	gint		status,
			cx, cy,
			cwidth, cheight,
			rwidth, rheight;

	if((iv == NULL) || (data == NULL))
		return(-1);

	dpy = data->xdisplay;
	xrootwin = data->xrootwin;
	if((dpy == NULL) || (xrootwin == None))
		return(-1);

	/* Finish processing any pending X operations to ensure that
	 * we do not haev any marks that we drew over the Window that
	 * we want to grab
	 */
	(void)XFlush(dpy);

	/* Grab the X server so that everything will hold still while
	 * we obtain the Window geometry
	 */
	(void)XGrabServer(dpy);

	/* Get grab window coordinates relative to root window */
	x = (int)data->x[0];
	y = (int)data->y[0];

	/* Find child of root window that the given coordinates lie
	 * under
	 */
	(void)XWarpPointer(
		dpy,
		None,				/* No source Window */
		xrootwin,
		0, 0,
		0, 0,
		x, y)
	;
	if(!XQueryPointer(
		dpy,
		xrootwin,
		&xrootrtn,
		&xwin,
		&x_return, &y_return,
		&xr2, &yr2,
		&mask
	))
	{
		last_screenshot_error =
"XQueryPointer() failed to find the clicked on Window";
		XUngrabServer(dpy);
		return(-1);
	}

	/* If the coordinates do not lie within a child Window then set
	 * the Window to grab as the root Window (desktop)
	 */
	if(xwin == None)
		xwin = xrootwin;

	/* At this point we have the X id of the root Window and the
	 * immediate child or the root Window that the pointer was over
	 *
	 * We need to handle two possible cases, if the root and child
	 * window are the same, then we grab the root window, otherwise
	 * we grab the portion of the root Window that the child Window
	 * is displayed at
	 *
	 * In either case the values for variables x, y, rwidth, rheight,
	 * cwidth, and cheight will be set
	 */
	if(xwin == xrootwin)
	{
		/* Grabbing the root Window
		 *
		 * Get the size of the root Window
		 */
		(void)XGetGeometry(
			dpy,
			xwin,
			&xrootrtn,
			&x_return, &y_return,
			&width_return, &height_return,
			&border_width_return,
			&depth_return
		);
		cx = (gint)x_return;
		cy = (gint)y_return;
		rwidth = cwidth = (gint)width_return;
		rheight = cheight = (gint)height_return;
	}
	else
	{
		/* Grabbing a child window, there are two posibilites now.
		 * The first is that we just grab the child window without
		 * the WM decorations. The second is that we grab the
		 * entire child window and its WM decorations if any
		 */

		/* Get size of the root Window first */
		(void)XGetGeometry(
			dpy,
			xrootwin,
			&xrootrtn,
			&x_return, &y_return,
			&width_return, &height_return,
			&border_width_return,
			&depth_return
		);
		rwidth = (gint)width_return;
		rheight = (gint)height_return;

		/* Now work on getting the clicked on child window */
		if(include_decorations)
		{
			/* Grab child window with its WM decorations or the
			 * child window has no WM decorations
			 */
			(void)XGetGeometry(
				dpy,
				xwin,
				&xrootrtn,
				&x_return, &y_return,
				&width_return, &height_return,
				&border_width_return,
				&depth_return
			);
			cx = (gint)x_return;
			cy = (gint)y_return;
			cwidth = (gint)width_return;
			cheight = (gint)height_return;
		}
		else
		{
			/* Do not grab WM decorations and this Window is
			 * suspected to have decorations
			 */
			Window tw;

			/* Get the position of the child Window */
			(void)XGetGeometry(
				dpy,
				xwin,
				&xrootrtn,
				&x_return, &y_return,
				&width_return, &height_return,
				&border_width_return,
				&depth_return
			);
			cx = (gint)x_return;
			cy = (gint)y_return;
			cwidth = (gint)width_return;
			cheight = (gint)height_return;

			/* Now get the next child if any */
			if(!XQueryPointer(
				dpy,
				xwin,
				&xrootrtn,
				&tw,
				&x_return, &y_return,
				&xr2, &yr2,
				&mask
			))
			{
				last_screenshot_error =
"XQueryPointer() failed to find undecorated clicked on Window";
				XUngrabServer(dpy);
				return(-1);
			}
			if(tw != None)
			{
				/* Got next child, so update xwin and its geometry */
				xwin = tw;
				(void)XGetGeometry(
					dpy,
					xwin,
					&xrootrtn,
					&x_return, &y_return,
					&width_return, &height_return,
					&border_width_return,
					&depth_return
				);
				cx += (gint)x_return;
				cy += (gint)y_return;
				cwidth = (gint)width_return;
				cheight = (gint)height_return;
			}
		}

	}
	/* At this point we now have the size of the root window in
	 * rwidth and rheight plus the child window geometry relative
	 * to the root window at x, y, cwidth, and cheight
	 */
	/* Clip the child Window's geometry */
	if((cx + cwidth) > rwidth)
		cwidth = rwidth - cx;
	if(cx < 0)
	{
		cwidth += cx;
		cx = 0;
	}
	if((cy + cheight) > rheight)
		cheight = rheight - cy;
	if(cy < 0)
	{
		cheight += cy;
		cy = 0;
	}
	/* Check if the child Window has no area */
	if((cwidth <= 0) || (cheight <= 0))
	{
		last_screenshot_error =
			"Window has no area (width and/or height <= 0)";
		XUngrabServer(dpy);
		return(-1);
	}

	/* We now have the rectangular area that we want to grab
	 * specified in cx, cy, cwidth, and cheight
	 *
	 * Set the return geometry values
	 */
	if(x_rtn != NULL)
		*x_rtn = cx;
	if(y_rtn != NULL)
		*y_rtn = cy;
	if(base_width_rtn != NULL)
		*base_width_rtn = rwidth;
	if(base_height_rtn != NULL)
		*base_height_rtn = rheight;

	/* Reset the status to -1, indicating error. If we were able to
	 * get the image then we'll set it to 0
	 */
	status = -1;

	/* Create a GdkWindow structure to reference the xroot_win */
	private = (GdkWindowPrivate *)g_malloc0(sizeof(GdkWindowPrivate));
	if(private != NULL)
	{
		const gint bpp = 4;			/* RGBA */
		gint bpl;
		guint8 *data_rgba;
		GdkRectangle rect;
		GdkWindow *window = (GdkWindow *)private;

		/* Set up the GdkWindow values with the X values for grabbing */
		private->parent = NULL;
		private->xwindow = xrootwin;
		private->xdisplay = dpy;
		private->x = 0;
		private->y = 0;
		private->width = rwidth;
		private->height = rheight;
		private->mapped = 1;

		/* Set the grab area */
		rect.x = cx;
		rect.y = cy;
		rect.width = cwidth;
		rect.height = cheight;

		/* Grab and get the RGBA image data */
		data_rgba = gdk_get_rgba_image(
			window,
			NULL,				/* No mask */
			&rect,
			&cwidth, &cheight, &bpl
		);
		if(data_rgba != NULL)
		{
			imgview_image_struct *img = ImgViewImageNew(
				cwidth, cheight,
				bpp, bpl
			);
			if(img != NULL)
			{
				(void)ImgViewImageAppendFrame(
					img,
					data_rgba,		/* Transfered */
					0l			/* No delay */
				);
				*img_rtn = img;
				status = 0;
			}
			else
			{
				g_free(data_rgba);
				last_screenshot_error = "Memory allocation error";
				status = -3;
			}
		}
		else
		{
			last_screenshot_error = "Unable to grab an X Drawable";
		}

		g_free(private);
	}

	/* Ungrab the X server */
	(void)XUngrabServer(dpy);

	return(status);
}
#endif	/* defined(HAVE_X) */

#if defined(HAVE_X)
/*
 *	Grabs a rectangular area.
 *
 *	The geometry of the rectangle will be obtained from the
 *	screenshot data, this geometry will then be used to get a
 *	rectangular image from the root Window.
 *
 *	Returns 0 on success or non-zero on error.
 */
static gint grab_X_rectangular(
	ScreenshotData *data,
	imgview_struct *iv,
	const gboolean include_decorations,
	imgview_image_struct **img_rtn,
	gint *x_rtn, gint *y_rtn,
	gint *base_width_rtn, gint *base_height_rtn
)
{
	int		x_return, y_return;
	unsigned int    width_return, height_return,
			border_width_return,
			depth_return;
	Display *dpy;
	Window		xrootwin,
			xrootrtn;		/* Dummy root for fetches */
	GdkWindowPrivate *private;
	gint		status,
			cx, cy,
			cwidth, cheight,
			rwidth, rheight;

	if((iv == NULL) || (data == NULL))
		return(-1);

	dpy = data->xdisplay;
	xrootwin = data->xrootwin;
	if((dpy == NULL) || (xrootwin == None))
		return(-1);

	/* Finish processing any pending X operations to ensure that
	 * we do not haev any marks that we drew over the Window that
	 * we want to grab
	 */
	(void)XFlush(dpy);

	/* Grab the X server so that everything will hold still while
	 * we obtain the Window geometry
	 */
	(void)XGrabServer(dpy);

	/* Get the size of the root window */
	(void)XGetGeometry(
		dpy,
		xrootwin,
		&xrootrtn,
		&x_return, &y_return,
		&width_return, &height_return,
		&border_width_return,
		&depth_return
	);
	rwidth = (gint)width_return;
	rheight = (gint)height_return;

	/* Get the size of the child area */
	cx = MIN(data->x[0], data->x[1]);
	cy = MIN(data->y[0], data->y[1]);
	cwidth = ABSOLUTE_VALUE(data->x[1] - data->x[0]);
	cheight = ABSOLUTE_VALUE(data->y[1] - data->y[0]);

	/* At this point we now have the size of the root window in
	 * rwidth and rheight plus the child window geometry relative
	 * to the root window at x, y, cwidth, and cheight.
	 */
	/* Clip the values of the child window geometry */
	if((cx + cwidth) > rwidth)
		cwidth = rwidth - cx;
	if(cx < 0)
	{
		cwidth += cx;
		cx = 0;
	}
	if((cy + cheight) > rheight)
		cheight = rheight - cy;
	if(cy < 0)
	{
		cheight += cy;
		cy = 0;
	}
	/* Check if the specified grab rectangle has no area */
	if((cwidth <= 0) || (cheight <= 0))
	{
		last_screenshot_error =
			"Grab rectangle has zero area";
		(void)XUngrabServer(dpy);
		return(-1);
	}

	/* We now have the rectangular area that we want to grab
	 * specified in cx, cy, cwidth, and cheight
	 *
	 * Set the return geometry values
	 */
	if(x_rtn != NULL)
		*x_rtn = cx;
	if(y_rtn != NULL)
		*y_rtn = cy;
	if(base_width_rtn != NULL)
		*base_width_rtn = rwidth;
	if(base_height_rtn != NULL)
		*base_height_rtn = rheight;

	/* Reset the status to -1, indicating error. If we were able to
	 * get the image then we'll set it to 0
	 */
	status = -1;

	/* Create a GdkWindow structure to reference the xroot_win */
	private = (GdkWindowPrivate *)g_malloc0(sizeof(GdkWindowPrivate));
	if(private != NULL)
	{
		const gint bpp = 4;		/* RGBA */
		gint bpl;
		guint8 *data_rgba;
		GdkRectangle rect;
		GdkWindow *window = (GdkWindow *)private;

		/* Set up the GdkWindow values with the X values for grabbing */
		private->parent = NULL;
		private->xwindow = xrootwin;
		private->xdisplay = dpy;
		private->x = 0;
		private->y = 0;
		private->width = rwidth;
		private->height = rheight;
		private->mapped = 1;

		/* Set the grab area */
		rect.x = cx;
		rect.y = cy;
		rect.width = cwidth;
		rect.height = cheight;

		/* Grab and get the RGBA image data */
		data_rgba = gdk_get_rgba_image(
			window,
			NULL,				/* No mask */
			&rect,
			&cwidth, &cheight, &bpl
		);
		if(data_rgba != NULL)
		{
			imgview_image_struct *img = ImgViewImageNew(
				cwidth, cheight,
				bpp, bpl
			);
			if(img != NULL)
			{
				(void)ImgViewImageAppendFrame(
					img,
					data_rgba,		/* Transfered */
					0l			/* No delay */
				);
				*img_rtn = img;
				status = 0;
			}
			else
			{
				g_free(data_rgba);
				last_screenshot_error = "Memory allocation error";
				status = -3;
			}
		}
		else
		{
			last_screenshot_error = "Unable to grab an X Drawable";
		}

		g_free(private);
	}

	/* Ungrab the X server */
	(void)XUngrabServer(dpy);

	return(status);
}
#endif	/* HAVE_X */


/*
 *	Grab window destroyed notify.
 *
 *	User to abort the grab when the grab window has been
 *	destroyed.
 */
static gint screenshot_window_destroy_cb(
	GtkWidget *widget, GdkEvent *event, gpointer data
)
{
	ScreenshotData *d = SCREENSHOT_DATA(data);
	if((widget == NULL) || (d == NULL))
		return(FALSE);

	/* Was a button pressed? */
	if(d->grab_button_number > 0)
	{
		d->grab_button_number = 0;

		/* If grabbing a rectangular area then update the
		 * rectangle's end coordinates and redraw the rectangular
		 * outline
		 */
		if(d->target == SCREENSHOT_TARGET_RECTANGULAR)
		{
			d->x[1] = d->x[0];
			d->y[1] = d->y[0];
#if defined(HAVE_X)
			draw_X_rectangular_area(d, False);
#endif
		}
	}

	/* Mark that the grab was aborted due to the grab window
	 * being destroyed
	 */
	d->target = SCREENSHOT_TARGET_ABORT;

	/* Pop off any remaining GTK main levels */
	if(d->nmain_levels > 0)
	{
		d->nmain_levels--;
		gtk_main_quit();
	}

	return(FALSE);
}

/*
 *	Grab delay timeout callback.
 *
 *	This is called when the delay time expires in
 *	iv_screenshot_start() and to break out of the GTK main set
 *	just after the setting of this callback.
 */
static gint screenshot_delay_timeout_cb(gpointer data)
{
	gint *nmain_levels = (gint *)data;
	if(*nmain_levels > 0)
	{
		*nmain_levels = (*nmain_levels) - 1;
		gtk_main_quit();
	}

	return(FALSE);		/* Don't call this timeout again */
}

/*
 *	Grab "button_press_event" signal callback.
 */
static gint screenshot_button_cb(
	GtkWidget *widget, GdkEventButton *button, gpointer data
)
{
	ScreenshotData *d = SCREENSHOT_DATA(data);
	if((button == NULL) || (d == NULL))
		return(FALSE);

	/* Do not process event if grabbing is not schedualed to
	 * be active yet
	 */
	if(!d->active)
		return(TRUE);

	/* Button press? */
	if(button->type == GDK_BUTTON_PRESS)
	{
		/* Get number of the button that was pressed */
		const gint button_num = button->button;

		/* Set ScreenshotTarget by which button number was pressed */
		switch(button_num)
		{
		  case GDK_BUTTON1:
			d->grab_button_number = button_num;
			d->target = SCREENSHOT_TARGET_WINDOW;
			d->x[0] = (gint)button->x;
			d->y[0] = (gint)button->y;
			break;

		  case GDK_BUTTON2:
			d->grab_button_number = button_num;
			d->target = SCREENSHOT_TARGET_RECTANGULAR;
			d->x[0] = (gint)button->x;
			d->y[0] = (gint)button->y;
			d->x[1] = d->x[0];
			d->y[1] = d->y[0];
			break;

		  case GDK_BUTTON3:
			d->grab_button_number = button_num;
			d->target = SCREENSHOT_TARGET_ABORT;
			break;

		  case GDK_BUTTON4:
		  case GDK_BUTTON5:
			/* Ignore the pointer's wheel */
			break;

		  default:
			/* All other button numbers assume abort */
			d->grab_button_number = button_num;
			d->target = SCREENSHOT_TARGET_ABORT;
			break;
		}
	}
	/* Assume GDK_BUTTON_RELEASE */
	else
	{
		/* Get number of the button that was released */
		const gint button_num = button->button;

		/* Set ScreenshotTarget by which button number was pressed */
		switch(button_num)
		{
		  case GDK_BUTTON4:
		  case GDK_BUTTON5:
			/* Ignore the pointer's wheel */
			break;

		  default:
			/* Was a valid grab button pressed? */
			if(d->grab_button_number > 0)
			{
				d->grab_button_number = 0;

				/* If grabbing a rectangular area then update
				 * the rectangle's end coordinates and redraw the
				 * rectangle outline
				 */
				if(d->target == SCREENSHOT_TARGET_RECTANGULAR)
				{
					d->x[1] = (gint)button->x;
					d->y[1] = (gint)button->y;
#if defined(HAVE_X)
					draw_X_rectangular_area(d, False);
#endif
				}

				/* Pop off all of our main levels */
				if(d->nmain_levels > 0)
				{
					d->nmain_levels--;
					gtk_main_quit();
				}
			}
			break;
		}
	}

	return(TRUE);
}

/*
 *	Grab "motion_notify_event" signal callback.
 */
static gint screenshot_motion_cb(
	GtkWidget *widget, GdkEventMotion *motion, gpointer data
)
{
	ScreenshotData *d = SCREENSHOT_DATA(data);
	if((motion == NULL) || (d == NULL))
		return(FALSE);

	/* Do not process event if grabbing is not schedualed to
	 * be active yet
	 */
	if(!d->active)
		return(TRUE);

	/* Button pressed? */
	if(d->grab_button_number > 0)
	{
		switch(d->target)
		{
		  case SCREENSHOT_TARGET_UNDETERMINED:
			/* Pending */
			break;

		  case SCREENSHOT_TARGET_WINDOW:
			/* Do nothing */
			break;

		  case SCREENSHOT_TARGET_RECTANGULAR:
			/* Update the grab coordinates */
			d->x[1] = (gint)motion->x;
			d->y[1] = (gint)motion->y;
#if defined(HAVE_X)
			draw_X_rectangular_area(d, True);
#endif
			break;

		  case SCREENSHOT_TARGET_ABORT:
			/* Do nothing */
			break;
		}
	}

	return(TRUE);
}


/*
 *	Returns the pointer to a string describing the last error or
 *	NULL if there was no error.
 */
const gchar *iv_screenshot_get_error(void)
{
	return(last_screenshot_error);
}

/*
 *	Take a screenshot.
 *
 *	The iv specifies the ImgView that is taking the screenshot..
 *
 *	The pointer will be grabbed to query which window or rectangular
 *	area to grab.
 *
 *	The delay_ms specifies the delay in milliseconds to wait before
 *	starting the grab.
 *
 *	If include_decorations is TRUE then WM decorations will be
 *	included in the user clicked on a toplevel window.
 *
 *	If resize_on_load is TRUE then the iv will be resized after
 *	the grab.
 *
 *	The grab_window specifies the GtkWindow to refer grab values
 *	to, it must be a valid and mapped GtkWindow, and it will be
 *	used to grab the pointer and report relative events to.
 *
 *	The grab_cursor specifies the cursor to display during the
 *	grab.
 *
 *	The img_rtn specifies the return image containing the grabbed
 *	image data or NULL on error or user abort.
 *
 *	The x_rtn and y_rtn specifies the return values for the
 *	coordinates the image was taken relative to the desktop.
 *
 *	The base_width_rtn and base_height_rtn specifies the return
 *	values for the size of the desktop.
 *
 *	Return values:
 *
 *		0	Success.
 *		-1	General error (see iv_screenshot_get_error()).
 *		-2	User did not select grab window/area properly.
 *		-3	System error.
 *		-4	User aborted.
 *
 *	When non-zero is returned the exact error message can be
 *	retrieved by calling iv_screenshot_get_error().
 */
gint iv_screenshot_start(
	imgview_struct *iv,
	const gulong delay_ms,
	const gboolean include_decorations,
	GtkWidget *grab_window,
	GdkCursor *grab_cursor,
	imgview_image_struct **img_rtn,
	gint *x_rtn, gint *y_rtn,
	gint *base_width_rtn, gint *base_height_rtn
)
{
#if defined(HAVE_X)
	int scr_num;
	Display *dpy;
	Window xrootwin;
	XGCValues gcv;
	GC gc_inverse;
#endif	/* HAVE_X */
	gint		status,
			nmain_levels;
	guint grab_signal_id[4];
	GdkWindow	*window,
			*root_win;
	ScreenshotData *data;

	/* Reset the last grab error string */
	last_screenshot_error = NULL;

	/* Reset the return values */
	if(x_rtn != NULL)
		*x_rtn = 0;
	if(y_rtn != NULL)
		*y_rtn = 0;
	if(base_width_rtn != NULL)
		*base_width_rtn = 0;
	if(base_height_rtn != NULL)
		*base_height_rtn = 0;
	if(img_rtn == NULL)
	{
		last_screenshot_error = "Return image pointer is NULL";
		return(-2);
	}
	*img_rtn = NULL;

	if(iv == NULL)
	{
		last_screenshot_error = "ImgView is NULL";
		return(-2);
	}
	if((grab_window != NULL) ?
		GTK_WIDGET_NO_WINDOW(grab_window) : TRUE
	)
	{
		last_screenshot_error = "Grab window is NULL";
		return(-1);
	}

	/* Wait for the specified delay */
	nmain_levels = 0;
	if(delay_ms > 0l)
	{
		/* Set the break main delay timeout callback and push a
		 * main level
		 */
		nmain_levels++;
		(void)gtk_timeout_add(
			delay_ms,
			screenshot_delay_timeout_cb,	/* Timeout Callback */
			&nmain_levels			/* Data */
		);
		gtk_main();
	}
	while(nmain_levels > 0)
	{
		nmain_levels--;
		gtk_main_quit();
	}


	/* Get the root GdkWindow */
	root_win = gdk_window_get_root();
	if(root_win == NULL)
	{
		last_screenshot_error = "Unable to obtain the root rindow";
		return(-1);
	}

	/* Get the grab_window's GdkWindow */
	window = grab_window->window;
	if(window == NULL)
	{
		last_screenshot_error = "Grab window's GdkWindow is NULL";
		return(-1);
	}

#if defined(HAVE_X)
	/* Get the X resources */
	dpy = GDK_WINDOW_XDISPLAY(root_win);
	scr_num = DefaultScreen(dpy);
	xrootwin = GDK_WINDOW_XWINDOW(root_win);
	if((dpy == NULL) || (xrootwin == None))
	{
		last_screenshot_error = "Unable obtain X resources";
		return(-3);
	}

	/* Sync X to ensure most recent characteristics are being
	 * displayed just before grabbing
	 */
	XSync(dpy, False);
#endif	/* HAVE_X */

	/* Create the screenshot data */
	data = SCREENSHOT_DATA(g_malloc0(sizeof(ScreenshotData)));
	if(data == NULL)
	{
		last_screenshot_error = "Memory allocation error";
		return(-3);
	}
	data->target = SCREENSHOT_TARGET_UNDETERMINED;
	data->active = FALSE;
	data->root_win = root_win;
	data->nmain_levels = 0;
	data->grab_button_number = 0;
#if defined(HAVE_X)
	data->xdisplay = dpy;
	data->xscreen_num = scr_num;
	data->xrootwin = xrootwin;
	/* Create the X Graphics Context for grab related drawing */
	gcv.function = GXinvert;
	gcv.plane_mask = AllPlanes;
	gcv.foreground = WhitePixel(dpy, scr_num);
	gcv.background = BlackPixel(dpy, scr_num);
	gcv.line_width = 1;
	gcv.line_style = LineSolid;
	gcv.cap_style = CapNotLast;
	gcv.join_style = JoinMiter;
	gcv.fill_style = FillSolid;
	gcv.fill_rule = EvenOddRule;
	gcv.subwindow_mode = IncludeInferiors;
	data->gc_inverse = gc_inverse = XCreateGC(
		dpy, xrootwin,
		GCFunction | GCForeground | GCBackground | GCPlaneMask |
		GCLineWidth | GCLineStyle | GCCapStyle | GCJoinStyle |
		GCFillStyle | GCFillRule | GCSubwindowMode,
		&gcv
	);
	data->have_last_rect = False;
#endif	/* HAVE_X */

	/* Install signal handlers for the grab GtkWindow */
	grab_signal_id[0] = gtk_signal_connect(
		GTK_OBJECT(grab_window), "destroy_event",
		GTK_SIGNAL_FUNC(screenshot_window_destroy_cb), data
	);
	grab_signal_id[1] = gtk_signal_connect(
		GTK_OBJECT(grab_window), "button_press_event",
		GTK_SIGNAL_FUNC(screenshot_button_cb), data
	);
	grab_signal_id[2] = gtk_signal_connect(
		GTK_OBJECT(grab_window), "button_release_event",
		GTK_SIGNAL_FUNC(screenshot_button_cb), data
	);
	grab_signal_id[3] = gtk_signal_connect(
		GTK_OBJECT(grab_window), "motion_notify_event",
		GTK_SIGNAL_FUNC(screenshot_motion_cb), data
	);


#define UNGRAB_POINTER	{			\
 if(gdk_pointer_is_grabbed()) {			\
  gdk_flush();					\
  gdk_pointer_ungrab(GDK_CURRENT_TIME);		\
  gdk_flush();					\
 }						\
}

#if defined(HAVE_X)
#define SCREENSHOT_DATA_DELETE(_d_)	{	\
 if((_d_) != NULL) {				\
  /* Delete the X resources */			\
  if((_d_)->gc_inverse != None)			\
   XFreeGC(dpy, (_d_)->gc_inverse);		\
											\
  g_free((_d_));				\
 }						\
}
#else
#define SCREENSHOT_DATA_DELETE(_d_)	{	\
 if((_d_) != NULL) {				\
  g_free((_d_));				\
 }						\
}
#endif

/* Ungrab pointer, disconnect signal handlers, and delete the grab
 * callback data
 */
#define CLEANUP_RETURN(_v_)	{		\
 gint i, m;					\
											\
 /* Ungrab the pointer as needed */		\
 UNGRAB_POINTER					\
											\
 /* Disconnect our tempory signal handlers */	\
 m = sizeof(grab_signal_id) / sizeof(guint);	\
 for(i = 0; i < m; i++) {			\
  gtk_signal_disconnect(			\
   GTK_OBJECT(grab_window),			\
   grab_signal_id[i]				\
  );						\
  grab_signal_id[i] = (guint)-1;		\
 }						\
											\
 /* Pop off any remaining GTK main levels */	\
 while(data->nmain_levels > 0) {		\
  data->nmain_levels--;				\
  gtk_main_quit();				\
 }						\
											\
 /* Delete the screenshot data */		\
 SCREENSHOT_DATA_DELETE(data);			\
											\
 return(_v_);					\
}

	/* Now mark grab as active, this will allow grab related events 
	 * to be processed in the signal callbacks we set above
	 */
	data->active = TRUE;

	/* Begin grabbing the pointer so grab related events get reported
	 * to and with respect to the given grab window
	 */
	gdk_flush();	/* Must flush GDK output before grabbing pointer */
	status = gdk_pointer_grab(
		window,
		FALSE,
		GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK |
			GDK_POINTER_MOTION_MASK,
		/* root_win */ NULL,
		grab_cursor,
		GDK_CURRENT_TIME
	);
	gdk_flush();	/* Must flush GDK output after grabbing pointer */
	/* Was there an error while attempting to grab the pointer? */
	switch(status)
	{
#ifdef GrabInvalidTime
	  case GrabInvalidTime:
		last_screenshot_error =
			"Invalid time specified for grabbing pointer";
		CLEANUP_RETURN(-3);
		break;
#endif
#ifdef GrabNotViewable
	  case GrabNotViewable:
		last_screenshot_error =
			"Grab window is not viewable to grab pointer";
		CLEANUP_RETURN(-3);
		break;
#endif
#ifdef GrabFrozen
	  case GrabFrozen:
		last_screenshot_error =
			"Grab is frozen and unable to grab pointer";
		CLEANUP_RETURN(-3);
		break;
#endif
#ifdef AlreadyGrabbed
	  case AlreadyGrabbed:
		last_screenshot_error =
"Another window is currently grabbing the pointer, please\n\
wait and try again later";
		CLEANUP_RETURN(-3);
		break;
#endif
#ifdef GrabSuccess
	  case GrabSuccess:
		break;
#endif
	}

	/* Enter our GTK+ main loop */
	data->nmain_levels++;
	gtk_main();

	/* Mark grab as expired (no longer active) so no more grab
	 * related events will be processed on our signal callbacks
	 */
	data->active = FALSE;

	/* Ungrab the pointer, we do not need it anymore */
	UNGRAB_POINTER


	/* Handle by grab type, assume error first */
	last_screenshot_error = "Specified screenshot operation is not supported";
	status = -1;
	switch(data->target)
	{
	  case SCREENSHOT_TARGET_UNDETERMINED:
		break;

	  case SCREENSHOT_TARGET_WINDOW:
#if defined(HAVE_X)
		status = grab_X_window(
			data,
			iv,
			include_decorations,
			img_rtn,
			x_rtn, y_rtn,
			base_width_rtn, base_height_rtn
		);
#endif
		break;

	  case SCREENSHOT_TARGET_RECTANGULAR:
#if defined(HAVE_X)
		status = grab_X_rectangular(
			data,
			iv,
			include_decorations,
			img_rtn,
			x_rtn, y_rtn,
			base_width_rtn, base_height_rtn
		);
#endif
		break;

	  case SCREENSHOT_TARGET_ABORT:
		last_screenshot_error = "User aborted";
		CLEANUP_RETURN(-4);
		break;
	}

	/* Ungrab the pointer, disconnect any signal handlers, delete
	 * the screenshot data, and return success
	 */
	CLEANUP_RETURN(0);
#undef CLEANUP_RETURN
#undef UNGRAB_POINTER   
#undef SCREENSHOT_DATA_DELETE
}
